﻿module CodecTests

open NUnit.Framework
open System.Text
open Kafunk

let ofMessages ms =
  MessageSet(ms |> Seq.map (fun m -> MessageSetItem (0L, Message.Size m, m)) |> Seq.toArray)

module Message =

  let create value key attrs =
    Message(0, 0y, (defaultArg attrs 0y), 0L, key, value)

let arraySegToFSharp (arr:Binary.Segment) =
  let sb = StringBuilder()
  sb.Append("[| ") |> ignore
  for byte in arr do
    sb.AppendFormat("{0}uy;", byte) |> ignore
  sb.Append(" |]") |> ignore
  sb.ToString()

[<Test>]
let ``Binary.writeInt32 should encode int32 negative``() =
  let x = -1
  let buf = Binary.zeros 4
  Binary.writeInt32 x buf |> ignore
  let x2 = Binary.readInt32 buf |> fst
  Assert.AreEqual(x, x2)

let toArraySeg (size:'a -> int) (write:'a * BinaryZipper -> unit) (a:'a) =
  let size = size a
  let buf = Binary.zeros size
  let bz = BinaryZipper (buf)
  write (a,bz)
  buf

[<Test>]
let ``Crc.crc32 message``() =
  let messageVer = 0s
  let m = Message.create (Binary.ofArray "hello world"B) (Binary.empty) None
  let bytes = toArraySeg (fun m -> Message.Size m) (fun (m,buf) -> Message.Write (m,buf)) m
  let crc32 = Crc.crc32 bytes.Array (bytes.Offset + 4) (bytes.Count - 4)
  let expected = 1940715388u
  Assert.AreEqual(expected, crc32)

[<Test>]
let ``Message.ComputeCrc``() =
  let magicByte = 0y
  let m = Message.create (Binary.ofArray "hello world"B) (Binary.empty) None
  //let bytes = toArraySeg (Message.size messageVer) (Message.write messageVer) m
  let bytes = toArraySeg (fun m -> Message.Size m) (fun (m,buf) -> Message.Write (m,buf)) m
  let m2 = Message.Read (BinaryZipper(bytes))
  let crc32 = Message.ComputeCrc (m2)
  let expected = int 1940715388u
  Assert.AreEqual(expected, m2.crc)
  Assert.AreEqual(expected, crc32)
  

[<Test>]
let ``Crc.crc32 string``() =
  let bytes = "hello world"B
  let crc32 = Crc.crc32 bytes 0 bytes.Length
  let expected = 0x0D4A1185u
  Assert.AreEqual(expected, crc32)

[<Test>]
let ``MessageSet.write should encode MessageSet``() =
  let expected =
    [
        0uy; 0uy; 0uy; 0uy; 0uy; 0uy; 0uy; 0uy; 0uy; 0uy; 0uy; 16uy; 45uy;
        70uy; 24uy; 62uy; 0uy; 0uy; 0uy; 0uy; 0uy; 1uy; 49uy; 0uy; 0uy; 0uy;
        1uy; 48uy; 0uy; 0uy; 0uy; 0uy; 0uy; 0uy; 0uy; 0uy; 0uy; 0uy; 0uy; 16uy;
        90uy; 65uy; 40uy; 168uy; 0uy; 0uy; 0uy; 0uy; 0uy; 1uy; 49uy; 0uy; 0uy;
        0uy; 1uy; 49uy; 0uy; 0uy; 0uy; 0uy; 0uy; 0uy; 0uy; 0uy; 0uy; 0uy; 0uy;
        16uy; 195uy; 72uy; 121uy; 18uy; 0uy; 0uy; 0uy; 0uy; 0uy; 1uy; 49uy; 0uy;
        0uy; 0uy; 1uy; 50uy
    ]
  let ms =
    [
      Message.create (Binary.ofArray "0"B) (Binary.ofArray "1"B) None
      Message.create (Binary.ofArray "1"B) (Binary.ofArray "1"B) None
      Message.create (Binary.ofArray "2"B) (Binary.ofArray "1"B) None
    ]
    |> ofMessages
  let size = MessageSet.Size (ms)
  let data = Binary.zeros size
  let bz = BinaryZipper (data)
  MessageSet.Write (size, ms, bz) |> ignore
  //let data = toArraySeg (MessageSet.size 0s) (MessageSet.write 0s) ms
  let encoded = data |> Binary.toArray |> Array.toList
  Assert.True ((expected = encoded))


let FetchResponseBinary = 
  Binary.ofArray [|
    0uy;0uy;0uy;1uy;0uy;4uy;116uy;101uy;115uy;116uy;0uy;0uy;0uy;1uy;0uy;0uy;
    0uy;0uy;0uy;0uy;0uy;0uy;0uy;0uy;0uy;0uy;0uy;8uy;0uy;0uy;0uy;37uy;0uy;0uy;
    0uy;0uy;0uy;0uy;0uy;7uy;0uy;0uy;0uy;25uy;115uy;172uy;247uy;124uy;0uy;0uy;
    255uy;255uy;255uy;255uy;0uy;0uy;0uy;11uy;104uy;101uy;108uy;108uy;111uy;
    32uy;119uy;111uy;114uy;108uy;100uy; |]

[<Test>]
let ``FetchResponse.read should decode FetchResponse``() =
  let data = FetchResponseBinary
  let data = BinaryZipper(data)
  let (res:FetchResponse) = FetchResponse.Read (0s, data)
  let topicName, ps = res.topics.[0]
  //let p, ec, _hwo, _, _, _, mss, ms = ps.[0]
  let p = ps.[0]   
  let ec = p.errorCode
  let mss = p.messageSetSize
  let ms = p.messageSet
  let p = p.partition
  //let o, _ms, m = ms.messages.[0]
  let x = ms.messages.[0]
  let o, m = x.offset, x.message
  Assert.AreEqual("test", topicName)
  Assert.AreEqual(p, 0)
  Assert.AreEqual(ec, 0s)
  Assert.AreEqual(mss, 37)
  Assert.AreEqual(o, 7L)
  Assert.AreEqual(1940715388, int m.crc)
  Assert.AreEqual("hello world", (m.value |> Binary.toString))

////[<Test>]
//let ``FetchResponse.read should decode FetchResponse API version 2``() =

//    let v2ReponseData =
//        Binary.ofArray [|
//            0uy;0uy;0uy;0uy         // throttle_time_ms (0ms)
//            0uy;0uy;0uy;1uy         // response array size (1 topic)
//            0uy;4uy                 // topic #1 name length (4 chars)
//            116uy;101uy;115uy;116uy // 't' 'e' 's' 't'
//            0uy;0uy;0uy;1uy         // topics[0] partition array size (1 partition)
//            0uy;0uy;0uy;1uy         // topics[0].partition[0].partition (#1)
//            0uy;0uy                 // topics[0].partition[0].error code (0)
//            0uy;0uy;0uy;0uy;0uy;0uy;6uy;95uy // topics[0].partition[0].highwatermark (1631)
//            0uy;0uy;1uy;64uy;       // "message set size"
//            0uy;0uy;0uy;0uy;0uy;0uy;0uy;64uy; // offset
//            0uy;0uy;114uy;181uy     // RecordBatch.Length
//            0uy;0uy;1uy;4uy         // RecordBatch.Partition Leader Epoch
//            2uy                     // Magic byte
//            0uy;0uy;0uy;0uy         // CRC32 sum TODO
//            0uy;8uy                 // batch attributes
//            0uy;0uy;0uy;0uy         // last offset delta
//            255uy;255uy;255uy;255uy;255uy;255uy;255uy;255uy // first timestamp
//            255uy;255uy;255uy;255uy;255uy;255uy;255uy;255uy // max timestamp
//            255uy;255uy;255uy;255uy;255uy;255uy;255uy;255uy // producer ID
//            255uy;255uy // producer epoch
//            255uy;255uy;255uy;255uy // first sequence
//            0uy;0uy;0uy;1uy         // number of records
//            130uy;4uy               // record length
//            0uy                     // record attributes
//            0uy                     // timestamp delta
//            0uy                     // offset delta
//            70uy                    // key length
//            0x31uy;0x30uy;0x36uy;0x30uy;0x31uy;0x30uy;0x63uy;0x65uy;0x38uy;0x66uy;0x66uy
//            0x36uy;0x34uy;0x35uy;0x39uy;0x66uy;0x62uy;0x37uy;0x63uy;0x66uy;0x34uy;0x37uy
//            0x64uy;0x34uy;0x32uy;0x35uy;0x31uy;0x33uy;0x38uy;0x30uy;0x64uy;0x31uy;0x3auy
//            0x32uy;0x31uy           // key (retail SKU ID : 21)
//            0xAEuy;0x03uy           // message length
//            0x7buy;0x22uy;0x72uy;0x65uy;0x74uy;0x61uy;0x69uy;0x6cuy;0x53uy;0x6buy;0x75uy
//            0x49uy;0x64uy;0x22uy;0x3auy;0x22uy;0x31uy;0x30uy;0x36uy;0x30uy;0x31uy;0x30uy
//            0x63uy;0x65uy;0x38uy;0x66uy;0x66uy;0x36uy;0x34uy;0x35uy;0x39uy;0x66uy;0x62uy
//            0x37uy;0x63uy;0x66uy;0x34uy;0x37uy;0x64uy;0x34uy;0x32uy;0x35uy;0x31uy;0x33uy
//            0x38uy;0x30uy;0x64uy;0x31uy;0x22uy;0x2cuy;0x22uy;0x6duy;0x61uy;0x72uy;0x74uy
//            0x49uy;0x64uy;0x22uy;0x3auy;0x32uy;0x31uy;0x2cuy;0x22uy;0x72uy;0x65uy;0x63uy
//            0x61uy;0x6cuy;0x63uy;0x75uy;0x6cuy;0x61uy;0x74uy;0x69uy;0x6fuy;0x6euy;0x54uy
//            0x68uy;0x72uy;0x65uy;0x73uy;0x68uy;0x6fuy;0x6cuy;0x64uy;0x22uy;0x3auy;0x30uy
//            0x2cuy;0x22uy;0x74uy;0x69uy;0x6duy;0x65uy;0x73uy;0x74uy;0x61uy;0x6duy;0x70uy
//            0x22uy;0x3auy;0x22uy;0x32uy;0x30uy;0x31uy;0x38uy;0x2duy;0x30uy;0x32uy;0x2duy
//            0x30uy;0x32uy;0x54uy;0x32uy;0x32uy;0x3auy;0x30uy;0x36uy;0x3auy;0x31uy;0x37uy
//            0x5auy;0x22uy;0x2cuy;0x22uy;0x72uy;0x65uy;0x61uy;0x73uy;0x6fuy;0x6euy;0x22uy
//            0x3auy;0x22uy;0x45uy;0x76uy;0x65uy;0x6euy;0x74uy;0x22uy;0x2cuy;0x22uy;0x63uy
//            0x6fuy;0x6duy;0x6duy;0x65uy;0x6euy;0x74uy;0x22uy;0x3auy;0x22uy;0x52uy;0x65uy
//            0x70uy;0x72uy;0x69uy;0x63uy;0x65uy;0x20uy;0x61uy;0x6cuy;0x6cuy;0x20uy;0x66uy
//            0x72uy;0x6fuy;0x6duy;0x20uy;0x4euy;0x6fuy;0x76uy;0x61uy;0x22uy;0x2cuy;0x22uy
//            0x69uy;0x64uy;0x22uy;0x3auy;0x22uy;0x66uy;0x65uy;0x62uy;0x37uy;0x30uy;0x66uy
//            0x63uy;0x32uy;0x65uy;0x65uy;0x61uy;0x31uy;0x34uy;0x32uy;0x66uy;0x39uy;0x39uy
//            0x66uy;0x30uy;0x34uy;0x62uy;0x38uy;0x63uy;0x32uy;0x39uy;0x38uy;0x36uy;0x31uy
//            0x61uy;0x34uy;0x62uy;0x30uy;0x22uy;0x7duy  // message
//            0uy                     // headers length TODO parse this!
//        |]
//        |> BinaryZipper

//    let res = FetchResponse.Read (2s, v2ReponseData)
//    let (topicName, topicData) = res.topics.[0]
//    let (partition, errorCode, highWatermark, _, _, _, messageSetSize, messageSet) = topicData.[0]
//    let x = messageSet.messages.[0]

//    Assert.AreEqual(0, res.throttleTime)
//    Assert.AreEqual(1, res.topics.Length)
//    Assert.AreEqual(1, topicData.Length)
//    Assert.AreEqual("test", topicName)
//    Assert.AreEqual(1, partition)
//    Assert.AreEqual(0, errorCode)
//    Assert.AreEqual(1631, highWatermark)
//    Assert.AreEqual(320, messageSetSize)

//    Assert.AreEqual(64, x.offset)
//    Assert.AreEqual(215, x.message.value.Count)

////[<Test>]
//let ``FetchResponse.read should decode unexpected RecordBatch response from v2 API call``() =
//  let responseData =
//    Binary.ofArray [|
//      0x00uy;0x00uy;0x00uy;
//      0x00uy;0x00uy;0x00uy;0x00uy;0x01uy;0x00uy;0x28uy;0x73uy;0x75uy;0x70uy;0x65uy;
//      0x72uy;0x6duy;0x61uy;0x6euy;0x2duy;0x6duy;0x65uy;0x6duy;0x62uy;0x65uy;0x72uy;
//      0x2duy;0x70uy;0x72uy;0x69uy;0x63uy;0x65uy;0x2duy;0x72uy;0x65uy;0x71uy;0x75uy;
//      0x65uy;0x73uy;0x74uy;0x2duy;0x73uy;0x68uy;0x6fuy;0x65uy;0x62uy;0x75uy;0x79uy;
//      0x2duy;0x70uy;0x34uy;0x00uy;0x00uy;0x00uy;0x01uy;0x00uy;0x00uy;0x00uy;0x01uy;
//      0x00uy;0x00uy;0x00uy;0x00uy;0x00uy;0x00uy;0x00uy;0x00uy;0x06uy;0x5fuy;0x00uy;
//      0x00uy;0x49uy;0xd4uy;0x00uy;0x00uy;0x00uy;0x00uy;0x00uy;0x00uy;0x00uy;0x40uy;
//      0x00uy;0x00uy;0x72uy;0xb5uy;0x00uy;0x00uy;0x01uy;0x04uy;0x02uy;0x6fuy;0xf9uy;
//      0xa8uy;0x82uy;0x00uy;0x08uy;0x00uy;0x00uy;0x00uy;0x70uy;0xffuy;0xffuy;0xffuy;
//      0xffuy;0xffuy;0xffuy;0xffuy;0xffuy;0x00uy;0x00uy;0x01uy;0x61uy;0x58uy;0x8euy;
//      0x54uy;0x9duy;0xffuy;0xffuy;0xffuy;0xffuy;0xffuy;0xffuy;0xffuy;0xffuy;0xffuy;
//      0xffuy;0xffuy;0xffuy;0xffuy;0xffuy;0x00uy;0x00uy;0x00uy;0x71uy;0x82uy;0x04uy;
//      0x00uy;0x00uy;0x00uy;0x46uy;0x31uy;0x30uy;0x36uy;0x30uy;0x31uy;0x30uy;0x63uy;
//      0x65uy;0x38uy;0x66uy;0x66uy;0x36uy;0x34uy;0x35uy;0x39uy;0x66uy;0x62uy;0x37uy;
//      0x63uy;0x66uy;0x34uy;0x37uy;0x64uy;0x34uy;0x32uy;0x35uy;0x31uy;0x33uy;0x38uy;
//      0x30uy;0x64uy;0x31uy;0x3auy;0x32uy;0x31uy;0xaeuy;0x03uy;0x7buy;0x22uy;0x72uy;
//      0x65uy;0x74uy;0x61uy;0x69uy;0x6cuy;0x53uy;0x6buy;0x75uy;0x49uy;0x64uy;0x22uy;
//      0x3auy;0x22uy;0x31uy;0x30uy;0x36uy;0x30uy;0x31uy;0x30uy;0x63uy;0x65uy;0x38uy;
//      0x66uy;0x66uy;0x36uy;0x34uy;0x35uy;0x39uy;0x66uy;0x62uy;0x37uy;0x63uy;0x66uy;
//      0x34uy;0x37uy;0x64uy;0x34uy;0x32uy;0x35uy;0x31uy;0x33uy;0x38uy;0x30uy;0x64uy;
//      0x31uy;0x22uy;0x2cuy;0x22uy;0x6duy;0x61uy;0x72uy;0x74uy;0x49uy;0x64uy;0x22uy;
//      0x3auy;0x32uy;0x31uy;0x2cuy;0x22uy;0x72uy;0x65uy;0x63uy;0x61uy;0x6cuy;0x63uy;
//      0x75uy;0x6cuy;0x61uy;0x74uy;0x69uy;0x6fuy;0x6euy;0x54uy;0x68uy;0x72uy;0x65uy;
//      0x73uy;0x68uy;0x6fuy;0x6cuy;0x64uy;0x22uy;0x3auy;0x30uy;0x2cuy;0x22uy;0x74uy;
//      0x69uy;0x6duy;0x65uy;0x73uy;0x74uy;0x61uy;0x6duy;0x70uy;0x22uy;0x3auy;0x22uy;
//      0x32uy;0x30uy;0x31uy;0x38uy;0x2duy;0x30uy;0x32uy;0x2duy;0x30uy;0x32uy;0x54uy;
//      0x32uy;0x32uy;0x3auy;0x30uy;0x36uy;0x3auy;0x31uy;0x37uy;0x5auy;0x22uy;0x2cuy;
//      0x22uy;0x72uy;0x65uy;0x61uy;0x73uy;0x6fuy;0x6euy;0x22uy;0x3auy;0x22uy;0x45uy;
//      0x76uy;0x65uy;0x6euy;0x74uy;0x22uy;0x2cuy;0x22uy;0x63uy;0x6fuy;0x6duy;0x6duy;
//      0x65uy;0x6euy;0x74uy;0x22uy;0x3auy;0x22uy;0x52uy;0x65uy;0x70uy;0x72uy;0x69uy;
//      0x63uy;0x65uy;0x20uy;0x61uy;0x6cuy;0x6cuy;0x20uy;0x66uy;0x72uy;0x6fuy;0x6duy;
//      0x20uy;0x4euy;0x6fuy;0x76uy;0x61uy;0x22uy;0x2cuy;0x22uy;0x69uy;0x64uy;0x22uy;
//      0x3auy;0x22uy;0x66uy;0x65uy;0x62uy;0x37uy;0x30uy;0x66uy;0x63uy;0x32uy;0x65uy;
//      0x65uy;0x61uy;0x31uy;0x34uy;0x32uy;0x66uy;0x39uy;0x39uy;0x66uy;0x30uy;0x34uy;
//      0x62uy;0x38uy;0x63uy;0x32uy;0x39uy;0x38uy;0x36uy;0x31uy;0x61uy;0x34uy;0x62uy;
//      0x30uy;0x22uy;0x7duy;0x00uy;0x82uy;0x04uy;0x00uy;0x00uy;0x02uy;0x46uy;0x31uy;
//      0x31uy;0x36uy;0x30uy;0x31uy;0x32uy;0x33uy;0x37uy;0x61uy;0x64uy;0x64uy;0x66uy;
//      0x34uy;0x35uy;0x65uy;0x63uy;0x62uy;0x64uy;0x32uy;0x34uy;0x37uy;0x38uy;0x62uy;
//      0x38uy;0x65uy;0x36uy;0x64uy;0x32uy;0x35uy;0x32uy;0x63uy;0x36uy;0x3auy;0x32uy;
//      0x31uy;0xaeuy;0x03uy;0x7buy;0x22uy;0x72uy;0x65uy;0x74uy;0x61uy;0x69uy;0x6cuy;
//      0x53uy;0x6buy;0x75uy;0x49uy;0x64uy;0x22uy;0x3auy;0x22uy;0x31uy;0x31uy;0x36uy;
//      0x30uy;0x31uy;0x32uy;0x33uy;0x37uy;0x61uy;0x64uy;0x64uy;0x66uy;0x34uy;0x35uy;
//      0x65uy;0x63uy;0x62uy;0x64uy;0x32uy;0x34uy;0x37uy;0x38uy;0x62uy;0x38uy;0x65uy;
//      0x36uy;0x64uy;0x32uy;0x35uy;0x32uy;0x63uy;0x36uy;0x22uy;0x2cuy;0x22uy;0x6duy;
//      0x61uy;0x72uy;0x74uy;0x49uy;0x64uy;0x22uy;0x3auy;0x32uy;0x31uy;0x2cuy;0x22uy;
//      0x72uy;0x65uy;0x63uy;0x61uy;0x6cuy;0x63uy;0x75uy;0x6cuy;0x61uy;0x74uy;0x69uy;
//      0x6fuy;0x6euy;0x54uy;0x68uy;0x72uy;0x65uy;0x73uy;0x68uy;0x6fuy;0x6cuy;0x64uy;
//      0x22uy;0x3auy;0x30uy;0x2cuy;0x22uy;0x74uy;0x69uy;0x6duy;0x65uy;0x73uy;0x74uy;
//      0x61uy;0x6duy;0x70uy;0x22uy;0x3auy;0x22uy;0x32uy;0x30uy;0x31uy;0x38uy;0x2duy;
//      0x30uy;0x32uy;0x2duy;0x30uy;0x32uy;0x54uy;0x32uy;0x32uy;0x3auy;0x30uy;0x36uy;
//      0x3auy;0x31uy;0x37uy;0x5auy;0x22uy;0x2cuy;0x22uy;0x72uy;0x65uy;0x61uy;0x73uy;
//      0x6fuy;0x6euy;0x22uy;0x3auy;0x22uy;0x45uy;0x76uy;0x65uy;0x6euy;0x74uy;0x22uy;
//      0x2cuy;0x22uy;0x63uy;0x6fuy;0x6duy;0x6duy;0x65uy;0x6euy;0x74uy;0x22uy;0x3auy;
//      0x22uy;0x52uy;0x65uy;0x70uy;0x72uy;0x69uy;0x63uy;0x65uy;0x20uy;0x61uy;0x6cuy;
//      0x6cuy;0x20uy;0x66uy;0x72uy;0x6fuy;0x6duy;0x20uy;0x4euy;0x6fuy;0x76uy;0x61uy;
//      0x22uy;0x2cuy;0x22uy;0x69uy;0x64uy;0x22uy;0x3auy;0x22uy;0x37uy;0x64uy;0x62uy;
//      0x66uy;0x65uy;0x33uy;0x34uy;0x61uy;0x34uy;0x65uy;0x32uy;0x31uy;0x34uy;0x61uy;
//      0x31uy;0x34uy;0x61uy;0x32uy;0x34uy;0x32uy;0x34uy;0x36uy;0x63uy;0x38uy;0x63uy;
//      0x37uy;0x32uy;0x34uy;0x31uy;0x31uy;0x62uy;0x63uy;0x22uy;0x7duy;0x00uy;0x82uy;
//      0x04uy;0x00uy;0x00uy;0x04uy;0x46uy;0x31uy;0x31uy;0x65uy;0x30uy;0x30uy;0x64uy;
//      0x31uy;0x30uy;0x33uy;0x38uy;0x30uy;0x31uy;0x34uy;0x33uy;0x39uy;0x37uy;0x61uy;
//      0x39uy;0x32uy;0x61uy;0x39uy;0x38uy;0x34uy;0x64uy;0x33uy;0x63uy;0x30uy;0x35uy;
//      0x61uy;0x36uy;0x62uy;0x32uy;0x3auy;0x32uy;0x31uy;0xaeuy;0x03uy;0x7buy;0x22uy;
//      0x72uy;0x65uy;0x74uy;0x61uy;0x69uy;0x6cuy;0x53uy;0x6buy;0x75uy;0x49uy;0x64uy;
//      0x22uy;0x3auy;0x22uy;0x31uy;0x31uy;0x65uy;0x30uy;0x30uy;0x64uy;0x31uy;0x30uy;
//      0x33uy;0x38uy;0x30uy;0x31uy;0x34uy;0x33uy;0x39uy;0x37uy;0x61uy;0x39uy;0x32uy;
//      0x61uy;0x39uy;0x38uy;0x34uy;0x64uy;0x33uy;0x63uy;0x30uy;0x35uy;0x61uy;0x36uy;
//      0x62uy;0x32uy;0x22uy;0x2cuy;0x22uy;0x6duy;0x61uy;0x72uy;0x74uy;0x49uy;0x64uy;
//      0x22uy;0x3auy;0x32uy;0x31uy;0x2cuy;0x22uy;0x72uy;0x65uy;0x63uy;0x61uy;0x6cuy;
//      0x63uy;0x75uy;0x6cuy;0x61uy;0x74uy;0x69uy;0x6fuy;0x6euy;0x54uy;0x68uy;0x72uy;
//      0x65uy;0x73uy;0x68uy;0x6fuy;0x6cuy;0x64uy;0x22uy;0x3auy;0x30uy;0x2cuy;0x22uy;
//      0x74uy;0x69uy;0x6duy;0x65uy;0x73uy;0x74uy;0x61uy;0x6duy;0x70uy;0x22uy;0x3auy;
//      0x22uy;0x32uy;0x30uy;0x31uy;0x38uy;0x2duy;0x30uy;0x32uy;0x2duy;0x30uy;0x32uy;
//      0x54uy;0x32uy;0x32uy;0x3auy;0x30uy;0x36uy;0x3auy;0x31uy;0x37uy;0x5auy;0x22uy;
//      0x2cuy;0x22uy;0x72uy;0x65uy;0x61uy;0x73uy;0x6fuy;0x6euy;0x22uy;0x3auy;0x22uy;
//      0x45uy;0x76uy;0x65uy;0x6euy;0x74uy;0x22uy;0x2cuy;0x22uy;0x63uy;0x6fuy;0x6duy;
//      0x6duy;0x65uy;0x6euy;0x74uy;0x22uy;0x3auy;0x22uy;0x52uy;0x65uy;0x70uy;0x72uy;
//      0x69uy;0x63uy;0x65uy;0x20uy;0x61uy;0x6cuy;0x6cuy;0x20uy;0x66uy;0x72uy;0x6fuy;
//      0x6duy;0x20uy;0x4euy;0x6fuy;0x76uy;0x61uy;0x22uy;0x2cuy;0x22uy;0x69uy;0x64uy;
//      0x22uy;0x3auy;0x22uy;0x31uy;0x33uy;0x63uy;0x66uy;0x37uy;0x33uy;0x32uy;0x32uy;
//      0x30uy;0x39uy;0x62uy;0x36uy;0x34uy;0x36uy;0x61uy;0x37uy;0x38uy;0x39uy;0x35uy;
//      0x38uy;0x66uy;0x34uy;0x33uy;0x65uy;0x66uy;0x64uy;0x36uy;0x36uy;0x32uy;0x36uy;
//      0x33uy;0x63uy;0x22uy;0x7duy;0x00uy;0x82uy;0x04uy;0x00uy;0x00uy;0x06uy;0x46uy;
//      0x31uy;0x32uy;0x34uy;0x30uy;0x30uy;0x34uy;0x36uy;0x30uy;0x63uy;0x66uy;0x63uy;
//      0x34uy;0x34uy;0x66uy;0x64uy;0x34uy;0x61uy;0x64uy;0x64uy;0x36uy;0x65uy;0x34uy;
//      0x32uy;0x36uy;0x63uy;0x39uy;0x65uy;0x33uy;0x36uy;0x65uy;0x39uy;0x66uy;0x3auy;
//      0x32uy;0x31uy;0xaeuy;0x03uy;0x7buy;0x22uy;0x72uy;0x65uy;0x74uy;0x61uy;0x69uy;
//      0x6cuy;0x53uy;0x6buy;0x75uy;0x49uy;0x64uy;0x22uy;0x3auy;0x22uy;0x31uy;0x32uy;
//      0x34uy;0x30uy;0x30uy;0x34uy;0x36uy;0x30uy;0x63uy;0x66uy;0x63uy;0x34uy;0x34uy;
//      0x66uy;0x64uy;0x34uy;0x61uy;0x64uy;0x64uy;0x36uy;0x65uy;0x34uy;0x32uy;0x36uy;
//      0x63uy;0x39uy;0x65uy;0x33uy;0x36uy;0x65uy;0x39uy;0x66uy;0x22uy;0x2cuy;0x22uy;
//      0x6duy;0x61uy;0x72uy;0x74uy;0x49uy;0x64uy;0x22uy;0x3auy;0x32uy;0x31uy;0x2cuy;
//      0x22uy;0x72uy;0x65uy;0x63uy;0x61uy;0x6cuy;0x63uy;0x75uy;0x6cuy;0x61uy;0x74uy;
//      0x69uy;0x6fuy;0x6euy;0x54uy;0x68uy;0x72uy;0x65uy;0x73uy;0x68uy;0x6fuy;0x6cuy;
//      0x64uy;0x22uy;0x3auy;0x30uy;0x2cuy;0x22uy;0x74uy;0x69uy;0x6duy;0x65uy;0x73uy;
//      0x74uy;0x61uy;0x6duy;0x70uy;0x22uy;0x3auy;0x22uy;0x32uy;0x30uy;0x31uy;0x38uy;
//      0x2duy;0x30uy;0x32uy;0x2duy;0x30uy;0x32uy;0x54uy;0x32uy;0x32uy;0x3auy;0x30uy;
//      0x36uy;0x3auy;0x31uy;0x37uy;0x5auy;0x22uy;0x2cuy;0x22uy;0x72uy;0x65uy;0x61uy;
//      0x73uy;0x6fuy;0x6euy;0x22uy;0x3auy;0x22uy;0x45uy;0x76uy;0x65uy;0x6euy;0x74uy;
//      0x22uy;0x2cuy;0x22uy;0x63uy;0x6fuy;0x6duy;0x6duy;0x65uy;0x6euy;0x74uy;0x22uy;
//      0x3auy;0x22uy;0x52uy;0x65uy;0x70uy;0x72uy;0x69uy;0x63uy;0x65uy;0x20uy;0x61uy;
//      0x6cuy;0x6cuy;0x20uy;0x66uy;0x72uy;0x6fuy;0x6duy;0x20uy;0x4euy;0x6fuy;0x76uy;
//      0x61uy;0x22uy;0x2cuy;0x22uy;0x69uy;0x64uy;0x22uy;0x3auy;0x22uy;0x33uy;0x65uy;
//      0x36uy;0x39uy;0x30uy;0x63uy;0x36uy;0x31uy;0x62uy;0x31uy;0x37uy;0x38uy;0x34uy;
//      0x32uy;0x34uy;0x65uy;0x38uy;0x37uy;0x66uy;0x31uy;0x66uy;0x30uy;0x39uy;0x32uy;
//      0x35uy;0x38uy;0x61uy;0x62uy;0x30uy;0x62uy;0x31uy;0x30uy;0x22uy;0x7duy;0x00uy;
//      0x82uy;0x04uy;0x00uy;0x00uy;0x08uy;0x46uy;0x31uy;0x33uy;0x36uy;0x30uy;0x30uy;
//      0x36uy;0x39uy;0x37uy;0x37uy;0x62uy;0x31uy;0x32uy;0x34uy;0x61uy;0x31uy;0x63uy;
//      0x38uy;0x31uy;0x39uy;0x31uy;0x31uy;0x31uy;0x37uy;0x33uy;0x33uy;0x63uy;0x34uy;
//      0x33uy;0x37uy;0x33uy;0x63uy;0x37uy;0x3auy;0x32uy;0x31uy;0xaeuy;0x03uy;0x7buy;
//      0x22uy;0x72uy;0x65uy;0x74uy;0x61uy;0x69uy;0x6cuy;0x53uy;0x6buy;0x75uy;0x49uy;
//      0x64uy;0x22uy;0x3auy;0x22uy;0x31uy;0x33uy;0x36uy;0x30uy;0x30uy;0x36uy;0x39uy;
//      0x37uy;0x37uy;0x62uy;0x31uy;0x32uy;0x34uy;0x61uy;0x31uy;0x63uy;0x38uy;0x31uy;
//      0x39uy;0x31uy;0x31uy;0x31uy;0x37uy;0x33uy;0x33uy;0x63uy;0x34uy;0x33uy;0x37uy;
//      0x33uy;0x63uy;0x37uy;0x22uy;0x2cuy;0x22uy;0x6duy;0x61uy;0x72uy;0x74uy;0x49uy;
//      0x64uy;0x22uy;0x3auy;0x32uy;0x31uy;0x2cuy;0x22uy;0x72uy;0x65uy;0x63uy;0x61uy;
//      0x6cuy;0x63uy;0x75uy;0x6cuy;0x61uy;0x74uy;0x69uy;0x6fuy;0x6euy;0x54uy;0x68uy;
//      0x72uy;0x65uy;0x73uy;0x68uy;0x6fuy;0x6cuy;0x64uy;0x22uy;0x3auy;0x30uy;0x2cuy;
//      0x22uy;0x74uy;0x69uy;0x6duy;0x65uy;0x73uy;0x74uy;0x61uy;0x6duy;0x70uy;0x22uy;
//      0x3auy;0x22uy;0x32uy;0x30uy;0x31uy;0x38uy;0x2duy;0x30uy;0x32uy;0x2duy;0x30uy;
//      0x32uy;0x54uy;0x32uy;0x32uy;0x3auy;0x30uy;0x36uy;0x3auy;0x31uy;0x37uy;0x5auy;
//      0x22uy;0x2cuy;0x22uy;0x72uy;0x65uy;0x61uy;0x73uy;0x6fuy;0x6euy;0x22uy;0x3auy;
//      0x22uy;0x45uy;0x76uy;0x65uy;0x6euy;0x74uy;0x22uy;0x2cuy;0x22uy;0x63uy;0x6fuy;
//      0x6duy;0x6duy;0x65uy;0x6euy;0x74uy;0x22uy;0x3auy;0x22uy;0x52uy;0x65uy;0x70uy;
//      0x72uy;0x69uy;0x63uy;0x65uy;0x20uy;0x61uy;0x6cuy;0x6cuy;0x20uy;0x66uy;0x72uy;
//      0x6fuy;0x6duy;0x20uy;0x4euy;0x6fuy;0x76uy;0x61uy;0x22uy;0x2cuy;0x22uy;0x69uy;
//      0x64uy;0x22uy;0x3auy;0x22uy;0x31uy;0x37uy;0x30uy;0x31uy;0x37uy;0x33uy;0x37uy;
//      0x62uy;0x62uy;0x37uy;0x63uy;0x30uy;0x34uy;0x36uy;0x38uy;0x66uy;0x61uy;0x34uy;
//      0x34uy;0x35uy;0x30uy;0x31uy;0x62uy;0x66uy;0x66uy;0x65uy;0x61uy;0x37uy;0x64uy;
//      0x39uy;0x32uy;0x35uy;0x22uy;0x7duy;0x00uy;0x82uy;0x04uy;0x00uy;0x00uy;0x0auy;
//      0x46uy;0x31uy;0x33uy;0x65uy;0x30uy;0x30uy;0x32uy;0x31uy;0x66uy;0x30uy;0x31uy;
//      0x33uy;0x30uy;0x34uy;0x38uy;0x66uy;0x35uy;0x39uy;0x33uy;0x65uy;0x31uy;0x61uy;
//      0x34uy;0x64uy;0x62uy;0x36uy;0x64uy;0x64uy;0x35uy;0x62uy;0x66uy;0x61uy;0x34uy;
//      0x3auy;0x32uy;0x31uy;0xaeuy;0x03uy;0x7buy;0x22uy;0x72uy;0x65uy;0x74uy;0x61uy;
//      0x69uy;0x6cuy;0x53uy;0x6buy;0x75uy;0x49uy;0x64uy;0x22uy;0x3auy;0x22uy;0x31uy;
//      0x33uy;0x65uy;0x30uy;0x30uy;0x32uy;0x31uy;0x66uy;0x30uy;0x31uy;0x33uy;0x30uy;
//      0x34uy;0x38uy;0x66uy;0x35uy;0x39uy;0x33uy;0x65uy;0x31uy;0x61uy;0x34uy;0x64uy;
//      0x62uy;0x36uy;0x64uy;0x64uy;0x35uy;0x62uy;0x66uy;0x61uy;0x34uy;0x22uy;0x2cuy;
//      0x22uy;0x6duy;0x61uy;0x72uy;0x74uy;0x49uy;0x64uy;0x22uy;0x3auy;0x32uy;0x31uy;
//      0x2cuy;0x22uy;0x72uy;0x65uy;0x63uy;0x61uy;0x6cuy;0x63uy;0x75uy;0x6cuy;0x61uy;
//      0x74uy;0x69uy;0x6fuy;0x6euy;0x54uy;0x68uy;0x72uy;0x65uy;0x73uy;0x68uy;0x6fuy;
//      0x6cuy;0x64uy;0x22uy;0x3auy;0x30uy;0x2cuy;0x22uy;0x74uy;0x69uy;0x6duy;0x65uy;
//      0x73uy;0x74uy;0x61uy;0x6duy;0x70uy;0x22uy;0x3auy;0x22uy;0x32uy;0x30uy;0x31uy;
//      0x38uy;0x2duy;0x30uy;0x32uy;0x2duy;0x30uy;0x32uy;0x54uy;0x32uy;0x32uy;0x3auy;
//      0x30uy;0x36uy;0x3auy;0x31uy;0x37uy;0x5auy;0x22uy;0x2cuy;0x22uy;0x72uy;0x65uy;
//      0x61uy;0x73uy;0x6fuy;0x6euy;0x22uy;0x3auy;0x22uy;0x45uy;0x76uy;0x65uy;0x6euy;
//      0x74uy;0x22uy;0x2cuy;0x22uy;0x63uy;0x6fuy;0x6duy;0x6duy;0x65uy;0x6euy;0x74uy;
//      0x22uy;0x3auy;0x22uy;0x52uy;0x65uy;0x70uy;0x72uy;0x69uy;0x63uy;0x65uy;0x20uy;
//      0x61uy;0x6cuy;0x6cuy;0x20uy;0x66uy;0x72uy;0x6fuy;0x6duy;0x20uy;0x4euy;0x6fuy;
//      0x76uy;0x61uy;0x22uy;0x2cuy;0x22uy;0x69uy;0x64uy;0x22uy;0x3auy;0x22uy;0x61uy;
//      0x36uy;0x35uy;0x30uy;0x31uy;0x38uy;0x64uy;0x65uy;0x61uy;0x61uy;0x66uy;0x63uy;
//      0x34uy;0x34uy;0x64uy;0x38uy;0x38uy;0x33uy;0x31uy;0x38uy;0x63uy;0x61uy;0x38uy;
//      0x64uy;0x36uy;0x66uy;0x61uy;0x39uy;0x30uy;0x31uy;0x63uy;0x34uy;0x22uy;0x7duy;
//      0x00uy;0x82uy;0x04uy;0x00uy;0x00uy;0x0cuy;0x46uy;0x31uy;0x34uy;0x62uy;0x30uy;
//      0x30uy;0x39uy;0x35uy;0x65uy;0x66uy;0x32uy;0x38uy;0x37uy;0x34uy;0x65uy;0x31uy;
//      0x30uy;0x61uy;0x65uy;0x62uy;0x63uy;0x61uy;0x39uy;0x61uy;0x31uy;0x62uy;0x33uy;
//      0x61uy;0x39uy;0x30uy;0x64uy;0x64uy;0x64uy;0x3auy;0x32uy;0x31uy;0xaeuy;0x03uy;
//      0x7buy;0x22uy;0x72uy;0x65uy;0x74uy;0x61uy;0x69uy;0x6cuy;0x53uy;0x6buy;0x75uy;
//      0x49uy;0x64uy;0x22uy;0x3auy;0x22uy;0x31uy;0x34uy;0x62uy;0x30uy;0x30uy;0x39uy;
//      0x35uy;0x65uy;0x66uy;0x32uy;0x38uy;0x37uy;0x34uy;0x65uy;0x31uy;0x30uy;0x61uy;
//      0x65uy;0x62uy;0x63uy;0x61uy;0x39uy;0x61uy;0x31uy;0x62uy;0x33uy;0x61uy;0x39uy;
//      0x30uy;0x64uy;0x64uy;0x64uy;0x22uy;0x2cuy;0x22uy;0x6duy;0x61uy;0x72uy;0x74uy;
//      0x49uy;0x64uy;0x22uy;0x3auy;0x32uy;0x31uy;0x2cuy;0x22uy;0x72uy;0x65uy;0x63uy;
//      0x61uy;0x6cuy;0x63uy;0x75uy;0x6cuy;0x61uy;0x74uy;0x69uy;0x6fuy;0x6euy;0x54uy;
//      0x68uy;0x72uy;0x65uy;0x73uy;0x68uy;0x6fuy;0x6cuy;0x64uy;0x22uy;0x3auy;0x30uy;
//      0x2cuy;0x22uy;0x74uy;0x69uy;0x6duy;0x65uy;0x73uy;0x74uy;0x61uy;0x6duy;0x70uy;
//      0x22uy;0x3auy;0x22uy;0x32uy;0x30uy;0x31uy;0x38uy;0x2duy;0x30uy;0x32uy;0x2duy;
//      0x30uy;0x32uy;0x54uy;0x32uy;0x32uy;0x3auy;0x30uy;0x36uy;0x3auy;0x31uy;0x37uy;
//      0x5auy;0x22uy;0x2cuy;0x22uy;0x72uy;0x65uy;0x61uy;0x73uy;0x6fuy;0x6euy;0x22uy;
//      0x3auy;0x22uy;0x45uy;0x76uy;0x65uy;0x6euy;0x74uy;0x22uy;0x2cuy;0x22uy;0x63uy;
//      0x6fuy;0x6duy;0x6duy;0x65uy;0x6euy;0x74uy;0x22uy;0x3auy;0x22uy;0x52uy;0x65uy;
//      0x70uy;0x72uy;0x69uy;0x63uy;0x65uy;0x20uy;0x61uy;0x6cuy;0x6cuy;0x20uy;0x66uy;
//      0x72uy;0x6fuy;0x6duy;0x20uy;0x4euy;0x6fuy;0x76uy;0x61uy;0x22uy;0x2cuy;0x22uy;
//      0x69uy;0x64uy;0x22uy;0x3auy;0x22uy;0x62uy;0x38uy;0x32uy;0x61uy;0x30uy;0x66uy;
//      0x30uy;0x65uy;0x39uy;0x61uy;0x31uy;0x61uy;0x34uy;0x61uy;0x30uy;0x38uy;0x62uy;
//      0x30uy;0x36uy;0x34uy;0x65uy;0x33uy;0x34uy;0x33uy;0x65uy;0x33uy;0x37uy;0x39uy;
//      0x36uy;0x61uy;0x30uy;0x32uy;0x22uy;0x7duy;0x00uy;0x82uy;0x04uy;0x00uy;0x00uy;
//      0x0euy;0x46uy;0x31uy;0x35uy;0x35uy;0x30uy;0x30uy;0x65uy;0x34uy;0x33uy;0x62uy;
//      0x62uy;0x33uy;0x38uy;0x34uy;0x66uy;0x33uy;0x62uy;0x38uy;0x61uy;0x66uy;0x39uy;
//      0x62uy;0x66uy;0x64uy;0x35uy;0x62uy;0x34uy;0x39uy;0x37uy;0x34uy;0x35uy;0x64uy;
//      0x61uy;0x3auy;0x32uy;0x31uy;0xaeuy;0x03uy;0x7buy;0x22uy;0x72uy;0x65uy;0x74uy;
//      0x61uy;0x69uy;0x6cuy;0x53uy;0x6buy;0x75uy;0x49uy;0x64uy;0x22uy;0x3auy;0x22uy;
//      0x31uy;0x35uy;0x35uy;0x30uy;0x30uy;0x65uy;0x34uy;0x33uy;0x62uy;0x62uy;0x33uy;
//      0x38uy;0x34uy;0x66uy;0x33uy;0x62uy;0x38uy;0x61uy;0x66uy;0x39uy;0x62uy;0x66uy;
//      0x64uy;0x35uy;0x62uy;0x34uy;0x39uy;0x37uy;0x34uy;0x35uy;0x64uy;0x61uy;0x22uy;
//      0x2cuy;0x22uy;0x6duy;0x61uy;0x72uy;0x74uy;0x49uy;0x64uy;0x22uy;0x3auy;0x32uy;
//      0x31uy;0x2cuy;0x22uy;0x72uy;0x65uy;0x63uy;0x61uy;0x6cuy;0x63uy;0x75uy;0x6cuy;
//      0x61uy;0x74uy;0x69uy;0x6fuy;0x6euy;0x54uy;0x68uy;0x72uy;0x65uy;0x73uy;0x68uy;
//      0x6fuy;0x6cuy;0x64uy;0x22uy;0x3auy;0x30uy;0x2cuy;0x22uy;0x74uy;0x69uy;0x6duy;
//      0x65uy;0x73uy;0x74uy;0x61uy;0x6duy;0x70uy;0x22uy;0x3auy;0x22uy;0x32uy;0x30uy;
//      0x31uy;0x38uy;0x2duy;0x30uy;0x32uy;0x2duy;0x30uy;0x32uy;0x54uy;0x32uy;0x32uy;
//      0x3auy;0x30uy;0x36uy;0x3auy;0x31uy;0x37uy;0x5auy;0x22uy;0x2cuy;0x22uy;0x72uy;
//      0x65uy;0x61uy;0x73uy;0x6fuy;0x6euy;0x22uy;0x3auy;0x22uy;0x45uy;0x76uy;0x65uy;
//      0x6euy;0x74uy;0x22uy;0x2cuy;0x22uy;0x63uy;0x6fuy;0x6duy;0x6duy;0x65uy;0x6euy;
//      0x74uy;0x22uy;0x3auy;0x22uy;0x52uy;0x65uy;0x70uy;0x72uy;0x69uy;0x63uy;0x65uy;
//      0x20uy;0x61uy;0x6cuy;0x6cuy;0x20uy;0x66uy;0x72uy;0x6fuy;0x6duy;0x20uy;0x4euy;
//      0x6fuy;0x76uy;0x61uy;0x22uy;0x2cuy;0x22uy;0x69uy;0x64uy;0x22uy;0x3auy;0x22uy;
//      0x66uy;0x30uy;0x35uy;0x61uy;0x30uy;0x37uy;0x35uy;0x30uy;0x32uy;0x62uy;0x63uy;
//      0x66uy;0x34uy;0x33uy;0x61uy;0x37uy;0x39uy;0x62uy;0x36uy;0x35uy;0x38uy;0x37uy;
//      0x63uy;0x66uy;0x31uy;0x63uy;0x38uy;0x37uy;0x61uy;0x36uy;0x63uy;0x36uy;0x22uy;
//      0x7duy;0x00uy;0x82uy;0x04uy;0x00uy;0x00uy;0x10uy;0x46uy;0x31uy;0x35uy;0x62uy;
//      0x30uy;0x30uy;0x66uy;0x63uy;0x64uy;0x62uy;0x62uy;0x36uy;0x66uy;0x34uy;0x64uy;
//      0x37uy;0x31uy;0x39uy;0x31uy;0x64uy;0x63uy;0x34uy;0x61uy;0x36uy;0x35uy;0x64uy;
//      0x61uy;0x63uy;0x62uy;0x30uy;0x32uy;0x32uy;0x33uy;0x3auy;0x32uy;0x31uy;0xaeuy;
//      0x03uy;0x7buy;0x22uy;0x72uy;0x65uy;0x74uy;0x61uy;0x69uy;0x6cuy;0x53uy;0x6buy;
//      0x75uy;0x49uy;0x64uy;0x22uy;0x3auy;0x22uy;0x31uy;0x35uy;0x62uy;0x30uy;0x30uy;
//      0x66uy;0x63uy;0x64uy;0x62uy;0x62uy;0x36uy;0x66uy;0x34uy;0x64uy;0x37uy;0x31uy;
//      0x39uy;0x31uy;0x64uy;0x63uy;0x34uy;0x61uy;0x36uy;0x35uy;0x64uy;0x61uy;0x63uy;
//      0x62uy;0x30uy;0x32uy;0x32uy;0x33uy;0x22uy;0x2cuy;0x22uy;0x6duy;0x61uy;0x72uy;
//      0x74uy;0x49uy;0x64uy;0x22uy;0x3auy;0x32uy;0x31uy;0x2cuy;0x22uy;0x72uy;0x65uy;
//      0x63uy;0x61uy;0x6cuy;0x63uy;0x75uy;0x6cuy;0x61uy;0x74uy;0x69uy;0x6fuy;0x6euy;
//      0x54uy;0x68uy;0x72uy;0x65uy;0x73uy;0x68uy;0x6fuy;0x6cuy;0x64uy;0x22uy;0x3auy;
//      0x30uy;0x2cuy;0x22uy;0x74uy;0x69uy;0x6duy;0x65uy;0x73uy;0x74uy;0x61uy;0x6duy;
//      0x70uy;0x22uy;0x3auy;0x22uy;0x32uy;0x30uy;0x31uy;0x38uy;0x2duy;0x30uy;0x32uy;
//      0x2duy;0x30uy;0x32uy;0x54uy;0x32uy;0x32uy;0x3auy;0x30uy;0x36uy;0x3auy;0x31uy;
//      0x37uy;0x5auy;0x22uy;0x2cuy;0x22uy;0x72uy;0x65uy;0x61uy;0x73uy;0x6fuy;0x6euy;
//      0x22uy;0x3auy;0x22uy;0x45uy;0x76uy;0x65uy;0x6euy;0x74uy;0x22uy;0x2cuy;0x22uy;
//      0x63uy;0x6fuy;0x6duy;0x6duy;0x65uy;0x6euy;0x74uy;0x22uy;0x3auy;0x22uy;0x52uy;
//      0x65uy;0x70uy;0x72uy;0x69uy;0x63uy;0x65uy;0x20uy;0x61uy;0x6cuy;0x6cuy;0x20uy;
//      0x66uy;0x72uy;0x6fuy;0x6duy;0x20uy;0x4euy;0x6fuy;0x76uy;0x61uy;0x22uy;0x2cuy;
//      0x22uy;0x69uy;0x64uy;0x22uy;0x3auy;0x22uy;0x36uy;0x39uy;0x64uy;0x35uy;0x33uy;
//      0x35uy;0x30uy;0x31uy;0x66uy;0x35uy;0x32uy;0x64uy;0x34uy;0x32uy;0x37uy;0x34uy;
//      0x39uy;0x63uy;0x35uy;0x35uy;0x33uy;0x32uy;0x30uy;0x66uy;0x34uy;0x35uy;0x63uy;
//      0x33uy;0x64uy;0x34uy;0x34uy;0x35uy;0x22uy;0x7duy;0x00uy;0x82uy;0x04uy;0x00uy;
//      0x00uy;0x12uy;0x46uy;0x31uy;0x36uy;0x39uy;0x30uy;0x30uy;0x36uy;0x33uy;0x65uy;
//      0x64uy;0x34uy;0x66uy;0x62uy;0x34uy;0x32uy;0x30uy;0x65uy;0x61uy;0x65uy;0x39uy;
//      0x31uy;0x36uy;0x63uy;0x33uy;0x64uy;0x31uy;0x62uy;0x39uy;0x38uy;0x31uy;0x30uy;
//      0x33uy;0x32uy;0x3auy;0x32uy;0x31uy;0xaeuy;0x03uy;0x7buy;0x22uy;0x72uy;0x65uy;
//      0x74uy;0x61uy;0x69uy;0x6cuy;0x53uy;0x6buy;0x75uy;0x49uy;0x64uy;0x22uy;0x3auy;
//      0x22uy;0x31uy;0x36uy;0x39uy;0x30uy;0x30uy;0x36uy;0x33uy;0x65uy;0x64uy;0x34uy;
//      0x66uy;0x62uy;0x34uy;0x32uy;0x30uy;0x65uy;0x61uy;0x65uy;0x39uy;0x31uy;0x36uy;
//      0x63uy;0x33uy;0x64uy;0x31uy;0x62uy;0x39uy;0x38uy;0x31uy;0x30uy;0x33uy;0x32uy;
//      0x22uy;0x2cuy;0x22uy;0x6duy;0x61uy;0x72uy;0x74uy;0x49uy;0x64uy;0x22uy;0x3auy;
//      0x32uy;0x31uy;0x2cuy;0x22uy;0x72uy;0x65uy;0x63uy;0x61uy;0x6cuy;0x63uy;0x75uy;
//      0x6cuy;0x61uy;0x74uy;0x69uy;0x6fuy;0x6euy;0x54uy;0x68uy;0x72uy;0x65uy;0x73uy;
//      0x68uy;0x6fuy;0x6cuy;0x64uy;0x22uy;0x3auy;0x30uy;0x2cuy;0x22uy;0x74uy;0x69uy;
//      0x6duy;0x65uy;0x73uy;0x74uy;0x61uy;0x6duy;0x70uy;0x22uy;0x3auy;0x22uy;0x32uy;
//      0x30uy;0x31uy;0x38uy;0x2duy;0x30uy;0x32uy;0x2duy;0x30uy;0x32uy;0x54uy;0x32uy;
//      0x32uy;0x3auy;0x30uy;0x36uy;0x3auy;0x31uy;0x37uy;0x5auy;0x22uy;0x2cuy;0x22uy;
//      0x72uy;0x65uy;0x61uy;0x73uy;0x6fuy;0x6euy;0x22uy;0x3auy;0x22uy;0x45uy;0x76uy;
//      0x65uy;0x6euy;0x74uy;0x22uy;0x2cuy;0x22uy;0x63uy;0x6fuy;0x6duy;0x6duy;0x65uy;
//      0x6euy;0x74uy;0x22uy;0x3auy;0x22uy;0x52uy;0x65uy;0x70uy;0x72uy;0x69uy;0x63uy;
//      0x65uy;0x20uy;0x61uy;0x6cuy;0x6cuy;0x20uy;0x66uy;0x72uy;0x6fuy;0x6duy;0x20uy;
//      0x4euy;0x6fuy;0x76uy;0x61uy;0x22uy;0x2cuy;0x22uy;0x69uy;0x64uy;0x22uy;0x3auy;
//      0x22uy;0x33uy;0x39uy;0x31uy;0x31uy;0x66uy;0x32uy;0x37uy;0x36uy;0x66uy;0x63uy;
//      0x62uy;0x35uy;0x34uy;0x31uy;0x32uy;0x34uy;0x38uy;0x65uy;0x37uy;0x31uy;0x31uy;
//      0x32uy;0x36uy;0x34uy;0x38uy;0x31uy;0x65uy;0x37uy;0x31uy;0x61uy;0x35uy;0x62uy;
//      0x22uy;0x7duy;0x00uy;0x82uy;0x04uy;0x00uy;0x00uy;0x14uy;0x46uy;0x31uy;0x37uy;
//      0x31uy;0x30uy;0x30uy;0x65uy;0x36uy;0x65uy;0x34uy;0x34uy;0x62uy;0x61uy;0x34uy;
//      0x31uy;0x37uy;0x66uy;0x62uy;0x36uy;0x39uy;0x61uy;0x64uy;0x32uy;0x35uy;0x30uy;
//      0x38uy;0x61uy;0x62uy;0x64uy;0x34uy;0x63uy;0x31uy;0x30uy;0x3auy;0x32uy;0x31uy;
//      0xaeuy;0x03uy;0x7buy;0x22uy;0x72uy;0x65uy;0x74uy;0x61uy;0x69uy;0x6cuy;0x53uy;
//      0x6buy;0x75uy;0x49uy;0x64uy;0x22uy;0x3auy;0x22uy;0x31uy;0x37uy;0x31uy;0x30uy;
//      0x30uy;0x65uy;0x36uy;0x65uy;0x34uy;0x34uy;0x62uy;0x61uy;0x34uy;0x31uy;0x37uy;
//      0x66uy;0x62uy;0x36uy;0x39uy;0x61uy;0x64uy;0x32uy;0x35uy;0x30uy;0x38uy;0x61uy;
//      0x62uy;0x64uy;0x34uy;0x63uy;0x31uy;0x30uy;0x22uy;0x2cuy;0x22uy;0x6duy;0x61uy;
//      0x72uy;0x74uy;0x49uy;0x64uy;0x22uy;0x3auy;0x32uy;0x31uy;0x2cuy;0x22uy;0x72uy;
//      0x65uy;0x63uy;0x61uy;0x6cuy;0x63uy;0x75uy;0x6cuy;0x61uy;0x74uy;0x69uy;0x6fuy;
//      0x6euy;0x54uy;0x68uy;0x72uy;0x65uy;0x73uy;0x68uy;0x6fuy;0x6cuy;0x64uy;0x22uy;
//      0x3auy;0x30uy;0x2cuy;0x22uy;0x74uy;0x69uy;0x6duy;0x65uy;0x73uy;0x74uy;0x61uy;
//      0x6duy;0x70uy;0x22uy;0x3auy;0x22uy;0x32uy;0x30uy;0x31uy;0x38uy;0x2duy;0x30uy;
//      0x32uy;0x2duy;0x30uy;0x32uy;0x54uy;0x32uy;0x32uy;0x3auy;0x30uy;0x36uy;0x3auy;
//      0x31uy;0x37uy;0x5auy;0x22uy;0x2cuy;0x22uy;0x72uy;0x65uy;0x61uy;0x73uy;0x6fuy;
//      0x6euy;0x22uy;0x3auy;0x22uy;0x45uy;0x76uy;0x65uy;0x6euy;0x74uy;0x22uy;0x2cuy;
//      0x22uy;0x63uy;0x6fuy;0x6duy;0x6duy;0x65uy;0x6euy;0x74uy;0x22uy;0x3auy;0x22uy;
//      0x52uy;0x65uy;0x70uy;0x72uy;0x69uy;0x63uy;0x65uy;0x20uy;0x61uy;0x6cuy;0x6cuy;
//      0x20uy;0x66uy;0x72uy;0x6fuy;0x6duy;0x20uy;0x4euy;0x6fuy;0x76uy;0x61uy;0x22uy;
//      0x2cuy;0x22uy;0x69uy;0x64uy;0x22uy;0x3auy;0x22uy;0x35uy;0x35uy;0x34uy;0x62uy;
//      0x65uy;0x37uy;0x32uy;0x37uy;0x30uy;0x30uy;0x37uy;0x34uy;0x34uy;0x32uy;0x61uy;
//      0x33uy;0x38uy;0x36uy;0x61uy;0x39uy;0x36uy;0x30uy;0x62uy;0x30uy;0x36uy;0x66uy;
//      0x62uy;0x35uy;0x61uy;0x34uy;0x66uy;0x38uy;0x22uy;0x7duy;0x00uy;0x82uy;0x04uy;
//      0x00uy;0x00uy;0x16uy;0x46uy;0x31uy;0x37uy;0x63uy;0x30uy;0x31uy;0x30uy;0x61uy;
//      0x62uy;0x33uy;0x65uy;0x39uy;0x62uy;0x34uy;0x36uy;0x35uy;0x33uy;0x39uy;0x33uy;
//      0x62uy;0x63uy;0x37uy;0x32uy;0x37uy;0x64uy;0x31uy;0x30uy;0x33uy;0x63uy;0x62uy;
//      0x62uy;0x38uy;0x62uy;0x3auy;0x32uy;0x31uy;0xaeuy;0x03uy;0x7buy;0x22uy;0x72uy;
//      0x65uy;0x74uy;0x61uy;0x69uy;0x6cuy;0x53uy;0x6buy;0x75uy;0x49uy;0x64uy;0x22uy;
//      0x3auy;0x22uy;0x31uy;0x37uy;0x63uy;0x30uy;0x31uy;0x30uy;0x61uy;0x62uy;0x33uy;
//      0x65uy;0x39uy;0x62uy;0x34uy;0x36uy;0x35uy;0x33uy;0x39uy;0x33uy;0x62uy;0x63uy;
//      0x37uy;0x32uy;0x37uy;0x64uy;0x31uy;0x30uy;0x33uy;0x63uy;0x62uy;0x62uy;0x38uy;
//      0x62uy;0x22uy;0x2cuy;0x22uy;0x6duy;0x61uy;0x72uy;0x74uy;0x49uy;0x64uy;0x22uy;
//      0x3auy;0x32uy;0x31uy;0x2cuy;0x22uy;0x72uy;0x65uy;0x63uy;0x61uy;0x6cuy;0x63uy;
//      0x75uy;0x6cuy;0x61uy;0x74uy;0x69uy;0x6fuy;0x6euy;0x54uy;0x68uy;0x72uy;0x65uy;
//      0x73uy;0x68uy;0x6fuy;0x6cuy;0x64uy;0x22uy;0x3auy;0x30uy;0x2cuy;0x22uy;0x74uy;
//      0x69uy;0x6duy;0x65uy;0x73uy;0x74uy;0x61uy;0x6duy;0x70uy;0x22uy;0x3auy;0x22uy;
//      0x32uy;0x30uy;0x31uy;0x38uy;0x2duy;0x30uy;0x32uy;0x2duy;0x30uy;0x32uy;0x54uy;
//      0x32uy;0x32uy;0x3auy;0x30uy;0x36uy;0x3auy;0x31uy;0x37uy;0x5auy;0x22uy;0x2cuy;
//      0x22uy;0x72uy;0x65uy;0x61uy;0x73uy;0x6fuy;0x6euy;0x22uy;0x3auy;0x22uy;0x45uy;
//      0x76uy;0x65uy;0x6euy;0x74uy;0x22uy;0x2cuy;0x22uy;0x63uy;0x6fuy;0x6duy;0x6duy;
//      0x65uy;0x6euy;0x74uy;0x22uy;0x3auy;0x22uy;0x52uy;0x65uy;0x70uy;0x72uy;0x69uy;
//      0x63uy;0x65uy;0x20uy;0x61uy;0x6cuy;0x6cuy;0x20uy;0x66uy;0x72uy;0x6fuy;0x6duy;
//      0x20uy;0x4euy;0x6fuy;0x76uy;0x61uy;0x22uy;0x2cuy;0x22uy;0x69uy;0x64uy;0x22uy;
//      0x3auy;0x22uy;0x64uy;0x36uy;0x38uy;0x31uy;0x61uy;0x34uy;0x62uy;0x33uy;0x30uy;
//      0x36uy;0x36uy;0x34uy;0x34uy;0x31uy;0x64uy;0x38uy;0x39uy;0x65uy;0x30uy;0x33uy;
//      0x63uy;0x39uy;0x32uy;0x39uy;0x39uy;0x62uy;0x62uy;0x65uy;0x65uy;0x33uy;0x66uy;
//      0x63uy;0x22uy;0x7duy;0x00uy;0x82uy;0x04uy;0x00uy;0x00uy;0x18uy;0x46uy;0x31uy;
//      0x63uy;0x36uy;0x30uy;0x31uy;0x34uy;0x61uy;0x35uy;0x63uy;0x36uy;0x36uy;0x62uy;
//      0x34uy;0x37uy;0x38uy;0x38uy;0x61uy;0x37uy;0x36uy;0x65uy;0x30uy;0x62uy;0x66uy;
//      0x30uy;0x65uy;0x39uy;0x63uy;0x39uy;0x35uy;0x64uy;0x36uy;0x37uy;0x3auy;0x32uy;
//      0x31uy;0xaeuy;0x03uy;0x7buy;0x22uy;0x72uy;0x65uy;0x74uy;0x61uy;0x69uy;0x6cuy;
//      0x53uy;0x6buy;0x75uy;0x49uy;0x64uy;0x22uy;0x3auy;0x22uy;0x31uy;0x63uy;0x36uy;
//      0x30uy;0x31uy;0x34uy;0x61uy;0x35uy;0x63uy;0x36uy;0x36uy;0x62uy;0x34uy;0x37uy;
//      0x38uy;0x38uy;0x61uy;0x37uy;0x36uy;0x65uy;0x30uy;0x62uy;0x66uy;0x30uy;0x65uy;
//      0x39uy;0x63uy;0x39uy;0x35uy;0x64uy;0x36uy;0x37uy;0x22uy;0x2cuy;0x22uy;0x6duy;
//      0x61uy;0x72uy;0x74uy;0x49uy;0x64uy;0x22uy;0x3auy;0x32uy;0x31uy;0x2cuy;0x22uy;
//      0x72uy;0x65uy;0x63uy;0x61uy;0x6cuy;0x63uy;0x75uy;0x6cuy;0x61uy;0x74uy;0x69uy;
//      0x6fuy;0x6euy;0x54uy;0x68uy;0x72uy;0x65uy;0x73uy;0x68uy;0x6fuy;0x6cuy;0x64uy;
//      0x22uy;0x3auy;0x30uy;0x2cuy;0x22uy;0x74uy;0x69uy;0x6duy;0x65uy;0x73uy;0x74uy;
//      0x61uy;0x6duy;0x70uy;0x22uy;0x3auy;0x22uy;0x32uy;0x30uy;0x31uy;0x38uy;0x2duy;
//      0x30uy;0x32uy;0x2duy;0x30uy;0x32uy;0x54uy;0x32uy;0x32uy;0x3auy;0x30uy;0x36uy;
//      0x3auy;0x31uy;0x37uy;0x5auy;0x22uy;0x2cuy;0x22uy;0x72uy;0x65uy;0x61uy;0x73uy;
//      0x6fuy;0x6euy;0x22uy;0x3auy;0x22uy;0x45uy;0x76uy;0x65uy;0x6euy;0x74uy;0x22uy;
//      0x2cuy;0x22uy;0x63uy;0x6fuy;0x6duy;0x6duy;0x65uy;0x6euy;0x74uy;0x22uy;0x3auy;
//      0x22uy;0x52uy;0x65uy;0x70uy;0x72uy;0x69uy;0x63uy;0x65uy;0x20uy;0x61uy;0x6cuy;
//      0x6cuy;0x20uy;0x66uy;0x72uy;0x6fuy;0x6duy;0x20uy;0x4euy;0x6fuy;0x76uy;0x61uy;
//      0x22uy;0x2cuy;0x22uy;0x69uy;0x64uy;0x22uy;0x3auy;0x22uy;0x65uy;0x39uy;0x34uy;
//      0x38uy;0x35uy;0x39uy;0x65uy;0x66uy;0x35uy;0x30uy;0x38uy;0x65uy;0x34uy;0x61uy;
//      0x38uy;0x62uy;0x38uy;0x37uy;0x38uy;0x64uy;0x30uy;0x39uy;0x39uy;0x37uy;0x64uy;
//      0x35uy;0x39uy;0x63uy;0x33uy;0x65uy;0x61uy;0x36uy;0x22uy;0x7duy;0x00uy;0x82uy;
//      0x04uy;0x00uy;0x00uy;0x1auy;0x46uy;0x31uy;0x38uy;0x39uy;0x30uy;0x30uy;0x61uy;
//      0x64uy;0x62uy;0x30uy;0x35uy;0x63uy;0x61uy;0x34uy;0x30uy;0x33uy;0x65uy;0x61uy;
//      0x39uy;0x61uy;0x36uy;0x32uy;0x32uy;0x33uy;0x38uy;0x37uy;0x61uy;0x64uy;0x39uy;
//      0x64uy;0x62uy;0x65uy;0x63uy;0x3auy;0x32uy;0x31uy;0xaeuy;0x03uy;0x7buy;0x22uy;
//      0x72uy;0x65uy;0x74uy;0x61uy;0x69uy;0x6cuy;0x53uy;0x6buy;0x75uy;0x49uy;0x64uy;
//      0x22uy;0x3auy;0x22uy;0x31uy;0x38uy;0x39uy;0x30uy;0x30uy;0x61uy;0x64uy;0x62uy;
//      0x30uy;0x35uy;0x63uy;0x61uy;0x34uy;0x30uy;0x33uy;0x65uy;0x61uy;0x39uy;0x61uy;
//      0x36uy;0x32uy;0x32uy;0x33uy;0x38uy;0x37uy;0x61uy;0x64uy;0x39uy;0x64uy;0x62uy;
//      0x65uy;0x63uy;0x22uy;0x2cuy;0x22uy;0x6duy;0x61uy;0x72uy;0x74uy;0x49uy;0x64uy;
//      0x22uy;0x3auy;0x32uy;0x31uy;0x2cuy;0x22uy;0x72uy;0x65uy;0x63uy;0x61uy;0x6cuy;
//      0x63uy;0x75uy;0x6cuy;0x61uy;0x74uy;0x69uy;0x6fuy;0x6euy;0x54uy;0x68uy;0x72uy;
//      0x65uy;0x73uy;0x68uy;0x6fuy;0x6cuy;0x64uy;0x22uy;0x3auy;0x30uy;0x2cuy;0x22uy;
//      0x74uy;0x69uy;0x6duy;0x65uy;0x73uy;0x74uy;0x61uy;0x6duy;0x70uy;0x22uy;0x3auy;
//      0x22uy;0x32uy;0x30uy;0x31uy;0x38uy;0x2duy;0x30uy;0x32uy;0x2duy;0x30uy;0x32uy;
//      0x54uy;0x32uy;0x32uy;0x3auy;0x30uy;0x36uy;0x3auy;0x31uy;0x37uy;0x5auy;0x22uy;
//      0x2cuy;0x22uy;0x72uy;0x65uy;0x61uy;0x73uy;0x6fuy;0x6euy;0x22uy;0x3auy;0x22uy;
//      0x45uy;0x76uy;0x65uy;0x6euy;0x74uy;0x22uy;0x2cuy;0x22uy;0x63uy;0x6fuy;0x6duy;
//      0x6duy;0x65uy;0x6euy;0x74uy;0x22uy;0x3auy;0x22uy;0x52uy;0x65uy;0x70uy;0x72uy;
//      0x69uy;0x63uy;0x65uy;0x20uy;0x61uy;0x6cuy;0x6cuy;0x20uy;0x66uy;0x72uy;0x6fuy;
//      0x6duy;0x20uy;0x4euy;0x6fuy;0x76uy;0x61uy;0x22uy;0x2cuy;0x22uy;0x69uy;0x64uy;
//      0x22uy;0x3auy;0x22uy;0x65uy;0x66uy;0x63uy;0x35uy;0x33uy;0x65uy;0x33uy;0x30uy;
//      0x30uy;0x38uy;0x63uy;0x35uy;0x34uy;0x63uy;0x66uy;0x36uy;0x61uy;0x38uy;0x62uy;
//      0x36uy;0x64uy;0x64uy;0x66uy;0x33uy;0x62uy;0x61uy;0x38uy;0x36uy;0x63uy;0x64uy;
//      0x64uy;0x62uy;0x22uy;0x7duy;0x00uy;0x82uy;0x04uy;0x00uy;0x00uy;0x1cuy;0x46uy;
//      0x31uy;0x39uy;0x36uy;0x30uy;0x30uy;0x61uy;0x63uy;0x64uy;0x65uy;0x66uy;0x38uy;
//      0x36uy;0x34uy;0x66uy;0x30uy;0x32uy;0x62uy;0x38uy;0x35uy;0x33uy;0x64uy;0x63uy;
//      0x65uy;0x37uy;0x30uy;0x62uy;0x31uy;0x32uy;0x39uy;0x64uy;0x37uy;0x36uy;0x3auy;
//      0x32uy;0x31uy;0xaeuy;0x03uy;0x7buy;0x22uy;0x72uy;0x65uy;0x74uy;0x61uy;0x69uy;
//      0x6cuy;0x53uy;0x6buy;0x75uy;0x49uy;0x64uy;0x22uy;0x3auy;0x22uy;0x31uy;0x39uy;
//      0x36uy;0x30uy;0x30uy;0x61uy;0x63uy;0x64uy;0x65uy;0x66uy;0x38uy;0x36uy;0x34uy;
//      0x66uy;0x30uy;0x32uy;0x62uy;0x38uy;0x35uy;0x33uy;0x64uy;0x63uy;0x65uy;0x37uy;
//      0x30uy;0x62uy;0x31uy;0x32uy;0x39uy;0x64uy;0x37uy;0x36uy;0x22uy;0x2cuy;0x22uy;
//      0x6duy;0x61uy;0x72uy;0x74uy;0x49uy;0x64uy;0x22uy;0x3auy;0x32uy;0x31uy;0x2cuy;
//      0x22uy;0x72uy;0x65uy;0x63uy;0x61uy;0x6cuy;0x63uy;0x75uy;0x6cuy;0x61uy;0x74uy;
//      0x69uy;0x6fuy;0x6euy;0x54uy;0x68uy;0x72uy;0x65uy;0x73uy;0x68uy;0x6fuy;0x6cuy;
//      0x64uy;0x22uy;0x3auy;0x30uy;0x2cuy;0x22uy;0x74uy;0x69uy;0x6duy;0x65uy;0x73uy;
//      0x74uy;0x61uy;0x6duy;0x70uy;0x22uy;0x3auy;0x22uy;0x32uy;0x30uy;0x31uy;0x38uy;
//      0x2duy;0x30uy;0x32uy;0x2duy;0x30uy;0x32uy;0x54uy;0x32uy;0x32uy;0x3auy;0x30uy;
//      0x36uy;0x3auy;0x31uy;0x37uy;0x5auy;0x22uy;0x2cuy;0x22uy;0x72uy;0x65uy;0x61uy;
//      0x73uy;0x6fuy;0x6euy;0x22uy;0x3auy;0x22uy;0x45uy;0x76uy;0x65uy;0x6euy;0x74uy;
//      0x22uy;0x2cuy;0x22uy;0x63uy;0x6fuy;0x6duy;0x6duy;0x65uy;0x6euy;0x74uy;0x22uy;
//      0x3auy;0x22uy;0x52uy;0x65uy;0x70uy;0x72uy;0x69uy;0x63uy;0x65uy;0x20uy;0x61uy;
//      0x6cuy;0x6cuy;0x20uy;0x66uy;0x72uy;0x6fuy;0x6duy;0x20uy;0x4euy;0x6fuy;0x76uy;
//      0x61uy;0x22uy;0x2cuy;0x22uy;0x69uy;0x64uy;0x22uy;0x3auy;0x22uy;0x65uy;0x37uy;
//      0x61uy;0x63uy;0x61uy;0x33uy;0x30uy;0x36uy;0x35uy;0x62uy;0x38uy;0x64uy;0x34uy;
//      0x30uy;0x34uy;0x61uy;0x39uy;0x63uy;0x33uy;0x39uy;0x61uy;0x38uy;0x37uy;0x33uy;
//      0x35uy;0x30uy;0x30uy;0x34uy;0x32uy;0x32uy;0x35uy;0x32uy;0x22uy;0x7duy;0x00uy;
//      0x82uy;0x04uy;0x00uy;0x00uy;0x1euy;0x46uy;0x31uy;0x62uy;0x30uy;0x30uy;0x30uy;
//      0x39uy;0x61uy;0x34uy;0x61uy;0x34uy;0x36uy;0x34uy;0x34uy;0x33uy;0x36uy;0x39uy;
//      0x38uy;0x36uy;0x39uy;0x36uy;0x64uy;0x61uy;0x30uy;0x61uy;0x32uy;0x63uy;0x38uy;
//      0x37uy;0x36uy;0x30uy;0x37uy;0x39uy;0x3auy;0x32uy;0x31uy;0xaeuy;0x03uy;0x7buy;
//      0x22uy;0x72uy;0x65uy;0x74uy;0x61uy;0x69uy;0x6cuy;0x53uy;0x6buy;0x75uy;0x49uy;
//      0x64uy;0x22uy;0x3auy;0x22uy;0x31uy;0x62uy;0x30uy;0x30uy;0x30uy;0x39uy;0x61uy;
//      0x34uy;0x61uy;0x34uy;0x36uy;0x34uy;0x34uy;0x33uy;0x36uy;0x39uy;0x38uy;0x36uy;
//      0x39uy;0x36uy;0x64uy;0x61uy;0x30uy;0x61uy;0x32uy;0x63uy;0x38uy;0x37uy;0x36uy;
//      0x30uy;0x37uy;0x39uy;0x22uy;0x2cuy;0x22uy;0x6duy;0x61uy;0x72uy;0x74uy;0x49uy;
//      0x64uy;0x22uy;0x3auy;0x32uy;0x31uy;0x2cuy;0x22uy;0x72uy;0x65uy;0x63uy;0x61uy;
//      0x6cuy;0x63uy;0x75uy;0x6cuy;0x61uy;0x74uy;0x69uy;0x6fuy;0x6euy;0x54uy;0x68uy;
//      0x72uy;0x65uy;0x73uy;0x68uy;0x6fuy;0x6cuy;0x64uy;0x22uy;0x3auy;0x30uy;0x2cuy;
//      0x22uy;0x74uy;0x69uy;0x6duy;0x65uy;0x73uy;0x74uy;0x61uy;0x6duy;0x70uy;0x22uy;
//      0x3auy;0x22uy;0x32uy;0x30uy;0x31uy;0x38uy;0x2duy;0x30uy;0x32uy;0x2duy;0x30uy;
//      0x32uy;0x54uy;0x32uy;0x32uy;0x3auy;0x30uy;0x36uy;0x3auy;0x31uy;0x37uy;0x5auy;
//      0x22uy;0x2cuy;0x22uy;0x72uy;0x65uy;0x61uy;0x73uy;0x6fuy;0x6euy;0x22uy;0x3auy;
//      0x22uy;0x45uy;0x76uy;0x65uy;0x6euy;0x74uy;0x22uy;0x2cuy;0x22uy;0x63uy;0x6fuy;
//      0x6duy;0x6duy;0x65uy;0x6euy;0x74uy;0x22uy;0x3auy;0x22uy;0x52uy;0x65uy;0x70uy;
//      0x72uy;0x69uy;0x63uy;0x65uy;0x20uy;0x61uy;0x6cuy;0x6cuy;0x20uy;0x66uy;0x72uy;
//      0x6fuy;0x6duy;0x20uy;0x4euy;0x6fuy;0x76uy;0x61uy;0x22uy;0x2cuy;0x22uy;0x69uy;
//      0x64uy;0x22uy;0x3auy;0x22uy;0x32uy;0x32uy;0x38uy;0x38uy;0x66uy;0x64uy;0x39uy;
//      0x37uy;0x66uy;0x64uy;0x64uy;0x36uy;0x34uy;0x38uy;0x61uy;0x35uy;0x61uy;0x31uy;
//      0x32uy;0x35uy;0x66uy;0x31uy;0x33uy;0x37uy;0x38uy;0x61uy;0x61uy;0x30uy;0x65uy;
//      0x64uy;0x65uy;0x66uy;0x22uy;0x7duy;0x00uy;0x82uy;0x04uy;0x00uy;0x00uy;0x20uy;
//      0x46uy;0x31uy;0x30uy;0x31uy;0x30uy;0x30uy;0x66uy;0x31uy;0x35uy;0x32uy;0x66uy;
//      0x31uy;0x36uy;0x34uy;0x33uy;0x64uy;0x66uy;0x62uy;0x39uy;0x30uy;0x38uy;0x30uy;
//      0x63uy;0x33uy;0x34uy;0x64uy;0x39uy;0x32uy;0x30uy;0x62uy;0x65uy;0x65uy;0x38uy;
//      0x3auy;0x32uy;0x31uy;0xaeuy;0x03uy;0x7buy;0x22uy;0x72uy;0x65uy;0x74uy;0x61uy;
//      0x69uy;0x6cuy;0x53uy;0x6buy;0x75uy;0x49uy;0x64uy;0x22uy;0x3auy;0x22uy;0x31uy;
//      0x30uy;0x31uy;0x30uy;0x30uy;0x66uy;0x31uy;0x35uy;0x32uy;0x66uy;0x31uy;0x36uy;
//      0x34uy;0x33uy;0x64uy;0x66uy;0x62uy;0x39uy;0x30uy;0x38uy;0x30uy;0x63uy;0x33uy;
//      0x34uy;0x64uy;0x39uy;0x32uy;0x30uy;0x62uy;0x65uy;0x65uy;0x38uy;0x22uy;0x2cuy;
//      0x22uy;0x6duy;0x61uy;0x72uy;0x74uy;0x49uy;0x64uy;0x22uy;0x3auy;0x32uy;0x31uy;
//      0x2cuy;0x22uy;0x72uy;0x65uy;0x63uy;0x61uy;0x6cuy;0x63uy;0x75uy;0x6cuy;0x61uy;
//      0x74uy;0x69uy;0x6fuy;0x6euy;0x54uy;0x68uy;0x72uy;0x65uy;0x73uy;0x68uy;0x6fuy;
//      0x6cuy;0x64uy;0x22uy;0x3auy;0x30uy;0x2cuy;0x22uy;0x74uy;0x69uy;0x6duy;0x65uy;
//      0x73uy;0x74uy;0x61uy;0x6duy;0x70uy;0x22uy;0x3auy;0x22uy;0x32uy;0x30uy;0x31uy;
//      0x38uy;0x2duy;0x30uy;0x32uy;0x2duy;0x30uy;0x32uy;0x54uy;0x32uy;0x32uy;0x3auy;
//      0x30uy;0x36uy;0x3auy;0x31uy;0x37uy;0x5auy;0x22uy;0x2cuy;0x22uy;0x72uy;0x65uy;
//      0x61uy;0x73uy;0x6fuy;0x6euy;0x22uy;0x3auy;0x22uy;0x45uy;0x76uy;0x65uy;0x6euy;
//      0x74uy;0x22uy;0x2cuy;0x22uy;0x63uy;0x6fuy;0x6duy;0x6duy;0x65uy;0x6euy;0x74uy;
//      0x22uy;0x3auy;0x22uy;0x52uy;0x65uy;0x70uy;0x72uy;0x69uy;0x63uy;0x65uy;0x20uy;
//      0x61uy;0x6cuy;0x6cuy;0x20uy;0x66uy;0x72uy;0x6fuy;0x6duy;0x20uy;0x4euy;0x6fuy;
//      0x76uy;0x61uy;0x22uy;0x2cuy;0x22uy;0x69uy;0x64uy;0x22uy;0x3auy;0x22uy;0x61uy;
//      0x38uy;0x61uy;0x62uy;0x61uy;0x30uy;0x63uy;0x39uy;0x65uy;0x66uy;0x34uy;0x32uy;
//      0x34uy;0x37uy;0x32uy;0x37uy;0x61uy;0x32uy;0x35uy;0x33uy;0x63uy;0x65uy;0x66uy;
//      0x63uy;0x36uy;0x65uy;0x33uy;0x35uy;0x32uy;0x65uy;0x61uy;0x61uy;0x22uy;0x7duy;
//      0x00uy;0x82uy;0x04uy;0x00uy;0x00uy;0x22uy;0x46uy;0x31uy;0x31uy;0x31uy;0x30uy;
//      0x30uy;0x33uy;0x65uy;0x37uy;0x34uy;0x38uy;0x62uy;0x31uy;0x34uy;0x63uy;0x64uy;
//      0x31uy;0x61uy;0x38uy;0x39uy;0x39uy;0x37uy;0x65uy;0x37uy;0x34uy;0x37uy;0x37uy;
//      0x37uy;0x64uy;0x39uy;0x65uy;0x61uy;0x37uy;0x3auy;0x32uy;0x31uy;0xaeuy;0x03uy;
//      0x7buy;0x22uy;0x72uy;0x65uy;0x74uy;0x61uy;0x69uy;0x6cuy;0x53uy;0x6buy;0x75uy;
//      0x49uy;0x64uy;0x22uy;0x3auy;0x22uy;0x31uy;0x31uy;0x31uy;0x30uy;0x30uy;0x33uy;
//      0x65uy;0x37uy;0x34uy;0x38uy;0x62uy;0x31uy;0x34uy;0x63uy;0x64uy;0x31uy;0x61uy;
//      0x38uy;0x39uy;0x39uy;0x37uy;0x65uy;0x37uy;0x34uy;0x37uy;0x37uy;0x37uy;0x64uy;
//      0x39uy;0x65uy;0x61uy;0x37uy;0x22uy;0x2cuy;0x22uy;0x6duy;0x61uy;0x72uy;0x74uy;
//      0x49uy;0x64uy;0x22uy;0x3auy;0x32uy;0x31uy;0x2cuy;0x22uy;0x72uy;0x65uy;0x63uy;
//      0x61uy;0x6cuy;0x63uy;0x75uy;0x6cuy;0x61uy;0x74uy;0x69uy;0x6fuy;0x6euy;0x54uy;
//      0x68uy;0x72uy;0x65uy;0x73uy;0x68uy;0x6fuy;0x6cuy;0x64uy;0x22uy;0x3auy;0x30uy;
//      0x2cuy;0x22uy;0x74uy;0x69uy;0x6duy;0x65uy;0x73uy;0x74uy;0x61uy;0x6duy;0x70uy;
//      0x22uy;0x3auy;0x22uy;0x32uy;0x30uy;0x31uy;0x38uy;0x2duy;0x30uy;0x32uy;0x2duy;
//      0x30uy;0x32uy;0x54uy;0x32uy;0x32uy;0x3auy;0x30uy;0x36uy;0x3auy;0x31uy;0x37uy;
//      0x5auy;0x22uy;0x2cuy;0x22uy;0x72uy;0x65uy;0x61uy;0x73uy;0x6fuy;0x6euy;0x22uy;
//      0x3auy;0x22uy;0x45uy;0x76uy;0x65uy;0x6euy;0x74uy;0x22uy;0x2cuy;0x22uy;0x63uy;
//      0x6fuy;0x6duy;0x6duy;0x65uy;0x6euy;0x74uy;0x22uy;0x3auy;0x22uy;0x52uy;0x65uy;
//      0x70uy;0x72uy;0x69uy;0x63uy;0x65uy;0x20uy;0x61uy;0x6cuy;0x6cuy;0x20uy;0x66uy;
//      0x72uy;0x6fuy;0x6duy;0x20uy;0x4euy;0x6fuy;0x76uy;0x61uy;0x22uy;0x2cuy;0x22uy;
//      0x69uy;0x64uy;0x22uy;0x3auy;0x22uy;0x35uy;0x61uy;0x30uy;0x65uy;0x35uy;0x63uy;
//      0x62uy;0x64uy;0x65uy;0x65uy;0x32uy;0x61uy;0x34uy;0x36uy;0x63uy;0x34uy;0x38uy;
//      0x62uy;0x61uy;0x62uy;0x61uy;0x61uy;0x61uy;0x62uy;0x65uy;0x65uy;0x63uy;0x31uy;
//      0x32uy;0x38uy;0x30uy;0x65uy;0x22uy;0x7duy;0x00uy;0x82uy;0x04uy;0x00uy;0x00uy;
//      0x24uy;0x46uy;0x31uy;0x66uy;0x34uy;0x30uy;0x30uy;0x34uy;0x65uy;0x36uy;0x61uy;
//      0x63uy;0x64uy;0x30uy;0x34uy;0x39uy;0x65uy;0x31uy;0x39uy;0x64uy;0x32uy;0x33uy;
//      0x62uy;0x66uy;0x36uy;0x63uy;0x31uy;0x62uy;0x64uy;0x65uy;0x35uy;0x30uy;0x39uy;
//      0x63uy;0x3auy;0x32uy;0x31uy;0xaeuy;0x03uy;0x7buy;0x22uy;0x72uy;0x65uy;0x74uy;
//      0x61uy;0x69uy;0x6cuy;0x53uy;0x6buy;0x75uy;0x49uy;0x64uy;0x22uy;0x3auy;0x22uy;
//      0x31uy;0x66uy;0x34uy;0x30uy;0x30uy;0x34uy;0x65uy;0x36uy;0x61uy;0x63uy;0x64uy;
//      0x30uy;0x34uy;0x39uy;0x65uy;0x31uy;0x39uy;0x64uy;0x32uy;0x33uy;0x62uy;0x66uy;
//      0x36uy;0x63uy;0x31uy;0x62uy;0x64uy;0x65uy;0x35uy;0x30uy;0x39uy;0x63uy;0x22uy;
//      0x2cuy;0x22uy;0x6duy;0x61uy;0x72uy;0x74uy;0x49uy;0x64uy;0x22uy;0x3auy;0x32uy;
//      0x31uy;0x2cuy;0x22uy;0x72uy;0x65uy;0x63uy;0x61uy;0x6cuy;0x63uy;0x75uy;0x6cuy;
//      0x61uy;0x74uy;0x69uy;0x6fuy;0x6euy;0x54uy;0x68uy;0x72uy;0x65uy;0x73uy;0x68uy;
//      0x6fuy;0x6cuy;0x64uy;0x22uy;0x3auy;0x30uy;0x2cuy;0x22uy;0x74uy;0x69uy;0x6duy;
//      0x65uy;0x73uy;0x74uy;0x61uy;0x6duy;0x70uy;0x22uy;0x3auy;0x22uy;0x32uy;0x30uy;
//      0x31uy;0x38uy;0x2duy;0x30uy;0x32uy;0x2duy;0x30uy;0x32uy;0x54uy;0x32uy;0x32uy;
//      0x3auy;0x30uy;0x36uy;0x3auy;0x31uy;0x37uy;0x5auy;0x22uy;0x2cuy;0x22uy;0x72uy;
//      0x65uy;0x61uy;0x73uy;0x6fuy;0x6euy;0x22uy;0x3auy;0x22uy;0x45uy;0x76uy;0x65uy;
//      0x6euy;0x74uy;0x22uy;0x2cuy;0x22uy;0x63uy;0x6fuy;0x6duy;0x6duy;0x65uy;0x6euy;
//      0x74uy;0x22uy;0x3auy;0x22uy;0x52uy;0x65uy;0x70uy;0x72uy;0x69uy;0x63uy;0x65uy;
//      0x20uy;0x61uy;0x6cuy;0x6cuy;0x20uy;0x66uy;0x72uy;0x6fuy;0x6duy;0x20uy;0x4euy;
//      0x6fuy;0x76uy;0x61uy;0x22uy;0x2cuy;0x22uy;0x69uy;0x64uy;0x22uy;0x3auy;0x22uy;
//      0x31uy;0x37uy;0x66uy;0x35uy;0x36uy;0x63uy;0x61uy;0x65uy;0x66uy;0x38uy;0x66uy;
//      0x39uy;0x34uy;0x39uy;0x66uy;0x30uy;0x61uy;0x34uy;0x61uy;0x66uy;0x38uy;0x65uy;
//      0x33uy;0x30uy;0x61uy;0x37uy;0x63uy;0x35uy;0x33uy;0x31uy;0x37uy;0x35uy;0x22uy;
//      0x7duy;0x00uy;0x82uy;0x04uy;0x00uy;0x00uy;0x26uy;0x46uy;0x31uy;0x31uy;0x62uy;
//      0x30uy;0x30uy;0x34uy;0x34uy;0x34uy;0x32uy;0x33uy;0x66uy;0x37uy;0x34uy;0x31uy;
//      0x61uy;0x32uy;0x62uy;0x30uy;0x61uy;0x66uy;0x33uy;0x35uy;0x65uy;0x30uy;0x37uy;
//      0x33uy;0x37uy;0x33uy;0x38uy;0x38uy;0x66uy;0x31uy;0x3auy;0x32uy;0x31uy;0xaeuy;
//      0x03uy;0x7buy;0x22uy;0x72uy;0x65uy;0x74uy;0x61uy;0x69uy;0x6cuy;0x53uy;0x6buy;
//      0x75uy;0x49uy;0x64uy;0x22uy;0x3auy;0x22uy;0x31uy;0x31uy;0x62uy;0x30uy;0x30uy;
//      0x34uy;0x34uy;0x34uy;0x32uy;0x33uy;0x66uy;0x37uy;0x34uy;0x31uy;0x61uy;0x32uy;
//      0x62uy;0x30uy;0x61uy;0x66uy;0x33uy;0x35uy;0x65uy;0x30uy;0x37uy;0x33uy;0x37uy;
//      0x33uy;0x38uy;0x38uy;0x66uy;0x31uy;0x22uy;0x2cuy;0x22uy;0x6duy;0x61uy;0x72uy;
//      0x74uy;0x49uy;0x64uy;0x22uy;0x3auy;0x32uy;0x31uy;0x2cuy;0x22uy;0x72uy;0x65uy;
//      0x63uy;0x61uy;0x6cuy;0x63uy;0x75uy;0x6cuy;0x61uy;0x74uy;0x69uy;0x6fuy;0x6euy;
//      0x54uy;0x68uy;0x72uy;0x65uy;0x73uy;0x68uy;0x6fuy;0x6cuy;0x64uy;0x22uy;0x3auy;
//      0x30uy;0x2cuy;0x22uy;0x74uy;0x69uy;0x6duy;0x65uy;0x73uy;0x74uy;0x61uy;0x6duy;
//      0x70uy;0x22uy;0x3auy;0x22uy;0x32uy;0x30uy;0x31uy;0x38uy;0x2duy;0x30uy;0x32uy;
//      0x2duy;0x30uy;0x32uy;0x54uy;0x32uy;0x32uy;0x3auy;0x30uy;0x36uy;0x3auy;0x31uy;
//      0x37uy;0x5auy;0x22uy;0x2cuy;0x22uy;0x72uy;0x65uy;0x61uy;0x73uy;0x6fuy;0x6euy;
//      0x22uy;0x3auy;0x22uy;0x45uy;0x76uy;0x65uy;0x6euy;0x74uy;0x22uy;0x2cuy;0x22uy;
//      0x63uy;0x6fuy;0x6duy;0x6duy;0x65uy;0x6euy;0x74uy;0x22uy;0x3auy;0x22uy;0x52uy;
//      0x65uy;0x70uy;0x72uy;0x69uy;0x63uy;0x65uy;0x20uy;0x61uy;0x6cuy;0x6cuy;0x20uy;
//      0x66uy;0x72uy;0x6fuy;0x6duy;0x20uy;0x4euy;0x6fuy;0x76uy;0x61uy;0x22uy;0x2cuy;
//      0x22uy;0x69uy;0x64uy;0x22uy;0x3auy;0x22uy;0x64uy;0x37uy;0x36uy;0x61uy;0x31uy;
//      0x37uy;0x30uy;0x33uy;0x63uy;0x30uy;0x38uy;0x37uy;0x34uy;0x39uy;0x63uy;0x32uy;
//      0x39uy;0x65uy;0x62uy;0x62uy;0x38uy;0x64uy;0x30uy;0x61uy;0x36uy;0x33uy;0x30uy;
//      0x64uy;0x32uy;0x39uy;0x34uy;0x39uy;0x22uy;0x7duy;0x00uy;0x82uy;0x04uy;0x00uy;
//      0x00uy;0x28uy;0x46uy;0x31uy;0x64uy;0x64uy;0x30uy;0x30uy;0x35uy;0x64uy;0x63uy;
//      0x39uy;0x39uy;0x39uy;0x37uy;0x34uy;0x62uy;0x61uy;0x34uy;0x62uy;0x62uy;0x35uy;
//      0x64uy;0x62uy;0x35uy;0x64uy;0x63uy;0x30uy;0x62uy;0x66uy;0x62uy;0x31uy;0x34uy;
//      0x65uy;0x31uy;0x3auy;0x32uy;0x31uy;0xaeuy;0x03uy;0x7buy;0x22uy;0x72uy;0x65uy;
//      0x74uy;0x61uy;0x69uy;0x6cuy;0x53uy;0x6buy;0x75uy;0x49uy;0x64uy;0x22uy;0x3auy;
//      0x22uy;0x31uy;0x64uy;0x64uy;0x30uy;0x30uy;0x35uy;0x64uy;0x63uy;0x39uy;0x39uy;
//      0x39uy;0x37uy;0x34uy;0x62uy;0x61uy;0x34uy;0x62uy;0x62uy;0x35uy;0x64uy;0x62uy;
//      0x35uy;0x64uy;0x63uy;0x30uy;0x62uy;0x66uy;0x62uy;0x31uy;0x34uy;0x65uy;0x31uy;
//      0x22uy;0x2cuy;0x22uy;0x6duy;0x61uy;0x72uy;0x74uy;0x49uy;0x64uy;0x22uy;0x3auy;
//      0x32uy;0x31uy;0x2cuy;0x22uy;0x72uy;0x65uy;0x63uy;0x61uy;0x6cuy;0x63uy;0x75uy;
//      0x6cuy;0x61uy;0x74uy;0x69uy;0x6fuy;0x6euy;0x54uy;0x68uy;0x72uy;0x65uy;0x73uy;
//      0x68uy;0x6fuy;0x6cuy;0x64uy;0x22uy;0x3auy;0x30uy;0x2cuy;0x22uy;0x74uy;0x69uy;
//      0x6duy;0x65uy;0x73uy;0x74uy;0x61uy;0x6duy;0x70uy;0x22uy;0x3auy;0x22uy;0x32uy;
//      0x30uy;0x31uy;0x38uy;0x2duy;0x30uy;0x32uy;0x2duy;0x30uy;0x32uy;0x54uy;0x32uy;
//      0x32uy;0x3auy;0x30uy;0x36uy;0x3auy;0x31uy;0x37uy;0x5auy;0x22uy;0x2cuy;0x22uy;
//      0x72uy;0x65uy;0x61uy;0x73uy;0x6fuy;0x6euy;0x22uy;0x3auy;0x22uy;0x45uy;0x76uy;
//      0x65uy;0x6euy;0x74uy;0x22uy;0x2cuy;0x22uy;0x63uy;0x6fuy;0x6duy;0x6duy;0x65uy;
//      0x6euy;0x74uy;0x22uy;0x3auy;0x22uy;0x52uy;0x65uy;0x70uy;0x72uy;0x69uy;0x63uy;
//      0x65uy;0x20uy;0x61uy;0x6cuy;0x6cuy;0x20uy;0x66uy;0x72uy;0x6fuy;0x6duy;0x20uy;
//      0x4euy;0x6fuy;0x76uy;0x61uy;0x22uy;0x2cuy;0x22uy;0x69uy;0x64uy;0x22uy;0x3auy;
//      0x22uy;0x38uy;0x31uy;0x34uy;0x34uy;0x30uy;0x31uy;0x34uy;0x36uy;0x64uy;0x35uy;
//      0x62uy;0x62uy;0x34uy;0x37uy;0x39uy;0x61uy;0x38uy;0x32uy;0x66uy;0x66uy;0x63uy;
//      0x66uy;0x35uy;0x65uy;0x66uy;0x30uy;0x62uy;0x39uy;0x38uy;0x61uy;0x63uy;0x64uy;
//      0x22uy;0x7duy;0x00uy;0x82uy;0x04uy;0x00uy;0x00uy;0x2auy;0x46uy;0x31uy;0x63uy;
//      0x66uy;0x30uy;0x31uy;0x30uy;0x38uy;0x30uy;0x30uy;0x30uy;0x36uy;0x31uy;0x34uy;
//      0x61uy;0x63uy;0x36uy;0x38uy;0x31uy;0x30uy;0x37uy;0x30uy;0x30uy;0x30uy;0x30uy;
//      0x61uy;0x39uy;0x36uy;0x65uy;0x62uy;0x33uy;0x62uy;0x36uy;0x3auy;0x32uy;0x31uy;
//      0xaeuy;0x03uy;0x7buy;0x22uy;0x72uy;0x65uy;0x74uy;0x61uy;0x69uy;0x6cuy;0x53uy;
//      0x6buy;0x75uy;0x49uy;0x64uy;0x22uy;0x3auy;0x22uy;0x31uy;0x63uy;0x66uy;0x30uy;
//      0x31uy;0x30uy;0x38uy;0x30uy;0x30uy;0x30uy;0x36uy;0x31uy;0x34uy;0x61uy;0x63uy;
//      0x36uy;0x38uy;0x31uy;0x30uy;0x37uy;0x30uy;0x30uy;0x30uy;0x30uy;0x61uy;0x39uy;
//      0x36uy;0x65uy;0x62uy;0x33uy;0x62uy;0x36uy;0x22uy;0x2cuy;0x22uy;0x6duy;0x61uy;
//      0x72uy;0x74uy;0x49uy;0x64uy;0x22uy;0x3auy;0x32uy;0x31uy;0x2cuy;0x22uy;0x72uy;
//      0x65uy;0x63uy;0x61uy;0x6cuy;0x63uy;0x75uy;0x6cuy;0x61uy;0x74uy;0x69uy;0x6fuy;
//      0x6euy;0x54uy;0x68uy;0x72uy;0x65uy;0x73uy;0x68uy;0x6fuy;0x6cuy;0x64uy;0x22uy;
//      0x3auy;0x30uy;0x2cuy;0x22uy;0x74uy;0x69uy;0x6duy;0x65uy;0x73uy;0x74uy;0x61uy;
//      0x6duy;0x70uy;0x22uy;0x3auy;0x22uy;0x32uy;0x30uy;0x31uy;0x38uy;0x2duy;0x30uy;
//      0x32uy;0x2duy;0x30uy;0x32uy;0x54uy;0x32uy;0x32uy;0x3auy;0x30uy;0x36uy;0x3auy;
//      0x31uy;0x37uy;0x5auy;0x22uy;0x2cuy;0x22uy;0x72uy;0x65uy;0x61uy;0x73uy;0x6fuy;
//      0x6euy;0x22uy;0x3auy;0x22uy;0x45uy;0x76uy;0x65uy;0x6euy;0x74uy;0x22uy;0x2cuy;
//      0x22uy;0x63uy;0x6fuy;0x6duy;0x6duy;0x65uy;0x6euy;0x74uy;0x22uy;0x3auy;0x22uy;
//      0x52uy;0x65uy;0x70uy;0x72uy;0x69uy;0x63uy;0x65uy;0x20uy;0x61uy;0x6cuy;0x6cuy;
//      0x20uy;0x66uy;0x72uy;0x6fuy;0x6duy;0x20uy;0x4euy;0x6fuy;0x76uy;0x61uy;0x22uy;
//      0x2cuy;0x22uy;0x69uy;0x64uy;0x22uy;0x3auy;0x22uy;0x39uy;0x35uy;0x63uy;0x64uy;
//      0x34uy;0x64uy;0x63uy;0x64uy;0x31uy;0x39uy;0x38uy;0x30uy;0x34uy;0x37uy;0x64uy;
//      0x64uy;0x61uy;0x35uy;0x64uy;0x66uy;0x37uy;0x35uy;0x38uy;0x35uy;0x64uy;0x66uy;
//      0x33uy;0x65uy;0x39uy;0x36uy;0x30uy;0x63uy;0x22uy;0x7duy;0x00uy;0x82uy;0x04uy;
//      0x00uy;0x00uy;0x2cuy;0x46uy;0x31uy;0x32uy;0x38uy;0x30uy;0x30uy;0x34uy;0x63uy;
//      0x34uy;0x31uy;0x62uy;0x31uy;0x66uy;0x34uy;0x34uy;0x62uy;0x32uy;0x38uy;0x38uy;
//      0x37uy;0x61uy;0x33uy;0x66uy;0x38uy;0x64uy;0x66uy;0x61uy;0x31uy;0x39uy;0x64uy;
//      0x66uy;0x65uy;0x36uy;0x3auy;0x32uy;0x31uy;0xaeuy;0x03uy;0x7buy;0x22uy;0x72uy;
//      0x65uy;0x74uy;0x61uy;0x69uy;0x6cuy;0x53uy;0x6buy;0x75uy;0x49uy;0x64uy;0x22uy;
//      0x3auy;0x22uy;0x31uy;0x32uy;0x38uy;0x30uy;0x30uy;0x34uy;0x63uy;0x34uy;0x31uy;
//      0x62uy;0x31uy;0x66uy;0x34uy;0x34uy;0x62uy;0x32uy;0x38uy;0x38uy;0x37uy;0x61uy;
//      0x33uy;0x66uy;0x38uy;0x64uy;0x66uy;0x61uy;0x31uy;0x39uy;0x64uy;0x66uy;0x65uy;
//      0x36uy;0x22uy;0x2cuy;0x22uy;0x6duy;0x61uy;0x72uy;0x74uy;0x49uy;0x64uy;0x22uy;
//      0x3auy;0x32uy;0x31uy;0x2cuy;0x22uy;0x72uy;0x65uy;0x63uy;0x61uy;0x6cuy;0x63uy;
//      0x75uy;0x6cuy;0x61uy;0x74uy;0x69uy;0x6fuy;0x6euy;0x54uy;0x68uy;0x72uy;0x65uy;
//      0x73uy;0x68uy;0x6fuy;0x6cuy;0x64uy;0x22uy;0x3auy;0x30uy;0x2cuy;0x22uy;0x74uy;
//      0x69uy;0x6duy;0x65uy;0x73uy;0x74uy;0x61uy;0x6duy;0x70uy;0x22uy;0x3auy;0x22uy;
//      0x32uy;0x30uy;0x31uy;0x38uy;0x2duy;0x30uy;0x32uy;0x2duy;0x30uy;0x32uy;0x54uy;
//      0x32uy;0x32uy;0x3auy;0x30uy;0x36uy;0x3auy;0x31uy;0x37uy;0x5auy;0x22uy;0x2cuy;
//      0x22uy;0x72uy;0x65uy;0x61uy;0x73uy;0x6fuy;0x6euy;0x22uy;0x3auy;0x22uy;0x45uy;
//      0x76uy;0x65uy;0x6euy;0x74uy;0x22uy;0x2cuy;0x22uy;0x63uy;0x6fuy;0x6duy;0x6duy;
//      0x65uy;0x6euy;0x74uy;0x22uy;0x3auy;0x22uy;0x52uy;0x65uy;0x70uy;0x72uy;0x69uy;
//      0x63uy;0x65uy;0x20uy;0x61uy;0x6cuy;0x6cuy;0x20uy;0x66uy;0x72uy;0x6fuy;0x6duy;
//      0x20uy;0x4euy;0x6fuy;0x76uy;0x61uy;0x22uy;0x2cuy;0x22uy;0x69uy;0x64uy;0x22uy;
//      0x3auy;0x22uy;0x62uy;0x61uy;0x63uy;0x32uy;0x32uy;0x32uy;0x35uy;0x33uy;0x32uy;
//      0x63uy;0x31uy;0x31uy;0x34uy;0x34uy;0x38uy;0x34uy;0x39uy;0x31uy;0x61uy;0x33uy;
//      0x62uy;0x64uy;0x66uy;0x31uy;0x65uy;0x33uy;0x39uy;0x32uy;0x64uy;0x38uy;0x30uy;
//      0x30uy;0x22uy;0x7duy;0x00uy;0x82uy;0x04uy;0x00uy;0x00uy;0x2euy;0x46uy;0x31uy;
//      0x62uy;0x62uy;0x30uy;0x30uy;0x35uy;0x63uy;0x34uy;0x64uy;0x32uy;0x32uy;0x36uy;
//      0x34uy;0x63uy;0x66uy;0x38uy;0x38uy;0x31uy;0x37uy;0x39uy;0x31uy;0x33uy;0x66uy;
//      0x61uy;0x65uy;0x36uy;0x66uy;0x61uy;0x30uy;0x33uy;0x61uy;0x65uy;0x3auy;0x32uy;
//      0x31uy;0xaeuy;0x03uy;0x7buy;0x22uy;0x72uy;0x65uy;0x74uy;0x61uy;0x69uy;0x6cuy;
//      0x53uy;0x6buy;0x75uy;0x49uy;0x64uy;0x22uy;0x3auy;0x22uy;0x31uy;0x62uy;0x62uy;
//      0x30uy;0x30uy;0x35uy;0x63uy;0x34uy;0x64uy;0x32uy;0x32uy;0x36uy;0x34uy;0x63uy;
//      0x66uy;0x38uy;0x38uy;0x31uy;0x37uy;0x39uy;0x31uy;0x33uy;0x66uy;0x61uy;0x65uy;
//      0x36uy;0x66uy;0x61uy;0x30uy;0x33uy;0x61uy;0x65uy;0x22uy;0x2cuy;0x22uy;0x6duy;
//      0x61uy;0x72uy;0x74uy;0x49uy;0x64uy;0x22uy;0x3auy;0x32uy;0x31uy;0x2cuy;0x22uy;
//      0x72uy;0x65uy;0x63uy;0x61uy;0x6cuy;0x63uy;0x75uy;0x6cuy;0x61uy;0x74uy;0x69uy;
//      0x6fuy;0x6euy;0x54uy;0x68uy;0x72uy;0x65uy;0x73uy;0x68uy;0x6fuy;0x6cuy;0x64uy;
//      0x22uy;0x3auy;0x30uy;0x2cuy;0x22uy;0x74uy;0x69uy;0x6duy;0x65uy;0x73uy;0x74uy;
//      0x61uy;0x6duy;0x70uy;0x22uy;0x3auy;0x22uy;0x32uy;0x30uy;0x31uy;0x38uy;0x2duy;
//      0x30uy;0x32uy;0x2duy;0x30uy;0x32uy;0x54uy;0x32uy;0x32uy;0x3auy;0x30uy;0x36uy;
//      0x3auy;0x31uy;0x37uy;0x5auy;0x22uy;0x2cuy;0x22uy;0x72uy;0x65uy;0x61uy;0x73uy;
//      0x6fuy;0x6euy;0x22uy;0x3auy;0x22uy;0x45uy;0x76uy;0x65uy;0x6euy;0x74uy;0x22uy;
//      0x2cuy;0x22uy;0x63uy;0x6fuy;0x6duy;0x6duy;0x65uy;0x6euy;0x74uy;0x22uy;0x3auy;
//      0x22uy;0x52uy;0x65uy;0x70uy;0x72uy;0x69uy;0x63uy;0x65uy;0x20uy;0x61uy;0x6cuy;
//      0x6cuy;0x20uy;0x66uy;0x72uy;0x6fuy;0x6duy;0x20uy;0x4euy;0x6fuy;0x76uy;0x61uy;
//      0x22uy;0x2cuy;0x22uy;0x69uy;0x64uy;0x22uy;0x3auy;0x22uy;0x36uy;0x37uy;0x37uy;
//      0x35uy;0x65uy;0x32uy;0x63uy;0x65uy;0x62uy;0x33uy;0x30uy;0x64uy;0x34uy;0x37uy;
//      0x61uy;0x36uy;0x61uy;0x65uy;0x33uy;0x34uy;0x66uy;0x34uy;0x63uy;0x65uy;0x31uy;
//      0x36uy;0x32uy;0x66uy;0x36uy;0x30uy;0x32uy;0x62uy;0x22uy;0x7duy;0x00uy;0x82uy;
//      0x04uy;0x00uy;0x00uy;0x30uy;0x46uy;0x31uy;0x62uy;0x33uy;0x30uy;0x30uy;0x33uy;
//      0x33uy;0x64uy;0x33uy;0x64uy;0x30uy;0x36uy;0x34uy;0x35uy;0x30uy;0x66uy;0x61uy;
//      0x62uy;0x32uy;0x61uy;0x32uy;0x65uy;0x34uy;0x63uy;0x32uy;0x39uy;0x34uy;0x33uy;
//      0x31uy;0x30uy;0x63uy;0x66uy;0x3auy;0x32uy;0x31uy;0xaeuy;0x03uy;0x7buy;0x22uy;
//      0x72uy;0x65uy;0x74uy;0x61uy;0x69uy;0x6cuy;0x53uy;0x6buy;0x75uy;0x49uy;0x64uy;
//      0x22uy;0x3auy;0x22uy;0x31uy;0x62uy;0x33uy;0x30uy;0x30uy;0x33uy;0x33uy;0x64uy;
//      0x33uy;0x64uy;0x30uy;0x36uy;0x34uy;0x35uy;0x30uy;0x66uy;0x61uy;0x62uy;0x32uy;
//      0x61uy;0x32uy;0x65uy;0x34uy;0x63uy;0x32uy;0x39uy;0x34uy;0x33uy;0x31uy;0x30uy;
//      0x63uy;0x66uy;0x22uy;0x2cuy;0x22uy;0x6duy;0x61uy;0x72uy;0x74uy;0x49uy;0x64uy;
//      0x22uy;0x3auy;0x32uy;0x31uy;0x2cuy;0x22uy;0x72uy;0x65uy;0x63uy;0x61uy;0x6cuy;
//      0x63uy;0x75uy;0x6cuy;0x61uy;0x74uy;0x69uy;0x6fuy;0x6euy;0x54uy;0x68uy;0x72uy;
//      0x65uy;0x73uy;0x68uy;0x6fuy;0x6cuy;0x64uy;0x22uy;0x3auy;0x30uy;0x2cuy;0x22uy;
//      0x74uy;0x69uy;0x6duy;0x65uy;0x73uy;0x74uy;0x61uy;0x6duy;0x70uy;0x22uy;0x3auy;
//      0x22uy;0x32uy;0x30uy;0x31uy;0x38uy;0x2duy;0x30uy;0x32uy;0x2duy;0x30uy;0x32uy;
//      0x54uy;0x32uy;0x32uy;0x3auy;0x30uy;0x36uy;0x3auy;0x31uy;0x37uy;0x5auy;0x22uy;
//      0x2cuy;0x22uy;0x72uy;0x65uy;0x61uy;0x73uy;0x6fuy;0x6euy;0x22uy;0x3auy;0x22uy;
//      0x45uy;0x76uy;0x65uy;0x6euy;0x74uy;0x22uy;0x2cuy;0x22uy;0x63uy;0x6fuy;0x6duy;
//      0x6duy;0x65uy;0x6euy;0x74uy;0x22uy;0x3auy;0x22uy;0x52uy;0x65uy;0x70uy;0x72uy;
//      0x69uy;0x63uy;0x65uy;0x20uy;0x61uy;0x6cuy;0x6cuy;0x20uy;0x66uy;0x72uy;0x6fuy;
//      0x6duy;0x20uy;0x4euy;0x6fuy;0x76uy;0x61uy;0x22uy;0x2cuy;0x22uy;0x69uy;0x64uy;
//      0x22uy;0x3auy;0x22uy;0x36uy;0x39uy;0x33uy;0x35uy;0x31uy;0x35uy;0x36uy;0x39uy;
//      0x34uy;0x61uy;0x34uy;0x37uy;0x34uy;0x62uy;0x66uy;0x32uy;0x39uy;0x34uy;0x35uy;
//      0x35uy;0x34uy;0x36uy;0x65uy;0x37uy;0x31uy;0x31uy;0x64uy;0x66uy;0x64uy;0x35uy;
//      0x63uy;0x38uy;0x22uy;0x7duy;0x00uy;0x82uy;0x04uy;0x00uy;0x00uy;0x32uy;0x46uy;
//      0x31uy;0x61uy;0x62uy;0x30uy;0x30uy;0x37uy;0x32uy;0x31uy;0x34uy;0x31uy;0x38uy;
//      0x61uy;0x34uy;0x61uy;0x64uy;0x34uy;0x38uy;0x61uy;0x63uy;0x38uy;0x65uy;0x65uy;
//      0x33uy;0x33uy;0x30uy;0x38uy;0x31uy;0x62uy;0x62uy;0x31uy;0x37uy;0x62uy;0x3auy;
//      0x32uy;0x31uy;0xaeuy;0x03uy;0x7buy;0x22uy;0x72uy;0x65uy;0x74uy;0x61uy;0x69uy;
//      0x6cuy;0x53uy;0x6buy;0x75uy;0x49uy;0x64uy;0x22uy;0x3auy;0x22uy;0x31uy;0x61uy;
//      0x62uy;0x30uy;0x30uy;0x37uy;0x32uy;0x31uy;0x34uy;0x31uy;0x38uy;0x61uy;0x34uy;
//      0x61uy;0x64uy;0x34uy;0x38uy;0x61uy;0x63uy;0x38uy;0x65uy;0x65uy;0x33uy;0x33uy;
//      0x30uy;0x38uy;0x31uy;0x62uy;0x62uy;0x31uy;0x37uy;0x62uy;0x22uy;0x2cuy;0x22uy;
//      0x6duy;0x61uy;0x72uy;0x74uy;0x49uy;0x64uy;0x22uy;0x3auy;0x32uy;0x31uy;0x2cuy;
//      0x22uy;0x72uy;0x65uy;0x63uy;0x61uy;0x6cuy;0x63uy;0x75uy;0x6cuy;0x61uy;0x74uy;
//      0x69uy;0x6fuy;0x6euy;0x54uy;0x68uy;0x72uy;0x65uy;0x73uy;0x68uy;0x6fuy;0x6cuy;
//      0x64uy;0x22uy;0x3auy;0x30uy;0x2cuy;0x22uy;0x74uy;0x69uy;0x6duy;0x65uy;0x73uy;
//      0x74uy;0x61uy;0x6duy;0x70uy;0x22uy;0x3auy;0x22uy;0x32uy;0x30uy;0x31uy;0x38uy;
//      0x2duy;0x30uy;0x32uy;0x2duy;0x30uy;0x32uy;0x54uy;0x32uy;0x32uy;0x3auy;0x30uy;
//      0x36uy;0x3auy;0x31uy;0x37uy;0x5auy;0x22uy;0x2cuy;0x22uy;0x72uy;0x65uy;0x61uy;
//      0x73uy;0x6fuy;0x6euy;0x22uy;0x3auy;0x22uy;0x45uy;0x76uy;0x65uy;0x6euy;0x74uy;
//      0x22uy;0x2cuy;0x22uy;0x63uy;0x6fuy;0x6duy;0x6duy;0x65uy;0x6euy;0x74uy;0x22uy;
//      0x3auy;0x22uy;0x52uy;0x65uy;0x70uy;0x72uy;0x69uy;0x63uy;0x65uy;0x20uy;0x61uy;
//      0x6cuy;0x6cuy;0x20uy;0x66uy;0x72uy;0x6fuy;0x6duy;0x20uy;0x4euy;0x6fuy;0x76uy;
//      0x61uy;0x22uy;0x2cuy;0x22uy;0x69uy;0x64uy;0x22uy;0x3auy;0x22uy;0x63uy;0x31uy;
//      0x62uy;0x66uy;0x36uy;0x62uy;0x35uy;0x34uy;0x63uy;0x33uy;0x34uy;0x62uy;0x34uy;
//      0x63uy;0x66uy;0x31uy;0x61uy;0x30uy;0x65uy;0x39uy;0x35uy;0x35uy;0x61uy;0x30uy;
//      0x38uy;0x38uy;0x64uy;0x62uy;0x38uy;0x65uy;0x39uy;0x38uy;0x22uy;0x7duy;0x00uy;
//      0x82uy;0x04uy;0x00uy;0x00uy;0x34uy;0x46uy;0x31uy;0x39uy;0x64uy;0x30uy;0x30uy;
//      0x37uy;0x64uy;0x62uy;0x62uy;0x66uy;0x34uy;0x33uy;0x34uy;0x37uy;0x65uy;0x36uy;
//      0x61uy;0x39uy;0x32uy;0x34uy;0x37uy;0x37uy;0x34uy;0x61uy;0x33uy;0x33uy;0x38uy;
//      0x33uy;0x34uy;0x62uy;0x35uy;0x37uy;0x3auy;0x32uy;0x31uy;0xaeuy;0x03uy;0x7buy;
//      0x22uy;0x72uy;0x65uy;0x74uy;0x61uy;0x69uy;0x6cuy;0x53uy;0x6buy;0x75uy;0x49uy;
//      0x64uy;0x22uy;0x3auy;0x22uy;0x31uy;0x39uy;0x64uy;0x30uy;0x30uy;0x37uy;0x64uy;
//      0x62uy;0x62uy;0x66uy;0x34uy;0x33uy;0x34uy;0x37uy;0x65uy;0x36uy;0x61uy;0x39uy;
//      0x32uy;0x34uy;0x37uy;0x37uy;0x34uy;0x61uy;0x33uy;0x33uy;0x38uy;0x33uy;0x34uy;
//      0x62uy;0x35uy;0x37uy;0x22uy;0x2cuy;0x22uy;0x6duy;0x61uy;0x72uy;0x74uy;0x49uy;
//      0x64uy;0x22uy;0x3auy;0x32uy;0x31uy;0x2cuy;0x22uy;0x72uy;0x65uy;0x63uy;0x61uy;
//      0x6cuy;0x63uy;0x75uy;0x6cuy;0x61uy;0x74uy;0x69uy;0x6fuy;0x6euy;0x54uy;0x68uy;
//      0x72uy;0x65uy;0x73uy;0x68uy;0x6fuy;0x6cuy;0x64uy;0x22uy;0x3auy;0x30uy;0x2cuy;
//      0x22uy;0x74uy;0x69uy;0x6duy;0x65uy;0x73uy;0x74uy;0x61uy;0x6duy;0x70uy;0x22uy;
//      0x3auy;0x22uy;0x32uy;0x30uy;0x31uy;0x38uy;0x2duy;0x30uy;0x32uy;0x2duy;0x30uy;
//      0x32uy;0x54uy;0x32uy;0x32uy;0x3auy;0x30uy;0x36uy;0x3auy;0x31uy;0x37uy;0x5auy;
//      0x22uy;0x2cuy;0x22uy;0x72uy;0x65uy;0x61uy;0x73uy;0x6fuy;0x6euy;0x22uy;0x3auy;
//      0x22uy;0x45uy;0x76uy;0x65uy;0x6euy;0x74uy;0x22uy;0x2cuy;0x22uy;0x63uy;0x6fuy;
//      0x6duy;0x6duy;0x65uy;0x6euy;0x74uy;0x22uy;0x3auy;0x22uy;0x52uy;0x65uy;0x70uy;
//      0x72uy;0x69uy;0x63uy;0x65uy;0x20uy;0x61uy;0x6cuy;0x6cuy;0x20uy;0x66uy;0x72uy;
//      0x6fuy;0x6duy;0x20uy;0x4euy;0x6fuy;0x76uy;0x61uy;0x22uy;0x2cuy;0x22uy;0x69uy;
//      0x64uy;0x22uy;0x3auy;0x22uy;0x33uy;0x62uy;0x34uy;0x63uy;0x30uy;0x36uy;0x34uy;
//      0x31uy;0x38uy;0x37uy;0x61uy;0x61uy;0x34uy;0x31uy;0x34uy;0x61uy;0x61uy;0x65uy;
//      0x65uy;0x31uy;0x39uy;0x34uy;0x30uy;0x37uy;0x63uy;0x33uy;0x32uy;0x32uy;0x63uy;
//      0x38uy;0x39uy;0x39uy;0x22uy;0x7duy;0x00uy;0x82uy;0x04uy;0x00uy;0x00uy;0x36uy;
//      0x46uy;0x31uy;0x34uy;0x38uy;0x30uy;0x30uy;0x65uy;0x31uy;0x32uy;0x64uy;0x34uy;
//      0x66uy;0x61uy;0x34uy;0x30uy;0x64uy;0x64uy;0x61uy;0x35uy;0x31uy;0x62uy;0x62uy;
//      0x65uy;0x63uy;0x34uy;0x32uy;0x66uy;0x38uy;0x38uy;0x63uy;0x39uy;0x65uy;0x38uy;
//      0x3auy;0x32uy;0x31uy;0xaeuy;0x03uy;0x7buy;0x22uy;0x72uy;0x65uy;0x74uy;0x61uy;
//      0x69uy;0x6cuy;0x53uy;0x6buy;0x75uy;0x49uy;0x64uy;0x22uy;0x3auy;0x22uy;0x31uy;
//      0x34uy;0x38uy;0x30uy;0x30uy;0x65uy;0x31uy;0x32uy;0x64uy;0x34uy;0x66uy;0x61uy;
//      0x34uy;0x30uy;0x64uy;0x64uy;0x61uy;0x35uy;0x31uy;0x62uy;0x62uy;0x65uy;0x63uy;
//      0x34uy;0x32uy;0x66uy;0x38uy;0x38uy;0x63uy;0x39uy;0x65uy;0x38uy;0x22uy;0x2cuy;
//      0x22uy;0x6duy;0x61uy;0x72uy;0x74uy;0x49uy;0x64uy;0x22uy;0x3auy;0x32uy;0x31uy;
//      0x2cuy;0x22uy;0x72uy;0x65uy;0x63uy;0x61uy;0x6cuy;0x63uy;0x75uy;0x6cuy;0x61uy;
//      0x74uy;0x69uy;0x6fuy;0x6euy;0x54uy;0x68uy;0x72uy;0x65uy;0x73uy;0x68uy;0x6fuy;
//      0x6cuy;0x64uy;0x22uy;0x3auy;0x30uy;0x2cuy;0x22uy;0x74uy;0x69uy;0x6duy;0x65uy;
//      0x73uy;0x74uy;0x61uy;0x6duy;0x70uy;0x22uy;0x3auy;0x22uy;0x32uy;0x30uy;0x31uy;
//      0x38uy;0x2duy;0x30uy;0x32uy;0x2duy;0x30uy;0x32uy;0x54uy;0x32uy;0x32uy;0x3auy;
//      0x30uy;0x36uy;0x3auy;0x31uy;0x37uy;0x5auy;0x22uy;0x2cuy;0x22uy;0x72uy;0x65uy;
//      0x61uy;0x73uy;0x6fuy;0x6euy;0x22uy;0x3auy;0x22uy;0x45uy;0x76uy;0x65uy;0x6euy;
//      0x74uy;0x22uy;0x2cuy;0x22uy;0x63uy;0x6fuy;0x6duy;0x6duy;0x65uy;0x6euy;0x74uy;
//      0x22uy;0x3auy;0x22uy;0x52uy;0x65uy;0x70uy;0x72uy;0x69uy;0x63uy;0x65uy;0x20uy;
//      0x61uy;0x6cuy;0x6cuy;0x20uy;0x66uy;0x72uy;0x6fuy;0x6duy;0x20uy;0x4euy;0x6fuy;
//      0x76uy;0x61uy;0x22uy;0x2cuy;0x22uy;0x69uy;0x64uy;0x22uy;0x3auy;0x22uy;0x31uy;
//      0x64uy;0x65uy;0x36uy;0x66uy;0x65uy;0x64uy;0x31uy;0x64uy;0x30uy;0x38uy;0x63uy;
//      0x34uy;0x37uy;0x30uy;0x62uy;0x39uy;0x62uy;0x64uy;0x32uy;0x37uy;0x63uy;0x61uy;
//      0x32uy;0x35uy;0x64uy;0x62uy;0x32uy;0x61uy;0x66uy;0x35uy;0x65uy;0x22uy;0x7duy;
//      0x00uy;0x82uy;0x04uy;0x00uy;0x00uy;0x38uy;0x46uy;0x31uy;0x38uy;0x65uy;0x30uy;
//      0x30uy;0x36uy;0x66uy;0x38uy;0x64uy;0x38uy;0x37uy;0x63uy;0x34uy;0x61uy;0x33uy;
//      0x30uy;0x62uy;0x61uy;0x65uy;0x35uy;0x35uy;0x65uy;0x35uy;0x33uy;0x34uy;0x65uy;
//      0x38uy;0x32uy;0x61uy;0x36uy;0x66uy;0x36uy;0x3auy;0x32uy;0x31uy;0xaeuy;0x03uy;
//      0x7buy;0x22uy;0x72uy;0x65uy;0x74uy;0x61uy;0x69uy;0x6cuy;0x53uy;0x6buy;0x75uy;
//      0x49uy;0x64uy;0x22uy;0x3auy;0x22uy;0x31uy;0x38uy;0x65uy;0x30uy;0x30uy;0x36uy;
//      0x66uy;0x38uy;0x64uy;0x38uy;0x37uy;0x63uy;0x34uy;0x61uy;0x33uy;0x30uy;0x62uy;
//      0x61uy;0x65uy;0x35uy;0x35uy;0x65uy;0x35uy;0x33uy;0x34uy;0x65uy;0x38uy;0x32uy;
//      0x61uy;0x36uy;0x66uy;0x36uy;0x22uy;0x2cuy;0x22uy;0x6duy;0x61uy;0x72uy;0x74uy;
//      0x49uy;0x64uy;0x22uy;0x3auy;0x32uy;0x31uy;0x2cuy;0x22uy;0x72uy;0x65uy;0x63uy;
//      0x61uy;0x6cuy;0x63uy;0x75uy;0x6cuy;0x61uy;0x74uy;0x69uy;0x6fuy;0x6euy;0x54uy;
//      0x68uy;0x72uy;0x65uy;0x73uy;0x68uy;0x6fuy;0x6cuy;0x64uy;0x22uy;0x3auy;0x30uy;
//      0x2cuy;0x22uy;0x74uy;0x69uy;0x6duy;0x65uy;0x73uy;0x74uy;0x61uy;0x6duy;0x70uy;
//      0x22uy;0x3auy;0x22uy;0x32uy;0x30uy;0x31uy;0x38uy;0x2duy;0x30uy;0x32uy;0x2duy;
//      0x30uy;0x32uy;0x54uy;0x32uy;0x32uy;0x3auy;0x30uy;0x36uy;0x3auy;0x31uy;0x37uy;
//      0x5auy;0x22uy;0x2cuy;0x22uy;0x72uy;0x65uy;0x61uy;0x73uy;0x6fuy;0x6euy;0x22uy;
//      0x3auy;0x22uy;0x45uy;0x76uy;0x65uy;0x6euy;0x74uy;0x22uy;0x2cuy;0x22uy;0x63uy;
//      0x6fuy;0x6duy;0x6duy;0x65uy;0x6euy;0x74uy;0x22uy;0x3auy;0x22uy;0x52uy;0x65uy;
//      0x70uy;0x72uy;0x69uy;0x63uy;0x65uy;0x20uy;0x61uy;0x6cuy;0x6cuy;0x20uy;0x66uy;
//      0x72uy;0x6fuy;0x6duy;0x20uy;0x4euy;0x6fuy;0x76uy;0x61uy;0x22uy;0x2cuy;0x22uy;
//      0x69uy;0x64uy;0x22uy;0x3auy;0x22uy;0x33uy;0x64uy;0x39uy;0x33uy;0x66uy;0x36uy;
//      0x31uy;0x36uy;0x34uy;0x35uy;0x32uy;0x64uy;0x34uy;0x33uy;0x37uy;0x36uy;0x38uy;
//      0x31uy;0x66uy;0x31uy;0x61uy;0x65uy;0x31uy;0x36uy;0x34uy;0x62uy;0x34uy;0x36uy;
//      0x39uy;0x31uy;0x32uy;0x65uy;0x22uy;0x7duy;0x00uy;0x82uy;0x04uy;0x00uy;0x00uy;
//      0x3auy;0x46uy;0x31uy;0x38uy;0x36uy;0x30uy;0x31uy;0x33uy;0x35uy;0x32uy;0x31uy;
//      0x30uy;0x33uy;0x36uy;0x34uy;0x32uy;0x63uy;0x39uy;0x38uy;0x36uy;0x32uy;0x37uy;
//      0x65uy;0x31uy;0x31uy;0x61uy;0x34uy;0x30uy;0x32uy;0x39uy;0x66uy;0x37uy;0x61uy;
//      0x63uy;0x3auy;0x32uy;0x31uy;0xaeuy;0x03uy;0x7buy;0x22uy;0x72uy;0x65uy;0x74uy;
//      0x61uy;0x69uy;0x6cuy;0x53uy;0x6buy;0x75uy;0x49uy;0x64uy;0x22uy;0x3auy;0x22uy;
//      0x31uy;0x38uy;0x36uy;0x30uy;0x31uy;0x33uy;0x35uy;0x32uy;0x31uy;0x30uy;0x33uy;
//      0x36uy;0x34uy;0x32uy;0x63uy;0x39uy;0x38uy;0x36uy;0x32uy;0x37uy;0x65uy;0x31uy;
//      0x31uy;0x61uy;0x34uy;0x30uy;0x32uy;0x39uy;0x66uy;0x37uy;0x61uy;0x63uy;0x22uy;
//      0x2cuy;0x22uy;0x6duy;0x61uy;0x72uy;0x74uy;0x49uy;0x64uy;0x22uy;0x3auy;0x32uy;
//      0x31uy;0x2cuy;0x22uy;0x72uy;0x65uy;0x63uy;0x61uy;0x6cuy;0x63uy;0x75uy;0x6cuy;
//      0x61uy;0x74uy;0x69uy;0x6fuy;0x6euy;0x54uy;0x68uy;0x72uy;0x65uy;0x73uy;0x68uy;
//      0x6fuy;0x6cuy;0x64uy;0x22uy;0x3auy;0x30uy;0x2cuy;0x22uy;0x74uy;0x69uy;0x6duy;
//      0x65uy;0x73uy;0x74uy;0x61uy;0x6duy;0x70uy;0x22uy;0x3auy;0x22uy;0x32uy;0x30uy;
//      0x31uy;0x38uy;0x2duy;0x30uy;0x32uy;0x2duy;0x30uy;0x32uy;0x54uy;0x32uy;0x32uy;
//      0x3auy;0x30uy;0x36uy;0x3auy;0x31uy;0x37uy;0x5auy;0x22uy;0x2cuy;0x22uy;0x72uy;
//      0x65uy;0x61uy;0x73uy;0x6fuy;0x6euy;0x22uy;0x3auy;0x22uy;0x45uy;0x76uy;0x65uy;
//      0x6euy;0x74uy;0x22uy;0x2cuy;0x22uy;0x63uy;0x6fuy;0x6duy;0x6duy;0x65uy;0x6euy;
//      0x74uy;0x22uy;0x3auy;0x22uy;0x52uy;0x65uy;0x70uy;0x72uy;0x69uy;0x63uy;0x65uy;
//      0x20uy;0x61uy;0x6cuy;0x6cuy;0x20uy;0x66uy;0x72uy;0x6fuy;0x6duy;0x20uy;0x4euy;
//      0x6fuy;0x76uy;0x61uy;0x22uy;0x2cuy;0x22uy;0x69uy;0x64uy;0x22uy;0x3auy;0x22uy;
//      0x33uy;0x37uy;0x65uy;0x34uy;0x37uy;0x63uy;0x31uy;0x30uy;0x62uy;0x64uy;0x62uy;
//      0x63uy;0x34uy;0x62uy;0x61uy;0x63uy;0x39uy;0x63uy;0x63uy;0x36uy;0x35uy;0x61uy;
//      0x30uy;0x38uy;0x36uy;0x66uy;0x61uy;0x30uy;0x31uy;0x36uy;0x61uy;0x63uy;0x22uy;
//      0x7duy;0x00uy;0x82uy;0x04uy;0x00uy;0x00uy;0x3cuy;0x46uy;0x31uy;0x36uy;0x34uy;
//      0x30uy;0x30uy;0x63uy;0x38uy;0x65uy;0x39uy;0x62uy;0x38uy;0x63uy;0x34uy;0x61uy;
//      0x32uy;0x35uy;0x62uy;0x35uy;0x30uy;0x66uy;0x33uy;0x39uy;0x30uy;0x35uy;0x30uy;
//      0x63uy;0x62uy;0x61uy;0x35uy;0x63uy;0x65uy;0x63uy;0x3auy;0x32uy;0x31uy;0xaeuy;
//      0x03uy;0x7buy;0x22uy;0x72uy;0x65uy;0x74uy;0x61uy;0x69uy;0x6cuy;0x53uy;0x6buy;
//      0x75uy;0x49uy;0x64uy;0x22uy;0x3auy;0x22uy;0x31uy;0x36uy;0x34uy;0x30uy;0x30uy;
//      0x63uy;0x38uy;0x65uy;0x39uy;0x62uy;0x38uy;0x63uy;0x34uy;0x61uy;0x32uy;0x35uy;
//      0x62uy;0x35uy;0x30uy;0x66uy;0x33uy;0x39uy;0x30uy;0x35uy;0x30uy;0x63uy;0x62uy;
//      0x61uy;0x35uy;0x63uy;0x65uy;0x63uy;0x22uy;0x2cuy;0x22uy;0x6duy;0x61uy;0x72uy;
//      0x74uy;0x49uy;0x64uy;0x22uy;0x3auy;0x32uy;0x31uy;0x2cuy;0x22uy;0x72uy;0x65uy;
//      0x63uy;0x61uy;0x6cuy;0x63uy;0x75uy;0x6cuy;0x61uy;0x74uy;0x69uy;0x6fuy;0x6euy;
//      0x54uy;0x68uy;0x72uy;0x65uy;0x73uy;0x68uy;0x6fuy;0x6cuy;0x64uy;0x22uy;0x3auy;
//      0x30uy;0x2cuy;0x22uy;0x74uy;0x69uy;0x6duy;0x65uy;0x73uy;0x74uy;0x61uy;0x6duy;
//      0x70uy;0x22uy;0x3auy;0x22uy;0x32uy;0x30uy;0x31uy;0x38uy;0x2duy;0x30uy;0x32uy;
//      0x2duy;0x30uy;0x32uy;0x54uy;0x32uy;0x32uy;0x3auy;0x30uy;0x36uy;0x3auy;0x31uy;
//      0x37uy;0x5auy;0x22uy;0x2cuy;0x22uy;0x72uy;0x65uy;0x61uy;0x73uy;0x6fuy;0x6euy;
//      0x22uy;0x3auy;0x22uy;0x45uy;0x76uy;0x65uy;0x6euy;0x74uy;0x22uy;0x2cuy;0x22uy;
//      0x63uy;0x6fuy;0x6duy;0x6duy;0x65uy;0x6euy;0x74uy;0x22uy;0x3auy;0x22uy;0x52uy;
//      0x65uy;0x70uy;0x72uy;0x69uy;0x63uy;0x65uy;0x20uy;0x61uy;0x6cuy;0x6cuy;0x20uy;
//      0x66uy;0x72uy;0x6fuy;0x6duy;0x20uy;0x4euy;0x6fuy;0x76uy;0x61uy;0x22uy;0x2cuy;
//      0x22uy;0x69uy;0x64uy;0x22uy;0x3auy;0x22uy;0x39uy;0x30uy;0x66uy;0x64uy;0x30uy;
//      0x38uy;0x62uy;0x61uy;0x34uy;0x30uy;0x62uy;0x32uy;0x34uy;0x66uy;0x36uy;0x33uy;
//      0x39uy;0x34uy;0x37uy;0x32uy;0x36uy;0x35uy;0x38uy;0x65uy;0x32uy;0x65uy;0x65uy;
//      0x66uy;0x65uy;0x34uy;0x30uy;0x34uy;0x22uy;0x7duy;0x00uy;0x82uy;0x04uy;0x00uy;
//      0x00uy;0x3euy;0x46uy;0x31uy;0x36uy;0x37uy;0x30uy;0x30uy;0x35uy;0x30uy;0x35uy;
//      0x39uy;0x31uy;0x31uy;0x39uy;0x34uy;0x63uy;0x30uy;0x62uy;0x38uy;0x63uy;0x66uy;
//      0x62uy;0x36uy;0x62uy;0x38uy;0x65uy;0x62uy;0x63uy;0x65uy;0x63uy;0x39uy;0x34uy;
//      0x66uy;0x36uy;0x3auy;0x32uy;0x31uy;0xaeuy;0x03uy;0x7buy;0x22uy;0x72uy;0x65uy;
//      0x74uy;0x61uy;0x69uy;0x6cuy;0x53uy;0x6buy;0x75uy;0x49uy;0x64uy;0x22uy;0x3auy;
//      0x22uy;0x31uy;0x36uy;0x37uy;0x30uy;0x30uy;0x35uy;0x30uy;0x35uy;0x39uy;0x31uy;
//      0x31uy;0x39uy;0x34uy;0x63uy;0x30uy;0x62uy;0x38uy;0x63uy;0x66uy;0x62uy;0x36uy;
//      0x62uy;0x38uy;0x65uy;0x62uy;0x63uy;0x65uy;0x63uy;0x39uy;0x34uy;0x66uy;0x36uy;
//      0x22uy;0x2cuy;0x22uy;0x6duy;0x61uy;0x72uy;0x74uy;0x49uy;0x64uy;0x22uy;0x3auy;
//      0x32uy;0x31uy;0x2cuy;0x22uy;0x72uy;0x65uy;0x63uy;0x61uy;0x6cuy;0x63uy;0x75uy;
//      0x6cuy;0x61uy;0x74uy;0x69uy;0x6fuy;0x6euy;0x54uy;0x68uy;0x72uy;0x65uy;0x73uy;
//      0x68uy;0x6fuy;0x6cuy;0x64uy;0x22uy;0x3auy;0x30uy;0x2cuy;0x22uy;0x74uy;0x69uy;
//      0x6duy;0x65uy;0x73uy;0x74uy;0x61uy;0x6duy;0x70uy;0x22uy;0x3auy;0x22uy;0x32uy;
//      0x30uy;0x31uy;0x38uy;0x2duy;0x30uy;0x32uy;0x2duy;0x30uy;0x32uy;0x54uy;0x32uy;
//      0x32uy;0x3auy;0x30uy;0x36uy;0x3auy;0x31uy;0x37uy;0x5auy;0x22uy;0x2cuy;0x22uy;
//      0x72uy;0x65uy;0x61uy;0x73uy;0x6fuy;0x6euy;0x22uy;0x3auy;0x22uy;0x45uy;0x76uy;
//      0x65uy;0x6euy;0x74uy;0x22uy;0x2cuy;0x22uy;0x63uy;0x6fuy;0x6duy;0x6duy;0x65uy;
//      0x6euy;0x74uy;0x22uy;0x3auy;0x22uy;0x52uy;0x65uy;0x70uy;0x72uy;0x69uy;0x63uy;
//      0x65uy;0x20uy;0x61uy;0x6cuy;0x6cuy;0x20uy;0x66uy;0x72uy;0x6fuy;0x6duy;0x20uy;
//      0x4euy;0x6fuy;0x76uy;0x61uy;0x22uy;0x2cuy;0x22uy;0x69uy;0x64uy;0x22uy;0x3auy;
//      0x22uy;0x36uy;0x33uy;0x66uy;0x65uy;0x36uy;0x37uy;0x32uy;0x30uy;0x63uy;0x31uy;
//      0x39uy;0x31uy;0x34uy;0x33uy;0x61uy;0x32uy;0x61uy;0x34uy;0x38uy;0x66uy;0x66uy;
//      0x64uy;0x39uy;0x62uy;0x33uy;0x37uy;0x30uy;0x38uy;0x34uy;0x30uy;0x62uy;0x65uy;
//      0x22uy;0x7duy;0x00uy;0x82uy;0x04uy;0x00uy;0x00uy;0x40uy;0x46uy;0x31uy;0x30uy;
//      0x30uy;0x30uy;0x30uy;0x30uy;0x64uy;0x38uy;0x39uy;0x33uy;0x61uy;0x65uy;0x34uy;
//      0x32uy;0x61uy;0x34uy;0x62uy;0x61uy;0x33uy;0x62uy;0x39uy;0x36uy;0x33uy;0x34uy;
//      0x38uy;0x37uy;0x31uy;0x32uy;0x31uy;0x63uy;0x64uy;0x30uy;0x3auy;0x32uy;0x31uy;
//      0xaeuy;0x03uy;0x7buy;0x22uy;0x72uy;0x65uy;0x74uy;0x61uy;0x69uy;0x6cuy;0x53uy;
//      0x6buy;0x75uy;0x49uy;0x64uy;0x22uy;0x3auy;0x22uy;0x31uy;0x30uy;0x30uy;0x30uy;
//      0x30uy;0x30uy;0x64uy;0x38uy;0x39uy;0x33uy;0x61uy;0x65uy;0x34uy;0x32uy;0x61uy;
//      0x34uy;0x62uy;0x61uy;0x33uy;0x62uy;0x39uy;0x36uy;0x33uy;0x34uy;0x38uy;0x37uy;
//      0x31uy;0x32uy;0x31uy;0x63uy;0x64uy;0x30uy;0x22uy;0x2cuy;0x22uy;0x6duy;0x61uy;
//      0x72uy;0x74uy;0x49uy;0x64uy;0x22uy;0x3auy;0x32uy;0x31uy;0x2cuy;0x22uy;0x72uy;
//      0x65uy;0x63uy;0x61uy;0x6cuy;0x63uy;0x75uy;0x6cuy;0x61uy;0x74uy;0x69uy;0x6fuy;
//      0x6euy;0x54uy;0x68uy;0x72uy;0x65uy;0x73uy;0x68uy;0x6fuy;0x6cuy;0x64uy;0x22uy;
//      0x3auy;0x30uy;0x2cuy;0x22uy;0x74uy;0x69uy;0x6duy;0x65uy;0x73uy;0x74uy;0x61uy;
//      0x6duy;0x70uy;0x22uy;0x3auy;0x22uy;0x32uy;0x30uy;0x31uy;0x38uy;0x2duy;0x30uy;
//      0x32uy;0x2duy;0x30uy;0x32uy;0x54uy;0x32uy;0x32uy;0x3auy;0x30uy;0x36uy;0x3auy;
//      0x31uy;0x37uy;0x5auy;0x22uy;0x2cuy;0x22uy;0x72uy;0x65uy;0x61uy;0x73uy;0x6fuy;
//      0x6euy;0x22uy;0x3auy;0x22uy;0x45uy;0x76uy;0x65uy;0x6euy;0x74uy;0x22uy;0x2cuy;
//      0x22uy;0x63uy;0x6fuy;0x6duy;0x6duy;0x65uy;0x6euy;0x74uy;0x22uy;0x3auy;0x22uy;
//      0x52uy;0x65uy;0x70uy;0x72uy;0x69uy;0x63uy;0x65uy;0x20uy;0x61uy;0x6cuy;0x6cuy;
//      0x20uy;0x66uy;0x72uy;0x6fuy;0x6duy;0x20uy;0x4euy;0x6fuy;0x76uy;0x61uy;0x22uy;
//      0x2cuy;0x22uy;0x69uy;0x64uy;0x22uy;0x3auy;0x22uy;0x32uy;0x63uy;0x37uy;0x65uy;
//      0x34uy;0x61uy;0x64uy;0x34uy;0x62uy;0x30uy;0x64uy;0x31uy;0x34uy;0x61uy;0x36uy;
//      0x63uy;0x61uy;0x65uy;0x66uy;0x61uy;0x33uy;0x36uy;0x35uy;0x64uy;0x64uy;0x31uy;
//      0x61uy;0x37uy;0x31uy;0x31uy;0x33uy;0x36uy;0x22uy;0x7duy;0x00uy;0x82uy;0x04uy;
//      0x00uy;0x00uy;0x42uy;0x46uy;0x31uy;0x30uy;0x66uy;0x30uy;0x30uy;0x38uy;0x31uy;
//      0x33uy;0x34uy;0x63uy;0x37uy;0x62uy;0x34uy;0x33uy;0x38uy;0x31uy;0x61uy;0x30uy;
//      0x62uy;0x36uy;0x66uy;0x65uy;0x32uy;0x33uy;0x34uy;0x62uy;0x64uy;0x35uy;0x30uy;
//      0x39uy;0x64uy;0x62uy;0x3auy;0x32uy;0x31uy;0xaeuy;0x03uy;0x7buy;0x22uy;0x72uy;
//      0x65uy;0x74uy;0x61uy;0x69uy;0x6cuy;0x53uy;0x6buy;0x75uy;0x49uy;0x64uy;0x22uy;
//      0x3auy;0x22uy;0x31uy;0x30uy;0x66uy;0x30uy;0x30uy;0x38uy;0x31uy;0x33uy;0x34uy;
//      0x63uy;0x37uy;0x62uy;0x34uy;0x33uy;0x38uy;0x31uy;0x61uy;0x30uy;0x62uy;0x36uy;
//      0x66uy;0x65uy;0x32uy;0x33uy;0x34uy;0x62uy;0x64uy;0x35uy;0x30uy;0x39uy;0x64uy;
//      0x62uy;0x22uy;0x2cuy;0x22uy;0x6duy;0x61uy;0x72uy;0x74uy;0x49uy;0x64uy;0x22uy;
//      0x3auy;0x32uy;0x31uy;0x2cuy;0x22uy;0x72uy;0x65uy;0x63uy;0x61uy;0x6cuy;0x63uy;
//      0x75uy;0x6cuy;0x61uy;0x74uy;0x69uy;0x6fuy;0x6euy;0x54uy;0x68uy;0x72uy;0x65uy;
//      0x73uy;0x68uy;0x6fuy;0x6cuy;0x64uy;0x22uy;0x3auy;0x30uy;0x2cuy;0x22uy;0x74uy;
//      0x69uy;0x6duy;0x65uy;0x73uy;0x74uy;0x61uy;0x6duy;0x70uy;0x22uy;0x3auy;0x22uy;
//      0x32uy;0x30uy;0x31uy;0x38uy;0x2duy;0x30uy;0x32uy;0x2duy;0x30uy;0x32uy;0x54uy;
//      0x32uy;0x32uy;0x3auy;0x30uy;0x36uy;0x3auy;0x31uy;0x37uy;0x5auy;0x22uy;0x2cuy;
//      0x22uy;0x72uy;0x65uy;0x61uy;0x73uy;0x6fuy;0x6euy;0x22uy;0x3auy;0x22uy;0x45uy;
//      0x76uy;0x65uy;0x6euy;0x74uy;0x22uy;0x2cuy;0x22uy;0x63uy;0x6fuy;0x6duy;0x6duy;
//      0x65uy;0x6euy;0x74uy;0x22uy;0x3auy;0x22uy;0x52uy;0x65uy;0x70uy;0x72uy;0x69uy;
//      0x63uy;0x65uy;0x20uy;0x61uy;0x6cuy;0x6cuy;0x20uy;0x66uy;0x72uy;0x6fuy;0x6duy;
//      0x20uy;0x4euy;0x6fuy;0x76uy;0x61uy;0x22uy;0x2cuy;0x22uy;0x69uy;0x64uy;0x22uy;
//      0x3auy;0x22uy;0x39uy;0x66uy;0x38uy;0x33uy;0x65uy;0x39uy;0x37uy;0x63uy;0x32uy;
//      0x63uy;0x64uy;0x64uy;0x34uy;0x32uy;0x36uy;0x30uy;0x61uy;0x65uy;0x33uy;0x34uy;
//      0x64uy;0x61uy;0x36uy;0x31uy;0x35uy;0x38uy;0x61uy;0x34uy;0x33uy;0x37uy;0x36uy;
//      0x66uy;0x22uy;0x7duy;0x00uy;0x82uy;0x04uy;0x00uy;0x00uy;0x44uy;0x46uy;0x31uy;
//      0x66uy;0x36uy;0x30uy;0x30uy;0x61uy;0x33uy;0x64uy;0x31uy;0x38uy;0x38uy;0x61uy;
//      0x34uy;0x34uy;0x39uy;0x34uy;0x39uy;0x30uy;0x64uy;0x33uy;0x35uy;0x39uy;0x30uy;
//      0x38uy;0x38uy;0x37uy;0x63uy;0x35uy;0x34uy;0x63uy;0x39uy;0x37uy;0x3auy;0x32uy;
//      0x31uy;0xaeuy;0x03uy;0x7buy;0x22uy;0x72uy;0x65uy;0x74uy;0x61uy;0x69uy;0x6cuy;
//      0x53uy;0x6buy;0x75uy;0x49uy;0x64uy;0x22uy;0x3auy;0x22uy;0x31uy;0x66uy;0x36uy;
//      0x30uy;0x30uy;0x61uy;0x33uy;0x64uy;0x31uy;0x38uy;0x38uy;0x61uy;0x34uy;0x34uy;
//      0x39uy;0x34uy;0x39uy;0x30uy;0x64uy;0x33uy;0x35uy;0x39uy;0x30uy;0x38uy;0x38uy;
//      0x37uy;0x63uy;0x35uy;0x34uy;0x63uy;0x39uy;0x37uy;0x22uy;0x2cuy;0x22uy;0x6duy;
//      0x61uy;0x72uy;0x74uy;0x49uy;0x64uy;0x22uy;0x3auy;0x32uy;0x31uy;0x2cuy;0x22uy;
//      0x72uy;0x65uy;0x63uy;0x61uy;0x6cuy;0x63uy;0x75uy;0x6cuy;0x61uy;0x74uy;0x69uy;
//      0x6fuy;0x6euy;0x54uy;0x68uy;0x72uy;0x65uy;0x73uy;0x68uy;0x6fuy;0x6cuy;0x64uy;
//      0x22uy;0x3auy;0x30uy;0x2cuy;0x22uy;0x74uy;0x69uy;0x6duy;0x65uy;0x73uy;0x74uy;
//      0x61uy;0x6duy;0x70uy;0x22uy;0x3auy;0x22uy;0x32uy;0x30uy;0x31uy;0x38uy;0x2duy;
//      0x30uy;0x32uy;0x2duy;0x30uy;0x32uy;0x54uy;0x32uy;0x32uy;0x3auy;0x30uy;0x36uy;
//      0x3auy;0x31uy;0x37uy;0x5auy;0x22uy;0x2cuy;0x22uy;0x72uy;0x65uy;0x61uy;0x73uy;
//      0x6fuy;0x6euy;0x22uy;0x3auy;0x22uy;0x45uy;0x76uy;0x65uy;0x6euy;0x74uy;0x22uy;
//      0x2cuy;0x22uy;0x63uy;0x6fuy;0x6duy;0x6duy;0x65uy;0x6euy;0x74uy;0x22uy;0x3auy;
//      0x22uy;0x52uy;0x65uy;0x70uy;0x72uy;0x69uy;0x63uy;0x65uy;0x20uy;0x61uy;0x6cuy;
//      0x6cuy;0x20uy;0x66uy;0x72uy;0x6fuy;0x6duy;0x20uy;0x4euy;0x6fuy;0x76uy;0x61uy;
//      0x22uy;0x2cuy;0x22uy;0x69uy;0x64uy;0x22uy;0x3auy;0x22uy;0x39uy;0x33uy;0x65uy;
//      0x39uy;0x62uy;0x38uy;0x39uy;0x63uy;0x64uy;0x38uy;0x32uy;0x32uy;0x34uy;0x62uy;
//      0x35uy;0x33uy;0x39uy;0x31uy;0x36uy;0x63uy;0x36uy;0x61uy;0x66uy;0x35uy;0x39uy;
//      0x39uy;0x39uy;0x36uy;0x62uy;0x63uy;0x31uy;0x66uy;0x22uy;0x7duy;0x00uy;0x82uy;
//      0x04uy;0x00uy;0x00uy;0x46uy;0x46uy;0x31uy;0x65uy;0x66uy;0x30uy;0x30uy;0x34uy;
//      0x30uy;0x34uy;0x63uy;0x63uy;0x33uy;0x33uy;0x34uy;0x30uy;0x66uy;0x38uy;0x38uy;
//      0x36uy;0x30uy;0x62uy;0x35uy;0x65uy;0x62uy;0x63uy;0x64uy;0x63uy;0x35uy;0x63uy;
//      0x65uy;0x66uy;0x31uy;0x39uy;0x3auy;0x32uy;0x31uy;0xaeuy;0x03uy;0x7buy;0x22uy;
//      0x72uy;0x65uy;0x74uy;0x61uy;0x69uy;0x6cuy;0x53uy;0x6buy;0x75uy;0x49uy;0x64uy;
//      0x22uy;0x3auy;0x22uy;0x31uy;0x65uy;0x66uy;0x30uy;0x30uy;0x34uy;0x30uy;0x34uy;
//      0x63uy;0x63uy;0x33uy;0x33uy;0x34uy;0x30uy;0x66uy;0x38uy;0x38uy;0x36uy;0x30uy;
//      0x62uy;0x35uy;0x65uy;0x62uy;0x63uy;0x64uy;0x63uy;0x35uy;0x63uy;0x65uy;0x66uy;
//      0x31uy;0x39uy;0x22uy;0x2cuy;0x22uy;0x6duy;0x61uy;0x72uy;0x74uy;0x49uy;0x64uy;
//      0x22uy;0x3auy;0x32uy;0x31uy;0x2cuy;0x22uy;0x72uy;0x65uy;0x63uy;0x61uy;0x6cuy;
//      0x63uy;0x75uy;0x6cuy;0x61uy;0x74uy;0x69uy;0x6fuy;0x6euy;0x54uy;0x68uy;0x72uy;
//      0x65uy;0x73uy;0x68uy;0x6fuy;0x6cuy;0x64uy;0x22uy;0x3auy;0x30uy;0x2cuy;0x22uy;
//      0x74uy;0x69uy;0x6duy;0x65uy;0x73uy;0x74uy;0x61uy;0x6duy;0x70uy;0x22uy;0x3auy;
//      0x22uy;0x32uy;0x30uy;0x31uy;0x38uy;0x2duy;0x30uy;0x32uy;0x2duy;0x30uy;0x32uy;
//      0x54uy;0x32uy;0x32uy;0x3auy;0x30uy;0x36uy;0x3auy;0x31uy;0x37uy;0x5auy;0x22uy;
//      0x2cuy;0x22uy;0x72uy;0x65uy;0x61uy;0x73uy;0x6fuy;0x6euy;0x22uy;0x3auy;0x22uy;
//      0x45uy;0x76uy;0x65uy;0x6euy;0x74uy;0x22uy;0x2cuy;0x22uy;0x63uy;0x6fuy;0x6duy;
//      0x6duy;0x65uy;0x6euy;0x74uy;0x22uy;0x3auy;0x22uy;0x52uy;0x65uy;0x70uy;0x72uy;
//      0x69uy;0x63uy;0x65uy;0x20uy;0x61uy;0x6cuy;0x6cuy;0x20uy;0x66uy;0x72uy;0x6fuy;
//      0x6duy;0x20uy;0x4euy;0x6fuy;0x76uy;0x61uy;0x22uy;0x2cuy;0x22uy;0x69uy;0x64uy;
//      0x22uy;0x3auy;0x22uy;0x66uy;0x32uy;0x30uy;0x32uy;0x32uy;0x61uy;0x34uy;0x65uy;
//      0x33uy;0x30uy;0x62uy;0x39uy;0x34uy;0x35uy;0x38uy;0x36uy;0x62uy;0x61uy;0x38uy;
//      0x30uy;0x31uy;0x61uy;0x39uy;0x36uy;0x63uy;0x65uy;0x32uy;0x32uy;0x37uy;0x62uy;
//      0x37uy;0x61uy;0x22uy;0x7duy;0x00uy;0x82uy;0x04uy;0x00uy;0x00uy;0x48uy;0x46uy;
//      0x31uy;0x65uy;0x34uy;0x30uy;0x30uy;0x61uy;0x34uy;0x64uy;0x39uy;0x61uy;0x36uy;
//      0x62uy;0x34uy;0x63uy;0x64uy;0x37uy;0x38uy;0x33uy;0x64uy;0x34uy;0x32uy;0x61uy;
//      0x30uy;0x39uy;0x31uy;0x39uy;0x61uy;0x64uy;0x33uy;0x61uy;0x63uy;0x38uy;0x3auy;
//      0x32uy;0x31uy;0xaeuy;0x03uy;0x7buy;0x22uy;0x72uy;0x65uy;0x74uy;0x61uy;0x69uy;
//      0x6cuy;0x53uy;0x6buy;0x75uy;0x49uy;0x64uy;0x22uy;0x3auy;0x22uy;0x31uy;0x65uy;
//      0x34uy;0x30uy;0x30uy;0x61uy;0x34uy;0x64uy;0x39uy;0x61uy;0x36uy;0x62uy;0x34uy;
//      0x63uy;0x64uy;0x37uy;0x38uy;0x33uy;0x64uy;0x34uy;0x32uy;0x61uy;0x30uy;0x39uy;
//      0x31uy;0x39uy;0x61uy;0x64uy;0x33uy;0x61uy;0x63uy;0x38uy;0x22uy;0x2cuy;0x22uy;
//      0x6duy;0x61uy;0x72uy;0x74uy;0x49uy;0x64uy;0x22uy;0x3auy;0x32uy;0x31uy;0x2cuy;
//      0x22uy;0x72uy;0x65uy;0x63uy;0x61uy;0x6cuy;0x63uy;0x75uy;0x6cuy;0x61uy;0x74uy;
//      0x69uy;0x6fuy;0x6euy;0x54uy;0x68uy;0x72uy;0x65uy;0x73uy;0x68uy;0x6fuy;0x6cuy;
//      0x64uy;0x22uy;0x3auy;0x30uy;0x2cuy;0x22uy;0x74uy;0x69uy;0x6duy;0x65uy;0x73uy;
//      0x74uy;0x61uy;0x6duy;0x70uy;0x22uy;0x3auy;0x22uy;0x32uy;0x30uy;0x31uy;0x38uy;
//      0x2duy;0x30uy;0x32uy;0x2duy;0x30uy;0x32uy;0x54uy;0x32uy;0x32uy;0x3auy;0x30uy;
//      0x36uy;0x3auy;0x31uy;0x37uy;0x5auy;0x22uy;0x2cuy;0x22uy;0x72uy;0x65uy;0x61uy;
//      0x73uy;0x6fuy;0x6euy;0x22uy;0x3auy;0x22uy;0x45uy;0x76uy;0x65uy;0x6euy;0x74uy;
//      0x22uy;0x2cuy;0x22uy;0x63uy;0x6fuy;0x6duy;0x6duy;0x65uy;0x6euy;0x74uy;0x22uy;
//      0x3auy;0x22uy;0x52uy;0x65uy;0x70uy;0x72uy;0x69uy;0x63uy;0x65uy;0x20uy;0x61uy;
//      0x6cuy;0x6cuy;0x20uy;0x66uy;0x72uy;0x6fuy;0x6duy;0x20uy;0x4euy;0x6fuy;0x76uy;
//      0x61uy;0x22uy;0x2cuy;0x22uy;0x69uy;0x64uy;0x22uy;0x3auy;0x22uy;0x61uy;0x35uy;
//      0x30uy;0x66uy;0x38uy;0x32uy;0x62uy;0x33uy;0x37uy;0x65uy;0x39uy;0x61uy;0x34uy;
//      0x35uy;0x63uy;0x37uy;0x38uy;0x30uy;0x32uy;0x31uy;0x34uy;0x30uy;0x65uy;0x37uy;
//      0x65uy;0x30uy;0x64uy;0x63uy;0x32uy;0x35uy;0x63uy;0x62uy;0x22uy;0x7duy;0x00uy;
//      0x82uy;0x04uy;0x00uy;0x00uy;0x4auy;0x46uy;0x31uy;0x32uy;0x66uy;0x30uy;0x30uy;
//      0x62uy;0x34uy;0x65uy;0x36uy;0x37uy;0x31uy;0x61uy;0x34uy;0x39uy;0x63uy;0x34uy;
//      0x38uy;0x63uy;0x66uy;0x65uy;0x33uy;0x35uy;0x64uy;0x33uy;0x37uy;0x31uy;0x63uy;
//      0x37uy;0x62uy;0x61uy;0x65uy;0x38uy;0x3auy;0x32uy;0x31uy;0xaeuy;0x03uy;0x7buy;
//      0x22uy;0x72uy;0x65uy;0x74uy;0x61uy;0x69uy;0x6cuy;0x53uy;0x6buy;0x75uy;0x49uy;
//      0x64uy;0x22uy;0x3auy;0x22uy;0x31uy;0x32uy;0x66uy;0x30uy;0x30uy;0x62uy;0x34uy;
//      0x65uy;0x36uy;0x37uy;0x31uy;0x61uy;0x34uy;0x39uy;0x63uy;0x34uy;0x38uy;0x63uy;
//      0x66uy;0x65uy;0x33uy;0x35uy;0x64uy;0x33uy;0x37uy;0x31uy;0x63uy;0x37uy;0x62uy;
//      0x61uy;0x65uy;0x38uy;0x22uy;0x2cuy;0x22uy;0x6duy;0x61uy;0x72uy;0x74uy;0x49uy;
//      0x64uy;0x22uy;0x3auy;0x32uy;0x31uy;0x2cuy;0x22uy;0x72uy;0x65uy;0x63uy;0x61uy;
//      0x6cuy;0x63uy;0x75uy;0x6cuy;0x61uy;0x74uy;0x69uy;0x6fuy;0x6euy;0x54uy;0x68uy;
//      0x72uy;0x65uy;0x73uy;0x68uy;0x6fuy;0x6cuy;0x64uy;0x22uy;0x3auy;0x30uy;0x2cuy;
//      0x22uy;0x74uy;0x69uy;0x6duy;0x65uy;0x73uy;0x74uy;0x61uy;0x6duy;0x70uy;0x22uy;
//      0x3auy;0x22uy;0x32uy;0x30uy;0x31uy;0x38uy;0x2duy;0x30uy;0x32uy;0x2duy;0x30uy;
//      0x32uy;0x54uy;0x32uy;0x32uy;0x3auy;0x30uy;0x36uy;0x3auy;0x31uy;0x37uy;0x5auy;
//      0x22uy;0x2cuy;0x22uy;0x72uy;0x65uy;0x61uy;0x73uy;0x6fuy;0x6euy;0x22uy;0x3auy;
//      0x22uy;0x45uy;0x76uy;0x65uy;0x6euy;0x74uy;0x22uy;0x2cuy;0x22uy;0x63uy;0x6fuy;
//      0x6duy;0x6duy;0x65uy;0x6euy;0x74uy;0x22uy;0x3auy;0x22uy;0x52uy;0x65uy;0x70uy;
//      0x72uy;0x69uy;0x63uy;0x65uy;0x20uy;0x61uy;0x6cuy;0x6cuy;0x20uy;0x66uy;0x72uy;
//      0x6fuy;0x6duy;0x20uy;0x4euy;0x6fuy;0x76uy;0x61uy;0x22uy;0x2cuy;0x22uy;0x69uy;
//      0x64uy;0x22uy;0x3auy;0x22uy;0x62uy;0x62uy;0x64uy;0x30uy;0x35uy;0x31uy;0x30uy;
//      0x34uy;0x63uy;0x35uy;0x35uy;0x36uy;0x34uy;0x39uy;0x30uy;0x39uy;0x61uy;0x63uy;
//      0x32uy;0x63uy;0x62uy;0x36uy;0x65uy;0x35uy;0x61uy;0x33uy;0x64uy;0x63uy;0x33uy;
//      0x38uy;0x36uy;0x62uy;0x22uy;0x7duy;0x00uy;0x82uy;0x04uy;0x00uy;0x00uy;0x4cuy;
//      0x46uy;0x31uy;0x64uy;0x38uy;0x30uy;0x30uy;0x38uy;0x35uy;0x34uy;0x64uy;0x33uy;
//      0x32uy;0x38uy;0x34uy;0x61uy;0x65uy;0x31uy;0x38uy;0x62uy;0x63uy;0x38uy;0x36uy;
//      0x63uy;0x38uy;0x30uy;0x66uy;0x33uy;0x30uy;0x32uy;0x62uy;0x31uy;0x38uy;0x61uy;
//      0x3auy;0x32uy;0x31uy;0xaeuy;0x03uy;0x7buy;0x22uy;0x72uy;0x65uy;0x74uy;0x61uy;
//      0x69uy;0x6cuy;0x53uy;0x6buy;0x75uy;0x49uy;0x64uy;0x22uy;0x3auy;0x22uy;0x31uy;
//      0x64uy;0x38uy;0x30uy;0x30uy;0x38uy;0x35uy;0x34uy;0x64uy;0x33uy;0x32uy;0x38uy;
//      0x34uy;0x61uy;0x65uy;0x31uy;0x38uy;0x62uy;0x63uy;0x38uy;0x36uy;0x63uy;0x38uy;
//      0x30uy;0x66uy;0x33uy;0x30uy;0x32uy;0x62uy;0x31uy;0x38uy;0x61uy;0x22uy;0x2cuy;
//      0x22uy;0x6duy;0x61uy;0x72uy;0x74uy;0x49uy;0x64uy;0x22uy;0x3auy;0x32uy;0x31uy;
//      0x2cuy;0x22uy;0x72uy;0x65uy;0x63uy;0x61uy;0x6cuy;0x63uy;0x75uy;0x6cuy;0x61uy;
//      0x74uy;0x69uy;0x6fuy;0x6euy;0x54uy;0x68uy;0x72uy;0x65uy;0x73uy;0x68uy;0x6fuy;
//      0x6cuy;0x64uy;0x22uy;0x3auy;0x30uy;0x2cuy;0x22uy;0x74uy;0x69uy;0x6duy;0x65uy;
//      0x73uy;0x74uy;0x61uy;0x6duy;0x70uy;0x22uy;0x3auy;0x22uy;0x32uy;0x30uy;0x31uy;
//      0x38uy;0x2duy;0x30uy;0x32uy;0x2duy;0x30uy;0x32uy;0x54uy;0x32uy;0x32uy;0x3auy;
//      0x30uy;0x36uy;0x3auy;0x31uy;0x37uy;0x5auy;0x22uy;0x2cuy;0x22uy;0x72uy;0x65uy;
//      0x61uy;0x73uy;0x6fuy;0x6euy;0x22uy;0x3auy;0x22uy;0x45uy;0x76uy;0x65uy;0x6euy;
//      0x74uy;0x22uy;0x2cuy;0x22uy;0x63uy;0x6fuy;0x6duy;0x6duy;0x65uy;0x6euy;0x74uy;
//      0x22uy;0x3auy;0x22uy;0x52uy;0x65uy;0x70uy;0x72uy;0x69uy;0x63uy;0x65uy;0x20uy;
//      0x61uy;0x6cuy;0x6cuy;0x20uy;0x66uy;0x72uy;0x6fuy;0x6duy;0x20uy;0x4euy;0x6fuy;
//      0x76uy;0x61uy;0x22uy;0x2cuy;0x22uy;0x69uy;0x64uy;0x22uy;0x3auy;0x22uy;0x32uy;
//      0x32uy;0x61uy;0x36uy;0x31uy;0x31uy;0x33uy;0x35uy;0x35uy;0x33uy;0x63uy;0x34uy;
//      0x34uy;0x39uy;0x31uy;0x30uy;0x39uy;0x34uy;0x65uy;0x39uy;0x34uy;0x65uy;0x61uy;
//      0x64uy;0x61uy;0x37uy;0x33uy;0x36uy;0x31uy;0x63uy;0x32uy;0x37uy;0x22uy;0x7duy;
//      0x00uy;0x82uy;0x04uy;0x00uy;0x00uy;0x4euy;0x46uy;0x31uy;0x63uy;0x65uy;0x30uy;
//      0x30uy;0x37uy;0x36uy;0x65uy;0x34uy;0x61uy;0x30uy;0x33uy;0x34uy;0x38uy;0x61uy;
//      0x65uy;0x62uy;0x63uy;0x38uy;0x66uy;0x36uy;0x33uy;0x35uy;0x39uy;0x66uy;0x66uy;
//      0x61uy;0x64uy;0x66uy;0x63uy;0x61uy;0x35uy;0x3auy;0x32uy;0x31uy;0xaeuy;0x03uy;
//      0x7buy;0x22uy;0x72uy;0x65uy;0x74uy;0x61uy;0x69uy;0x6cuy;0x53uy;0x6buy;0x75uy;
//      0x49uy;0x64uy;0x22uy;0x3auy;0x22uy;0x31uy;0x63uy;0x65uy;0x30uy;0x30uy;0x37uy;
//      0x36uy;0x65uy;0x34uy;0x61uy;0x30uy;0x33uy;0x34uy;0x38uy;0x61uy;0x65uy;0x62uy;
//      0x63uy;0x38uy;0x66uy;0x36uy;0x33uy;0x35uy;0x39uy;0x66uy;0x66uy;0x61uy;0x64uy;
//      0x66uy;0x63uy;0x61uy;0x35uy;0x22uy;0x2cuy;0x22uy;0x6duy;0x61uy;0x72uy;0x74uy;
//      0x49uy;0x64uy;0x22uy;0x3auy;0x32uy;0x31uy;0x2cuy;0x22uy;0x72uy;0x65uy;0x63uy;
//      0x61uy;0x6cuy;0x63uy;0x75uy;0x6cuy;0x61uy;0x74uy;0x69uy;0x6fuy;0x6euy;0x54uy;
//      0x68uy;0x72uy;0x65uy;0x73uy;0x68uy;0x6fuy;0x6cuy;0x64uy;0x22uy;0x3auy;0x30uy;
//      0x2cuy;0x22uy;0x74uy;0x69uy;0x6duy;0x65uy;0x73uy;0x74uy;0x61uy;0x6duy;0x70uy;
//      0x22uy;0x3auy;0x22uy;0x32uy;0x30uy;0x31uy;0x38uy;0x2duy;0x30uy;0x32uy;0x2duy;
//      0x30uy;0x32uy;0x54uy;0x32uy;0x32uy;0x3auy;0x30uy;0x36uy;0x3auy;0x31uy;0x37uy;
//      0x5auy;0x22uy;0x2cuy;0x22uy;0x72uy;0x65uy;0x61uy;0x73uy;0x6fuy;0x6euy;0x22uy;
//      0x3auy;0x22uy;0x45uy;0x76uy;0x65uy;0x6euy;0x74uy;0x22uy;0x2cuy;0x22uy;0x63uy;
//      0x6fuy;0x6duy;0x6duy;0x65uy;0x6euy;0x74uy;0x22uy;0x3auy;0x22uy;0x52uy;0x65uy;
//      0x70uy;0x72uy;0x69uy;0x63uy;0x65uy;0x20uy;0x61uy;0x6cuy;0x6cuy;0x20uy;0x66uy;
//      0x72uy;0x6fuy;0x6duy;0x20uy;0x4euy;0x6fuy;0x76uy;0x61uy;0x22uy;0x2cuy;0x22uy;
//      0x69uy;0x64uy;0x22uy;0x3auy;0x22uy;0x33uy;0x33uy;0x37uy;0x66uy;0x63uy;0x64uy;
//      0x39uy;0x66uy;0x32uy;0x64uy;0x33uy;0x30uy;0x34uy;0x64uy;0x30uy;0x35uy;0x62uy;
//      0x34uy;0x38uy;0x63uy;0x66uy;0x39uy;0x34uy;0x31uy;0x30uy;0x64uy;0x32uy;0x61uy;
//      0x33uy;0x32uy;0x33uy;0x30uy;0x22uy;0x7duy;0x00uy;0x82uy;0x04uy;0x00uy;0x00uy;
//      0x50uy;0x46uy;0x31uy;0x34uy;0x36uy;0x30uy;0x30uy;0x62uy;0x64uy;0x38uy;0x33uy;
//      0x61uy;0x39uy;0x37uy;0x34uy;0x38uy;0x36uy;0x38uy;0x39uy;0x33uy;0x36uy;0x65uy;
//      0x33uy;0x37uy;0x38uy;0x39uy;0x32uy;0x62uy;0x37uy;0x31uy;0x39uy;0x34uy;0x39uy;
//      0x31uy;0x3auy;0x32uy;0x31uy;0xaeuy;0x03uy;0x7buy;0x22uy;0x72uy;0x65uy;0x74uy;
//      0x61uy;0x69uy;0x6cuy;0x53uy;0x6buy;0x75uy;0x49uy;0x64uy;0x22uy;0x3auy;0x22uy;
//      0x31uy;0x34uy;0x36uy;0x30uy;0x30uy;0x62uy;0x64uy;0x38uy;0x33uy;0x61uy;0x39uy;
//      0x37uy;0x34uy;0x38uy;0x36uy;0x38uy;0x39uy;0x33uy;0x36uy;0x65uy;0x33uy;0x37uy;
//      0x38uy;0x39uy;0x32uy;0x62uy;0x37uy;0x31uy;0x39uy;0x34uy;0x39uy;0x31uy;0x22uy;
//      0x2cuy;0x22uy;0x6duy;0x61uy;0x72uy;0x74uy;0x49uy;0x64uy;0x22uy;0x3auy;0x32uy;
//      0x31uy;0x2cuy;0x22uy;0x72uy;0x65uy;0x63uy;0x61uy;0x6cuy;0x63uy;0x75uy;0x6cuy;
//      0x61uy;0x74uy;0x69uy;0x6fuy;0x6euy;0x54uy;0x68uy;0x72uy;0x65uy;0x73uy;0x68uy;
//      0x6fuy;0x6cuy;0x64uy;0x22uy;0x3auy;0x30uy;0x2cuy;0x22uy;0x74uy;0x69uy;0x6duy;
//      0x65uy;0x73uy;0x74uy;0x61uy;0x6duy;0x70uy;0x22uy;0x3auy;0x22uy;0x32uy;0x30uy;
//      0x31uy;0x38uy;0x2duy;0x30uy;0x32uy;0x2duy;0x30uy;0x32uy;0x54uy;0x32uy;0x32uy;
//      0x3auy;0x30uy;0x36uy;0x3auy;0x31uy;0x37uy;0x5auy;0x22uy;0x2cuy;0x22uy;0x72uy;
//      0x65uy;0x61uy;0x73uy;0x6fuy;0x6euy;0x22uy;0x3auy;0x22uy;0x45uy;0x76uy;0x65uy;
//      0x6euy;0x74uy;0x22uy;0x2cuy;0x22uy;0x63uy;0x6fuy;0x6duy;0x6duy;0x65uy;0x6euy;
//      0x74uy;0x22uy;0x3auy;0x22uy;0x52uy;0x65uy;0x70uy;0x72uy;0x69uy;0x63uy;0x65uy;
//      0x20uy;0x61uy;0x6cuy;0x6cuy;0x20uy;0x66uy;0x72uy;0x6fuy;0x6duy;0x20uy;0x4euy;
//      0x6fuy;0x76uy;0x61uy;0x22uy;0x2cuy;0x22uy;0x69uy;0x64uy;0x22uy;0x3auy;0x22uy;
//      0x61uy;0x61uy;0x65uy;0x62uy;0x39uy;0x37uy;0x63uy;0x64uy;0x34uy;0x34uy;0x31uy;
//      0x39uy;0x34uy;0x30uy;0x66uy;0x32uy;0x61uy;0x39uy;0x64uy;0x35uy;0x36uy;0x38uy;
//      0x30uy;0x33uy;0x61uy;0x64uy;0x64uy;0x30uy;0x30uy;0x35uy;0x38uy;0x61uy;0x22uy;
//      0x7duy;0x00uy;0x82uy;0x04uy;0x00uy;0x00uy;0x52uy;0x46uy;0x31uy;0x34uy;0x66uy;
//      0x30uy;0x30uy;0x35uy;0x32uy;0x66uy;0x65uy;0x31uy;0x30uy;0x30uy;0x34uy;0x37uy;
//      0x61uy;0x36uy;0x61uy;0x31uy;0x61uy;0x31uy;0x39uy;0x39uy;0x36uy;0x63uy;0x36uy;
//      0x65uy;0x62uy;0x32uy;0x37uy;0x30uy;0x30uy;0x30uy;0x3auy;0x32uy;0x31uy;0xaeuy;
//      0x03uy;0x7buy;0x22uy;0x72uy;0x65uy;0x74uy;0x61uy;0x69uy;0x6cuy;0x53uy;0x6buy;
//      0x75uy;0x49uy;0x64uy;0x22uy;0x3auy;0x22uy;0x31uy;0x34uy;0x66uy;0x30uy;0x30uy;
//      0x35uy;0x32uy;0x66uy;0x65uy;0x31uy;0x30uy;0x30uy;0x34uy;0x37uy;0x61uy;0x36uy;
//      0x61uy;0x31uy;0x61uy;0x31uy;0x39uy;0x39uy;0x36uy;0x63uy;0x36uy;0x65uy;0x62uy;
//      0x32uy;0x37uy;0x30uy;0x30uy;0x30uy;0x22uy;0x2cuy;0x22uy;0x6duy;0x61uy;0x72uy;
//      0x74uy;0x49uy;0x64uy;0x22uy;0x3auy;0x32uy;0x31uy;0x2cuy;0x22uy;0x72uy;0x65uy;
//      0x63uy;0x61uy;0x6cuy;0x63uy;0x75uy;0x6cuy;0x61uy;0x74uy;0x69uy;0x6fuy;0x6euy;
//      0x54uy;0x68uy;0x72uy;0x65uy;0x73uy;0x68uy;0x6fuy;0x6cuy;0x64uy;0x22uy;0x3auy;
//      0x30uy;0x2cuy;0x22uy;0x74uy;0x69uy;0x6duy;0x65uy;0x73uy;0x74uy;0x61uy;0x6duy;
//      0x70uy;0x22uy;0x3auy;0x22uy;0x32uy;0x30uy;0x31uy;0x38uy;0x2duy;0x30uy;0x32uy;
//      0x2duy;0x30uy;0x32uy;0x54uy;0x32uy;0x32uy;0x3auy;0x30uy;0x36uy;0x3auy;0x31uy;
//      0x37uy;0x5auy;0x22uy;0x2cuy;0x22uy;0x72uy;0x65uy;0x61uy;0x73uy;0x6fuy;0x6euy;
//      0x22uy;0x3auy;0x22uy;0x45uy;0x76uy;0x65uy;0x6euy;0x74uy;0x22uy;0x2cuy;0x22uy;
//      0x63uy;0x6fuy;0x6duy;0x6duy;0x65uy;0x6euy;0x74uy;0x22uy;0x3auy;0x22uy;0x52uy;
//      0x65uy;0x70uy;0x72uy;0x69uy;0x63uy;0x65uy;0x20uy;0x61uy;0x6cuy;0x6cuy;0x20uy;
//      0x66uy;0x72uy;0x6fuy;0x6duy;0x20uy;0x4euy;0x6fuy;0x76uy;0x61uy;0x22uy;0x2cuy;
//      0x22uy;0x69uy;0x64uy;0x22uy;0x3auy;0x22uy;0x62uy;0x35uy;0x39uy;0x34uy;0x38uy;
//      0x36uy;0x37uy;0x32uy;0x37uy;0x66uy;0x62uy;0x36uy;0x34uy;0x32uy;0x31uy;0x66uy;
//      0x39uy;0x33uy;0x37uy;0x38uy;0x38uy;0x65uy;0x37uy;0x61uy;0x35uy;0x62uy;0x64uy;
//      0x39uy;0x65uy;0x35uy;0x38uy;0x35uy;0x22uy;0x7duy;0x00uy;0x82uy;0x04uy;0x00uy;
//      0x00uy;0x54uy;0x46uy;0x31uy;0x62uy;0x36uy;0x30uy;0x31uy;0x34uy;0x37uy;0x32uy;
//      0x38uy;0x30uy;0x38uy;0x63uy;0x34uy;0x64uy;0x34uy;0x36uy;0x39uy;0x63uy;0x62uy;
//      0x35uy;0x31uy;0x32uy;0x31uy;0x33uy;0x65uy;0x32uy;0x37uy;0x38uy;0x38uy;0x35uy;
//      0x64uy;0x30uy;0x3auy;0x32uy;0x31uy;0xaeuy;0x03uy;0x7buy;0x22uy;0x72uy;0x65uy;
//      0x74uy;0x61uy;0x69uy;0x6cuy;0x53uy;0x6buy;0x75uy;0x49uy;0x64uy;0x22uy;0x3auy;
//      0x22uy;0x31uy;0x62uy;0x36uy;0x30uy;0x31uy;0x34uy;0x37uy;0x32uy;0x38uy;0x30uy;
//      0x38uy;0x63uy;0x34uy;0x64uy;0x34uy;0x36uy;0x39uy;0x63uy;0x62uy;0x35uy;0x31uy;
//      0x32uy;0x31uy;0x33uy;0x65uy;0x32uy;0x37uy;0x38uy;0x38uy;0x35uy;0x64uy;0x30uy;
//      0x22uy;0x2cuy;0x22uy;0x6duy;0x61uy;0x72uy;0x74uy;0x49uy;0x64uy;0x22uy;0x3auy;
//      0x32uy;0x31uy;0x2cuy;0x22uy;0x72uy;0x65uy;0x63uy;0x61uy;0x6cuy;0x63uy;0x75uy;
//      0x6cuy;0x61uy;0x74uy;0x69uy;0x6fuy;0x6euy;0x54uy;0x68uy;0x72uy;0x65uy;0x73uy;
//      0x68uy;0x6fuy;0x6cuy;0x64uy;0x22uy;0x3auy;0x30uy;0x2cuy;0x22uy;0x74uy;0x69uy;
//      0x6duy;0x65uy;0x73uy;0x74uy;0x61uy;0x6duy;0x70uy;0x22uy;0x3auy;0x22uy;0x32uy;
//      0x30uy;0x31uy;0x38uy;0x2duy;0x30uy;0x32uy;0x2duy;0x30uy;0x32uy;0x54uy;0x32uy;
//      0x32uy;0x3auy;0x30uy;0x36uy;0x3auy;0x31uy;0x37uy;0x5auy;0x22uy;0x2cuy;0x22uy;
//      0x72uy;0x65uy;0x61uy;0x73uy;0x6fuy;0x6euy;0x22uy;0x3auy;0x22uy;0x45uy;0x76uy;
//      0x65uy;0x6euy;0x74uy;0x22uy;0x2cuy;0x22uy;0x63uy;0x6fuy;0x6duy;0x6duy;0x65uy;
//      0x6euy;0x74uy;0x22uy;0x3auy;0x22uy;0x52uy;0x65uy;0x70uy;0x72uy;0x69uy;0x63uy;
//      0x65uy;0x20uy;0x61uy;0x6cuy;0x6cuy;0x20uy;0x66uy;0x72uy;0x6fuy;0x6duy;0x20uy;
//      0x4euy;0x6fuy;0x76uy;0x61uy;0x22uy;0x2cuy;0x22uy;0x69uy;0x64uy;0x22uy;0x3auy;
//      0x22uy;0x30uy;0x61uy;0x37uy;0x64uy;0x34uy;0x62uy;0x39uy;0x62uy;0x65uy;0x64uy;
//      0x66uy;0x62uy;0x34uy;0x65uy;0x63uy;0x61uy;0x62uy;0x61uy;0x65uy;0x33uy;0x66uy;
//      0x36uy;0x63uy;0x32uy;0x34uy;0x38uy;0x36uy;0x62uy;0x35uy;0x37uy;0x36uy;0x36uy;
//      0x22uy;0x7duy;0x00uy;0x82uy;0x04uy;0x00uy;0x00uy;0x56uy;0x46uy;0x31uy;0x61uy;
//      0x66uy;0x30uy;0x30uy;0x65uy;0x36uy;0x65uy;0x36uy;0x66uy;0x63uy;0x63uy;0x34uy;
//      0x33uy;0x62uy;0x62uy;0x62uy;0x38uy;0x38uy;0x64uy;0x39uy;0x62uy;0x63uy;0x33uy;
//      0x39uy;0x63uy;0x33uy;0x30uy;0x34uy;0x63uy;0x63uy;0x30uy;0x3auy;0x32uy;0x31uy;
//      0xaeuy;0x03uy;0x7buy;0x22uy;0x72uy;0x65uy;0x74uy;0x61uy;0x69uy;0x6cuy;0x53uy;
//      0x6buy;0x75uy;0x49uy;0x64uy;0x22uy;0x3auy;0x22uy;0x31uy;0x61uy;0x66uy;0x30uy;
//      0x30uy;0x65uy;0x36uy;0x65uy;0x36uy;0x66uy;0x63uy;0x63uy;0x34uy;0x33uy;0x62uy;
//      0x62uy;0x62uy;0x38uy;0x38uy;0x64uy;0x39uy;0x62uy;0x63uy;0x33uy;0x39uy;0x63uy;
//      0x33uy;0x30uy;0x34uy;0x63uy;0x63uy;0x30uy;0x22uy;0x2cuy;0x22uy;0x6duy;0x61uy;
//      0x72uy;0x74uy;0x49uy;0x64uy;0x22uy;0x3auy;0x32uy;0x31uy;0x2cuy;0x22uy;0x72uy;
//      0x65uy;0x63uy;0x61uy;0x6cuy;0x63uy;0x75uy;0x6cuy;0x61uy;0x74uy;0x69uy;0x6fuy;
//      0x6euy;0x54uy;0x68uy;0x72uy;0x65uy;0x73uy;0x68uy;0x6fuy;0x6cuy;0x64uy;0x22uy;
//      0x3auy;0x30uy;0x2cuy;0x22uy;0x74uy;0x69uy;0x6duy;0x65uy;0x73uy;0x74uy;0x61uy;
//      0x6duy;0x70uy;0x22uy;0x3auy;0x22uy;0x32uy;0x30uy;0x31uy;0x38uy;0x2duy;0x30uy;
//      0x32uy;0x2duy;0x30uy;0x32uy;0x54uy;0x32uy;0x32uy;0x3auy;0x30uy;0x36uy;0x3auy;
//      0x31uy;0x37uy;0x5auy;0x22uy;0x2cuy;0x22uy;0x72uy;0x65uy;0x61uy;0x73uy;0x6fuy;
//      0x6euy;0x22uy;0x3auy;0x22uy;0x45uy;0x76uy;0x65uy;0x6euy;0x74uy;0x22uy;0x2cuy;
//      0x22uy;0x63uy;0x6fuy;0x6duy;0x6duy;0x65uy;0x6euy;0x74uy;0x22uy;0x3auy;0x22uy;
//      0x52uy;0x65uy;0x70uy;0x72uy;0x69uy;0x63uy;0x65uy;0x20uy;0x61uy;0x6cuy;0x6cuy;
//      0x20uy;0x66uy;0x72uy;0x6fuy;0x6duy;0x20uy;0x4euy;0x6fuy;0x76uy;0x61uy;0x22uy;
//      0x2cuy;0x22uy;0x69uy;0x64uy;0x22uy;0x3auy;0x22uy;0x61uy;0x33uy;0x34uy;0x37uy;
//      0x35uy;0x33uy;0x31uy;0x39uy;0x61uy;0x62uy;0x32uy;0x39uy;0x34uy;0x65uy;0x35uy;
//      0x36uy;0x39uy;0x66uy;0x31uy;0x66uy;0x32uy;0x63uy;0x32uy;0x61uy;0x32uy;0x37uy;
//      0x33uy;0x33uy;0x66uy;0x31uy;0x35uy;0x61uy;0x22uy;0x7duy;0x00uy;0x82uy;0x04uy;
//      0x00uy;0x00uy;0x58uy;0x46uy;0x31uy;0x61uy;0x38uy;0x30uy;0x31uy;0x65uy;0x31uy;
//      0x34uy;0x31uy;0x32uy;0x34uy;0x62uy;0x34uy;0x32uy;0x38uy;0x62uy;0x61uy;0x66uy;
//      0x36uy;0x38uy;0x61uy;0x36uy;0x35uy;0x32uy;0x63uy;0x37uy;0x32uy;0x36uy;0x66uy;
//      0x36uy;0x39uy;0x62uy;0x3auy;0x32uy;0x31uy;0xaeuy;0x03uy;0x7buy;0x22uy;0x72uy;
//      0x65uy;0x74uy;0x61uy;0x69uy;0x6cuy;0x53uy;0x6buy;0x75uy;0x49uy;0x64uy;0x22uy;
//      0x3auy;0x22uy;0x31uy;0x61uy;0x38uy;0x30uy;0x31uy;0x65uy;0x31uy;0x34uy;0x31uy;
//      0x32uy;0x34uy;0x62uy;0x34uy;0x32uy;0x38uy;0x62uy;0x61uy;0x66uy;0x36uy;0x38uy;
//      0x61uy;0x36uy;0x35uy;0x32uy;0x63uy;0x37uy;0x32uy;0x36uy;0x66uy;0x36uy;0x39uy;
//      0x62uy;0x22uy;0x2cuy;0x22uy;0x6duy;0x61uy;0x72uy;0x74uy;0x49uy;0x64uy;0x22uy;
//      0x3auy;0x32uy;0x31uy;0x2cuy;0x22uy;0x72uy;0x65uy;0x63uy;0x61uy;0x6cuy;0x63uy;
//      0x75uy;0x6cuy;0x61uy;0x74uy;0x69uy;0x6fuy;0x6euy;0x54uy;0x68uy;0x72uy;0x65uy;
//      0x73uy;0x68uy;0x6fuy;0x6cuy;0x64uy;0x22uy;0x3auy;0x30uy;0x2cuy;0x22uy;0x74uy;
//      0x69uy;0x6duy;0x65uy;0x73uy;0x74uy;0x61uy;0x6duy;0x70uy;0x22uy;0x3auy;0x22uy;
//      0x32uy;0x30uy;0x31uy;0x38uy;0x2duy;0x30uy;0x32uy;0x2duy;0x30uy;0x32uy;0x54uy;
//      0x32uy;0x32uy;0x3auy;0x30uy;0x36uy;0x3auy;0x31uy;0x37uy;0x5auy;0x22uy;0x2cuy;
//      0x22uy;0x72uy;0x65uy;0x61uy;0x73uy;0x6fuy;0x6euy;0x22uy;0x3auy;0x22uy;0x45uy;
//      0x76uy;0x65uy;0x6euy;0x74uy;0x22uy;0x2cuy;0x22uy;0x63uy;0x6fuy;0x6duy;0x6duy;
//      0x65uy;0x6euy;0x74uy;0x22uy;0x3auy;0x22uy;0x52uy;0x65uy;0x70uy;0x72uy;0x69uy;
//      0x63uy;0x65uy;0x20uy;0x61uy;0x6cuy;0x6cuy;0x20uy;0x66uy;0x72uy;0x6fuy;0x6duy;
//      0x20uy;0x4euy;0x6fuy;0x76uy;0x61uy;0x22uy;0x2cuy;0x22uy;0x69uy;0x64uy;0x22uy;
//      0x3auy;0x22uy;0x32uy;0x32uy;0x66uy;0x30uy;0x62uy;0x37uy;0x30uy;0x39uy;0x35uy;
//      0x39uy;0x34uy;0x38uy;0x34uy;0x64uy;0x33uy;0x62uy;0x38uy;0x37uy;0x30uy;0x66uy;
//      0x31uy;0x66uy;0x62uy;0x64uy;0x33uy;0x39uy;0x31uy;0x34uy;0x38uy;0x32uy;0x30uy;
//      0x33uy;0x22uy;0x7duy;0x00uy;0x82uy;0x04uy;0x00uy;0x00uy;0x5auy;0x46uy;0x31uy;
//      0x61uy;0x30uy;0x30uy;0x30uy;0x65uy;0x38uy;0x61uy;0x31uy;0x66uy;0x35uy;0x65uy;
//      0x34uy;0x36uy;0x65uy;0x38uy;0x61uy;0x37uy;0x66uy;0x33uy;0x62uy;0x30uy;0x65uy;
//      0x34uy;0x66uy;0x35uy;0x39uy;0x61uy;0x62uy;0x39uy;0x66uy;0x36uy;0x3auy;0x32uy;
//      0x31uy;0xaeuy;0x03uy;0x7buy;0x22uy;0x72uy;0x65uy;0x74uy;0x61uy;0x69uy;0x6cuy;
//      0x53uy;0x6buy;0x75uy;0x49uy;0x64uy;0x22uy;0x3auy;0x22uy;0x31uy;0x61uy;0x30uy;
//      0x30uy;0x30uy;0x65uy;0x38uy;0x61uy;0x31uy;0x66uy;0x35uy;0x65uy;0x34uy;0x36uy;
//      0x65uy;0x38uy;0x61uy;0x37uy;0x66uy;0x33uy;0x62uy;0x30uy;0x65uy;0x34uy;0x66uy;
//      0x35uy;0x39uy;0x61uy;0x62uy;0x39uy;0x66uy;0x36uy;0x22uy;0x2cuy;0x22uy;0x6duy;
//      0x61uy;0x72uy;0x74uy;0x49uy;0x64uy;0x22uy;0x3auy;0x32uy;0x31uy;0x2cuy;0x22uy;
//      0x72uy;0x65uy;0x63uy;0x61uy;0x6cuy;0x63uy;0x75uy;0x6cuy;0x61uy;0x74uy;0x69uy;
//      0x6fuy;0x6euy;0x54uy;0x68uy;0x72uy;0x65uy;0x73uy;0x68uy;0x6fuy;0x6cuy;0x64uy;
//      0x22uy;0x3auy;0x30uy;0x2cuy;0x22uy;0x74uy;0x69uy;0x6duy;0x65uy;0x73uy;0x74uy;
//      0x61uy;0x6duy;0x70uy;0x22uy;0x3auy;0x22uy;0x32uy;0x30uy;0x31uy;0x38uy;0x2duy;
//      0x30uy;0x32uy;0x2duy;0x30uy;0x32uy;0x54uy;0x32uy;0x32uy;0x3auy;0x30uy;0x36uy;
//      0x3auy;0x31uy;0x37uy;0x5auy;0x22uy;0x2cuy;0x22uy;0x72uy;0x65uy;0x61uy;0x73uy;
//      0x6fuy;0x6euy;0x22uy;0x3auy;0x22uy;0x45uy;0x76uy;0x65uy;0x6euy;0x74uy;0x22uy;
//      0x2cuy;0x22uy;0x63uy;0x6fuy;0x6duy;0x6duy;0x65uy;0x6euy;0x74uy;0x22uy;0x3auy;
//      0x22uy;0x52uy;0x65uy;0x70uy;0x72uy;0x69uy;0x63uy;0x65uy;0x20uy;0x61uy;0x6cuy;
//      0x6cuy;0x20uy;0x66uy;0x72uy;0x6fuy;0x6duy;0x20uy;0x4euy;0x6fuy;0x76uy;0x61uy;
//      0x22uy;0x2cuy;0x22uy;0x69uy;0x64uy;0x22uy;0x3auy;0x22uy;0x62uy;0x65uy;0x31uy;
//      0x35uy;0x34uy;0x34uy;0x38uy;0x66uy;0x36uy;0x64uy;0x65uy;0x30uy;0x34uy;0x65uy;
//      0x66uy;0x63uy;0x62uy;0x38uy;0x30uy;0x31uy;0x61uy;0x33uy;0x61uy;0x33uy;0x38uy;
//      0x34uy;0x66uy;0x35uy;0x31uy;0x66uy;0x66uy;0x66uy;0x22uy;0x7duy;0x00uy;0x82uy;
//      0x04uy;0x00uy;0x00uy;0x5cuy;0x46uy;0x31uy;0x39uy;0x37uy;0x30uy;0x30uy;0x66uy;
//      0x62uy;0x64uy;0x31uy;0x32uy;0x33uy;0x36uy;0x34uy;0x66uy;0x35uy;0x31uy;0x61uy;
//      0x31uy;0x31uy;0x63uy;0x33uy;0x65uy;0x32uy;0x38uy;0x66uy;0x31uy;0x35uy;0x31uy;
//      0x32uy;0x31uy;0x39uy;0x34uy;0x3auy;0x32uy;0x31uy;0xaeuy;0x03uy;0x7buy;0x22uy;
//      0x72uy;0x65uy;0x74uy;0x61uy;0x69uy;0x6cuy;0x53uy;0x6buy;0x75uy;0x49uy;0x64uy;
//      0x22uy;0x3auy;0x22uy;0x31uy;0x39uy;0x37uy;0x30uy;0x30uy;0x66uy;0x62uy;0x64uy;
//      0x31uy;0x32uy;0x33uy;0x36uy;0x34uy;0x66uy;0x35uy;0x31uy;0x61uy;0x31uy;0x31uy;
//      0x63uy;0x33uy;0x65uy;0x32uy;0x38uy;0x66uy;0x31uy;0x35uy;0x31uy;0x32uy;0x31uy;
//      0x39uy;0x34uy;0x22uy;0x2cuy;0x22uy;0x6duy;0x61uy;0x72uy;0x74uy;0x49uy;0x64uy;
//      0x22uy;0x3auy;0x32uy;0x31uy;0x2cuy;0x22uy;0x72uy;0x65uy;0x63uy;0x61uy;0x6cuy;
//      0x63uy;0x75uy;0x6cuy;0x61uy;0x74uy;0x69uy;0x6fuy;0x6euy;0x54uy;0x68uy;0x72uy;
//      0x65uy;0x73uy;0x68uy;0x6fuy;0x6cuy;0x64uy;0x22uy;0x3auy;0x30uy;0x2cuy;0x22uy;
//      0x74uy;0x69uy;0x6duy;0x65uy;0x73uy;0x74uy;0x61uy;0x6duy;0x70uy;0x22uy;0x3auy;
//      0x22uy;0x32uy;0x30uy;0x31uy;0x38uy;0x2duy;0x30uy;0x32uy;0x2duy;0x30uy;0x32uy;
//      0x54uy;0x32uy;0x32uy;0x3auy;0x30uy;0x36uy;0x3auy;0x31uy;0x37uy;0x5auy;0x22uy;
//      0x2cuy;0x22uy;0x72uy;0x65uy;0x61uy;0x73uy;0x6fuy;0x6euy;0x22uy;0x3auy;0x22uy;
//      0x45uy;0x76uy;0x65uy;0x6euy;0x74uy;0x22uy;0x2cuy;0x22uy;0x63uy;0x6fuy;0x6duy;
//      0x6duy;0x65uy;0x6euy;0x74uy;0x22uy;0x3auy;0x22uy;0x52uy;0x65uy;0x70uy;0x72uy;
//      0x69uy;0x63uy;0x65uy;0x20uy;0x61uy;0x6cuy;0x6cuy;0x20uy;0x66uy;0x72uy;0x6fuy;
//      0x6duy;0x20uy;0x4euy;0x6fuy;0x76uy;0x61uy;0x22uy;0x2cuy;0x22uy;0x69uy;0x64uy;
//      0x22uy;0x3auy;0x22uy;0x35uy;0x63uy;0x64uy;0x31uy;0x37uy;0x64uy;0x31uy;0x65uy;
//      0x35uy;0x34uy;0x35uy;0x64uy;0x34uy;0x65uy;0x64uy;0x30uy;0x38uy;0x30uy;0x37uy;
//      0x36uy;0x62uy;0x32uy;0x66uy;0x66uy;0x65uy;0x31uy;0x64uy;0x38uy;0x32uy;0x30uy;
//      0x30uy;0x30uy;0x22uy;0x7duy;0x00uy;0x82uy;0x04uy;0x00uy;0x00uy;0x5euy;0x46uy;
//      0x31uy;0x38uy;0x66uy;0x30uy;0x30uy;0x32uy;0x39uy;0x30uy;0x30uy;0x34uy;0x34uy;
//      0x63uy;0x34uy;0x62uy;0x64uy;0x36uy;0x61uy;0x35uy;0x66uy;0x37uy;0x37uy;0x32uy;
//      0x61uy;0x35uy;0x66uy;0x33uy;0x32uy;0x36uy;0x37uy;0x37uy;0x63uy;0x63uy;0x3auy;
//      0x32uy;0x31uy;0xaeuy;0x03uy;0x7buy;0x22uy;0x72uy;0x65uy;0x74uy;0x61uy;0x69uy;
//      0x6cuy;0x53uy;0x6buy;0x75uy;0x49uy;0x64uy;0x22uy;0x3auy;0x22uy;0x31uy;0x38uy;
//      0x66uy;0x30uy;0x30uy;0x32uy;0x39uy;0x30uy;0x30uy;0x34uy;0x34uy;0x63uy;0x34uy;
//      0x62uy;0x64uy;0x36uy;0x61uy;0x35uy;0x66uy;0x37uy;0x37uy;0x32uy;0x61uy;0x35uy;
//      0x66uy;0x33uy;0x32uy;0x36uy;0x37uy;0x37uy;0x63uy;0x63uy;0x22uy;0x2cuy;0x22uy;
//      0x6duy;0x61uy;0x72uy;0x74uy;0x49uy;0x64uy;0x22uy;0x3auy;0x32uy;0x31uy;0x2cuy;
//      0x22uy;0x72uy;0x65uy;0x63uy;0x61uy;0x6cuy;0x63uy;0x75uy;0x6cuy;0x61uy;0x74uy;
//      0x69uy;0x6fuy;0x6euy;0x54uy;0x68uy;0x72uy;0x65uy;0x73uy;0x68uy;0x6fuy;0x6cuy;
//      0x64uy;0x22uy;0x3auy;0x30uy;0x2cuy;0x22uy;0x74uy;0x69uy;0x6duy;0x65uy;0x73uy;
//      0x74uy;0x61uy;0x6duy;0x70uy;0x22uy;0x3auy;0x22uy;0x32uy;0x30uy;0x31uy;0x38uy;
//      0x2duy;0x30uy;0x32uy;0x2duy;0x30uy;0x32uy;0x54uy;0x32uy;0x32uy;0x3auy;0x30uy;
//      0x36uy;0x3auy;0x31uy;0x37uy;0x5auy;0x22uy;0x2cuy;0x22uy;0x72uy;0x65uy;0x61uy;
//      0x73uy;0x6fuy;0x6euy;0x22uy;0x3auy;0x22uy;0x45uy;0x76uy;0x65uy;0x6euy;0x74uy;
//      0x22uy;0x2cuy;0x22uy;0x63uy;0x6fuy;0x6duy;0x6duy;0x65uy;0x6euy;0x74uy;0x22uy;
//      0x3auy;0x22uy;0x52uy;0x65uy;0x70uy;0x72uy;0x69uy;0x63uy;0x65uy;0x20uy;0x61uy;
//      0x6cuy;0x6cuy;0x20uy;0x66uy;0x72uy;0x6fuy;0x6duy;0x20uy;0x4euy;0x6fuy;0x76uy;
//      0x61uy;0x22uy;0x2cuy;0x22uy;0x69uy;0x64uy;0x22uy;0x3auy;0x22uy;0x32uy;0x36uy;
//      0x33uy;0x64uy;0x35uy;0x30uy;0x61uy;0x39uy;0x38uy;0x35uy;0x33uy;0x38uy;0x34uy;
//      0x62uy;0x66uy;0x63uy;0x62uy;0x39uy;0x38uy;0x35uy;0x38uy;0x64uy;0x66uy;0x34uy;
//      0x35uy;0x37uy;0x32uy;0x31uy;0x63uy;0x36uy;0x34uy;0x32uy;0x22uy;0x7duy;0x00uy;
//      0x82uy;0x04uy;0x00uy;0x00uy;0x60uy;0x46uy;0x31uy;0x30uy;0x32uy;0x30uy;0x30uy;
//      0x33uy;0x66uy;0x31uy;0x36uy;0x35uy;0x34uy;0x35uy;0x34uy;0x34uy;0x35uy;0x64uy;
//      0x38uy;0x61uy;0x37uy;0x39uy;0x30uy;0x34uy;0x65uy;0x62uy;0x62uy;0x66uy;0x66uy;
//      0x37uy;0x35uy;0x62uy;0x30uy;0x38uy;0x3auy;0x32uy;0x31uy;0xaeuy;0x03uy;0x7buy;
//      0x22uy;0x72uy;0x65uy;0x74uy;0x61uy;0x69uy;0x6cuy;0x53uy;0x6buy;0x75uy;0x49uy;
//      0x64uy;0x22uy;0x3auy;0x22uy;0x31uy;0x30uy;0x32uy;0x30uy;0x30uy;0x33uy;0x66uy;
//      0x31uy;0x36uy;0x35uy;0x34uy;0x35uy;0x34uy;0x34uy;0x35uy;0x64uy;0x38uy;0x61uy;
//      0x37uy;0x39uy;0x30uy;0x34uy;0x65uy;0x62uy;0x62uy;0x66uy;0x66uy;0x37uy;0x35uy;
//      0x62uy;0x30uy;0x38uy;0x22uy;0x2cuy;0x22uy;0x6duy;0x61uy;0x72uy;0x74uy;0x49uy;
//      0x64uy;0x22uy;0x3auy;0x32uy;0x31uy;0x2cuy;0x22uy;0x72uy;0x65uy;0x63uy;0x61uy;
//      0x6cuy;0x63uy;0x75uy;0x6cuy;0x61uy;0x74uy;0x69uy;0x6fuy;0x6euy;0x54uy;0x68uy;
//      0x72uy;0x65uy;0x73uy;0x68uy;0x6fuy;0x6cuy;0x64uy;0x22uy;0x3auy;0x30uy;0x2cuy;
//      0x22uy;0x74uy;0x69uy;0x6duy;0x65uy;0x73uy;0x74uy;0x61uy;0x6duy;0x70uy;0x22uy;
//      0x3auy;0x22uy;0x32uy;0x30uy;0x31uy;0x38uy;0x2duy;0x30uy;0x32uy;0x2duy;0x30uy;
//      0x32uy;0x54uy;0x32uy;0x32uy;0x3auy;0x30uy;0x36uy;0x3auy;0x31uy;0x37uy;0x5auy;
//      0x22uy;0x2cuy;0x22uy;0x72uy;0x65uy;0x61uy;0x73uy;0x6fuy;0x6euy;0x22uy;0x3auy;
//      0x22uy;0x45uy;0x76uy;0x65uy;0x6euy;0x74uy;0x22uy;0x2cuy;0x22uy;0x63uy;0x6fuy;
//      0x6duy;0x6duy;0x65uy;0x6euy;0x74uy;0x22uy;0x3auy;0x22uy;0x52uy;0x65uy;0x70uy;
//      0x72uy;0x69uy;0x63uy;0x65uy;0x20uy;0x61uy;0x6cuy;0x6cuy;0x20uy;0x66uy;0x72uy;
//      0x6fuy;0x6duy;0x20uy;0x4euy;0x6fuy;0x76uy;0x61uy;0x22uy;0x2cuy;0x22uy;0x69uy;
//      0x64uy;0x22uy;0x3auy;0x22uy;0x39uy;0x39uy;0x38uy;0x32uy;0x64uy;0x35uy;0x32uy;
//      0x33uy;0x37uy;0x38uy;0x39uy;0x33uy;0x34uy;0x65uy;0x64uy;0x63uy;0x61uy;0x32uy;
//      0x33uy;0x62uy;0x33uy;0x30uy;0x37uy;0x34uy;0x63uy;0x62uy;0x33uy;0x64uy;0x30uy;
//      0x30uy;0x61uy;0x39uy;0x22uy;0x7duy;0x00uy;0x82uy;0x04uy;0x00uy;0x00uy;0x62uy;
//      0x46uy;0x31uy;0x31uy;0x30uy;0x30uy;0x30uy;0x64uy;0x37uy;0x34uy;0x30uy;0x36uy;
//      0x34uy;0x64uy;0x34uy;0x35uy;0x64uy;0x63uy;0x62uy;0x63uy;0x32uy;0x63uy;0x31uy;
//      0x36uy;0x35uy;0x62uy;0x31uy;0x66uy;0x34uy;0x39uy;0x32uy;0x62uy;0x39uy;0x31uy;
//      0x3auy;0x32uy;0x31uy;0xaeuy;0x03uy;0x7buy;0x22uy;0x72uy;0x65uy;0x74uy;0x61uy;
//      0x69uy;0x6cuy;0x53uy;0x6buy;0x75uy;0x49uy;0x64uy;0x22uy;0x3auy;0x22uy;0x31uy;
//      0x31uy;0x30uy;0x30uy;0x30uy;0x64uy;0x37uy;0x34uy;0x30uy;0x36uy;0x34uy;0x64uy;
//      0x34uy;0x35uy;0x64uy;0x63uy;0x62uy;0x63uy;0x32uy;0x63uy;0x31uy;0x36uy;0x35uy;
//      0x62uy;0x31uy;0x66uy;0x34uy;0x39uy;0x32uy;0x62uy;0x39uy;0x31uy;0x22uy;0x2cuy;
//      0x22uy;0x6duy;0x61uy;0x72uy;0x74uy;0x49uy;0x64uy;0x22uy;0x3auy;0x32uy;0x31uy;
//      0x2cuy;0x22uy;0x72uy;0x65uy;0x63uy;0x61uy;0x6cuy;0x63uy;0x75uy;0x6cuy;0x61uy;
//      0x74uy;0x69uy;0x6fuy;0x6euy;0x54uy;0x68uy;0x72uy;0x65uy;0x73uy;0x68uy;0x6fuy;
//      0x6cuy;0x64uy;0x22uy;0x3auy;0x30uy;0x2cuy;0x22uy;0x74uy;0x69uy;0x6duy;0x65uy;
//      0x73uy;0x74uy;0x61uy;0x6duy;0x70uy;0x22uy;0x3auy;0x22uy;0x32uy;0x30uy;0x31uy;
//      0x38uy;0x2duy;0x30uy;0x32uy;0x2duy;0x30uy;0x32uy;0x54uy;0x32uy;0x32uy;0x3auy;
//      0x30uy;0x36uy;0x3auy;0x31uy;0x37uy;0x5auy;0x22uy;0x2cuy;0x22uy;0x72uy;0x65uy;
//      0x61uy;0x73uy;0x6fuy;0x6euy;0x22uy;0x3auy;0x22uy;0x45uy;0x76uy;0x65uy;0x6euy;
//      0x74uy;0x22uy;0x2cuy;0x22uy;0x63uy;0x6fuy;0x6duy;0x6duy;0x65uy;0x6euy;0x74uy;
//      0x22uy;0x3auy;0x22uy;0x52uy;0x65uy;0x70uy;0x72uy;0x69uy;0x63uy;0x65uy;0x20uy;
//      0x61uy;0x6cuy;0x6cuy;0x20uy;0x66uy;0x72uy;0x6fuy;0x6duy;0x20uy;0x4euy;0x6fuy;
//      0x76uy;0x61uy;0x22uy;0x2cuy;0x22uy;0x69uy;0x64uy;0x22uy;0x3auy;0x22uy;0x63uy;
//      0x38uy;0x35uy;0x33uy;0x62uy;0x35uy;0x30uy;0x38uy;0x30uy;0x39uy;0x66uy;0x32uy;
//      0x34uy;0x34uy;0x62uy;0x39uy;0x61uy;0x37uy;0x35uy;0x61uy;0x35uy;0x63uy;0x63uy;
//      0x62uy;0x64uy;0x37uy;0x37uy;0x36uy;0x39uy;0x36uy;0x61uy;0x66uy;0x22uy;0x7duy;
//      0x00uy;0x82uy;0x04uy;0x00uy;0x00uy;0x64uy;0x46uy;0x31uy;0x31uy;0x34uy;0x30uy;
//      0x30uy;0x34uy;0x62uy;0x39uy;0x36uy;0x37uy;0x31uy;0x33uy;0x34uy;0x30uy;0x63uy;
//      0x31uy;0x39uy;0x38uy;0x63uy;0x63uy;0x39uy;0x30uy;0x34uy;0x30uy;0x62uy;0x38uy;
//      0x63uy;0x39uy;0x32uy;0x62uy;0x36uy;0x65uy;0x3auy;0x32uy;0x31uy;0xaeuy;0x03uy;
//      0x7buy;0x22uy;0x72uy;0x65uy;0x74uy;0x61uy;0x69uy;0x6cuy;0x53uy;0x6buy;0x75uy;
//      0x49uy;0x64uy;0x22uy;0x3auy;0x22uy;0x31uy;0x31uy;0x34uy;0x30uy;0x30uy;0x34uy;
//      0x62uy;0x39uy;0x36uy;0x37uy;0x31uy;0x33uy;0x34uy;0x30uy;0x63uy;0x31uy;0x39uy;
//      0x38uy;0x63uy;0x63uy;0x39uy;0x30uy;0x34uy;0x30uy;0x62uy;0x38uy;0x63uy;0x39uy;
//      0x32uy;0x62uy;0x36uy;0x65uy;0x22uy;0x2cuy;0x22uy;0x6duy;0x61uy;0x72uy;0x74uy;
//      0x49uy;0x64uy;0x22uy;0x3auy;0x32uy;0x31uy;0x2cuy;0x22uy;0x72uy;0x65uy;0x63uy;
//      0x61uy;0x6cuy;0x63uy;0x75uy;0x6cuy;0x61uy;0x74uy;0x69uy;0x6fuy;0x6euy;0x54uy;
//      0x68uy;0x72uy;0x65uy;0x73uy;0x68uy;0x6fuy;0x6cuy;0x64uy;0x22uy;0x3auy;0x30uy;
//      0x2cuy;0x22uy;0x74uy;0x69uy;0x6duy;0x65uy;0x73uy;0x74uy;0x61uy;0x6duy;0x70uy;
//      0x22uy;0x3auy;0x22uy;0x32uy;0x30uy;0x31uy;0x38uy;0x2duy;0x30uy;0x32uy;0x2duy;
//      0x30uy;0x32uy;0x54uy;0x32uy;0x32uy;0x3auy;0x30uy;0x36uy;0x3auy;0x31uy;0x37uy;
//      0x5auy;0x22uy;0x2cuy;0x22uy;0x72uy;0x65uy;0x61uy;0x73uy;0x6fuy;0x6euy;0x22uy;
//      0x3auy;0x22uy;0x45uy;0x76uy;0x65uy;0x6euy;0x74uy;0x22uy;0x2cuy;0x22uy;0x63uy;
//      0x6fuy;0x6duy;0x6duy;0x65uy;0x6euy;0x74uy;0x22uy;0x3auy;0x22uy;0x52uy;0x65uy;
//      0x70uy;0x72uy;0x69uy;0x63uy;0x65uy;0x20uy;0x61uy;0x6cuy;0x6cuy;0x20uy;0x66uy;
//      0x72uy;0x6fuy;0x6duy;0x20uy;0x4euy;0x6fuy;0x76uy;0x61uy;0x22uy;0x2cuy;0x22uy;
//      0x69uy;0x64uy;0x22uy;0x3auy;0x22uy;0x34uy;0x31uy;0x39uy;0x65uy;0x30uy;0x39uy;
//      0x30uy;0x64uy;0x64uy;0x36uy;0x63uy;0x66uy;0x34uy;0x32uy;0x34uy;0x64uy;0x39uy;
//      0x63uy;0x66uy;0x65uy;0x30uy;0x33uy;0x31uy;0x66uy;0x33uy;0x64uy;0x62uy;0x39uy;
//      0x65uy;0x65uy;0x32uy;0x64uy;0x22uy;0x7duy;0x00uy;0x82uy;0x04uy;0x00uy;0x00uy;
//      0x66uy;0x46uy;0x31uy;0x65uy;0x64uy;0x30uy;0x30uy;0x35uy;0x35uy;0x63uy;0x34uy;
//      0x30uy;0x30uy;0x37uy;0x34uy;0x36uy;0x35uy;0x64uy;0x61uy;0x36uy;0x35uy;0x37uy;
//      0x36uy;0x37uy;0x37uy;0x64uy;0x39uy;0x39uy;0x62uy;0x37uy;0x32uy;0x32uy;0x65uy;
//      0x33uy;0x3auy;0x32uy;0x31uy;0xaeuy;0x03uy;0x7buy;0x22uy;0x72uy;0x65uy;0x74uy;
//      0x61uy;0x69uy;0x6cuy;0x53uy;0x6buy;0x75uy;0x49uy;0x64uy;0x22uy;0x3auy;0x22uy;
//      0x31uy;0x65uy;0x64uy;0x30uy;0x30uy;0x35uy;0x35uy;0x63uy;0x34uy;0x30uy;0x30uy;
//      0x37uy;0x34uy;0x36uy;0x35uy;0x64uy;0x61uy;0x36uy;0x35uy;0x37uy;0x36uy;0x37uy;
//      0x37uy;0x64uy;0x39uy;0x39uy;0x62uy;0x37uy;0x32uy;0x32uy;0x65uy;0x33uy;0x22uy;
//      0x2cuy;0x22uy;0x6duy;0x61uy;0x72uy;0x74uy;0x49uy;0x64uy;0x22uy;0x3auy;0x32uy;
//      0x31uy;0x2cuy;0x22uy;0x72uy;0x65uy;0x63uy;0x61uy;0x6cuy;0x63uy;0x75uy;0x6cuy;
//      0x61uy;0x74uy;0x69uy;0x6fuy;0x6euy;0x54uy;0x68uy;0x72uy;0x65uy;0x73uy;0x68uy;
//      0x6fuy;0x6cuy;0x64uy;0x22uy;0x3auy;0x30uy;0x2cuy;0x22uy;0x74uy;0x69uy;0x6duy;
//      0x65uy;0x73uy;0x74uy;0x61uy;0x6duy;0x70uy;0x22uy;0x3auy;0x22uy;0x32uy;0x30uy;
//      0x31uy;0x38uy;0x2duy;0x30uy;0x32uy;0x2duy;0x30uy;0x32uy;0x54uy;0x32uy;0x32uy;
//      0x3auy;0x30uy;0x36uy;0x3auy;0x31uy;0x37uy;0x5auy;0x22uy;0x2cuy;0x22uy;0x72uy;
//      0x65uy;0x61uy;0x73uy;0x6fuy;0x6euy;0x22uy;0x3auy;0x22uy;0x45uy;0x76uy;0x65uy;
//      0x6euy;0x74uy;0x22uy;0x2cuy;0x22uy;0x63uy;0x6fuy;0x6duy;0x6duy;0x65uy;0x6euy;
//      0x74uy;0x22uy;0x3auy;0x22uy;0x52uy;0x65uy;0x70uy;0x72uy;0x69uy;0x63uy;0x65uy;
//      0x20uy;0x61uy;0x6cuy;0x6cuy;0x20uy;0x66uy;0x72uy;0x6fuy;0x6duy;0x20uy;0x4euy;
//      0x6fuy;0x76uy;0x61uy;0x22uy;0x2cuy;0x22uy;0x69uy;0x64uy;0x22uy;0x3auy;0x22uy;
//      0x32uy;0x38uy;0x64uy;0x32uy;0x33uy;0x35uy;0x39uy;0x30uy;0x61uy;0x38uy;0x32uy;
//      0x37uy;0x34uy;0x32uy;0x38uy;0x36uy;0x38uy;0x33uy;0x39uy;0x39uy;0x30uy;0x35uy;
//      0x39uy;0x63uy;0x35uy;0x39uy;0x61uy;0x64uy;0x34uy;0x30uy;0x63uy;0x34uy;0x22uy;
//      0x7duy;0x00uy;0x82uy;0x04uy;0x00uy;0x00uy;0x68uy;0x46uy;0x31uy;0x65uy;0x35uy;
//      0x30uy;0x30uy;0x39uy;0x64uy;0x37uy;0x62uy;0x65uy;0x37uy;0x38uy;0x34uy;0x36uy;
//      0x61uy;0x31uy;0x62uy;0x66uy;0x32uy;0x36uy;0x38uy;0x66uy;0x33uy;0x63uy;0x31uy;
//      0x35uy;0x61uy;0x62uy;0x36uy;0x63uy;0x65uy;0x66uy;0x3auy;0x32uy;0x31uy;0xaeuy;
//      0x03uy;0x7buy;0x22uy;0x72uy;0x65uy;0x74uy;0x61uy;0x69uy;0x6cuy;0x53uy;0x6buy;
//      0x75uy;0x49uy;0x64uy;0x22uy;0x3auy;0x22uy;0x31uy;0x65uy;0x35uy;0x30uy;0x30uy;
//      0x39uy;0x64uy;0x37uy;0x62uy;0x65uy;0x37uy;0x38uy;0x34uy;0x36uy;0x61uy;0x31uy;
//      0x62uy;0x66uy;0x32uy;0x36uy;0x38uy;0x66uy;0x33uy;0x63uy;0x31uy;0x35uy;0x61uy;
//      0x62uy;0x36uy;0x63uy;0x65uy;0x66uy;0x22uy;0x2cuy;0x22uy;0x6duy;0x61uy;0x72uy;
//      0x74uy;0x49uy;0x64uy;0x22uy;0x3auy;0x32uy;0x31uy;0x2cuy;0x22uy;0x72uy;0x65uy;
//      0x63uy;0x61uy;0x6cuy;0x63uy;0x75uy;0x6cuy;0x61uy;0x74uy;0x69uy;0x6fuy;0x6euy;
//      0x54uy;0x68uy;0x72uy;0x65uy;0x73uy;0x68uy;0x6fuy;0x6cuy;0x64uy;0x22uy;0x3auy;
//      0x30uy;0x2cuy;0x22uy;0x74uy;0x69uy;0x6duy;0x65uy;0x73uy;0x74uy;0x61uy;0x6duy;
//      0x70uy;0x22uy;0x3auy;0x22uy;0x32uy;0x30uy;0x31uy;0x38uy;0x2duy;0x30uy;0x32uy;
//      0x2duy;0x30uy;0x32uy;0x54uy;0x32uy;0x32uy;0x3auy;0x30uy;0x36uy;0x3auy;0x31uy;
//      0x37uy;0x5auy;0x22uy;0x2cuy;0x22uy;0x72uy;0x65uy;0x61uy;0x73uy;0x6fuy;0x6euy;
//      0x22uy;0x3auy;0x22uy;0x45uy;0x76uy;0x65uy;0x6euy;0x74uy;0x22uy;0x2cuy;0x22uy;
//      0x63uy;0x6fuy;0x6duy;0x6duy;0x65uy;0x6euy;0x74uy;0x22uy;0x3auy;0x22uy;0x52uy;
//      0x65uy;0x70uy;0x72uy;0x69uy;0x63uy;0x65uy;0x20uy;0x61uy;0x6cuy;0x6cuy;0x20uy;
//      0x66uy;0x72uy;0x6fuy;0x6duy;0x20uy;0x4euy;0x6fuy;0x76uy;0x61uy;0x22uy;0x2cuy;
//      0x22uy;0x69uy;0x64uy;0x22uy;0x3auy;0x22uy;0x39uy;0x36uy;0x65uy;0x39uy;0x34uy;
//      0x62uy;0x32uy;0x37uy;0x63uy;0x32uy;0x65uy;0x64uy;0x34uy;0x61uy;0x32uy;0x33uy;
//      0x39uy;0x34uy;0x39uy;0x36uy;0x36uy;0x63uy;0x63uy;0x63uy;0x64uy;0x66uy;0x63uy;
//      0x65uy;0x36uy;0x65uy;0x38uy;0x37uy;0x22uy;0x7duy;0x00uy;0x82uy;0x04uy;0x00uy;
//      0x00uy;0x6auy;0x46uy;0x31uy;0x64uy;0x65uy;0x30uy;0x30uy;0x61uy;0x34uy;0x65uy;
//      0x65uy;0x66uy;0x30uy;0x31uy;0x34uy;0x35uy;0x32uy;0x36uy;0x39uy;0x39uy;0x30uy;
//      0x65uy;0x32uy;0x65uy;0x35uy;0x63uy;0x62uy;0x64uy;0x31uy;0x63uy;0x33uy;0x37uy;
//      0x36uy;0x34uy;0x3auy;0x32uy;0x31uy;0xaeuy;0x03uy;0x7buy;0x22uy;0x72uy;0x65uy;
//      0x74uy;0x61uy;0x69uy;0x6cuy;0x53uy;0x6buy;0x75uy;0x49uy;0x64uy;0x22uy;0x3auy;
//      0x22uy;0x31uy;0x64uy;0x65uy;0x30uy;0x30uy;0x61uy;0x34uy;0x65uy;0x65uy;0x66uy;
//      0x30uy;0x31uy;0x34uy;0x35uy;0x32uy;0x36uy;0x39uy;0x39uy;0x30uy;0x65uy;0x32uy;
//      0x65uy;0x35uy;0x63uy;0x62uy;0x64uy;0x31uy;0x63uy;0x33uy;0x37uy;0x36uy;0x34uy;
//      0x22uy;0x2cuy;0x22uy;0x6duy;0x61uy;0x72uy;0x74uy;0x49uy;0x64uy;0x22uy;0x3auy;
//      0x32uy;0x31uy;0x2cuy;0x22uy;0x72uy;0x65uy;0x63uy;0x61uy;0x6cuy;0x63uy;0x75uy;
//      0x6cuy;0x61uy;0x74uy;0x69uy;0x6fuy;0x6euy;0x54uy;0x68uy;0x72uy;0x65uy;0x73uy;
//      0x68uy;0x6fuy;0x6cuy;0x64uy;0x22uy;0x3auy;0x30uy;0x2cuy;0x22uy;0x74uy;0x69uy;
//      0x6duy;0x65uy;0x73uy;0x74uy;0x61uy;0x6duy;0x70uy;0x22uy;0x3auy;0x22uy;0x32uy;
//      0x30uy;0x31uy;0x38uy;0x2duy;0x30uy;0x32uy;0x2duy;0x30uy;0x32uy;0x54uy;0x32uy;
//      0x32uy;0x3auy;0x30uy;0x36uy;0x3auy;0x31uy;0x37uy;0x5auy;0x22uy;0x2cuy;0x22uy;
//      0x72uy;0x65uy;0x61uy;0x73uy;0x6fuy;0x6euy;0x22uy;0x3auy;0x22uy;0x45uy;0x76uy;
//      0x65uy;0x6euy;0x74uy;0x22uy;0x2cuy;0x22uy;0x63uy;0x6fuy;0x6duy;0x6duy;0x65uy;
//      0x6euy;0x74uy;0x22uy;0x3auy;0x22uy;0x52uy;0x65uy;0x70uy;0x72uy;0x69uy;0x63uy;
//      0x65uy;0x20uy;0x61uy;0x6cuy;0x6cuy;0x20uy;0x66uy;0x72uy;0x6fuy;0x6duy;0x20uy;
//      0x4euy;0x6fuy;0x76uy;0x61uy;0x22uy;0x2cuy;0x22uy;0x69uy;0x64uy;0x22uy;0x3auy;
//      0x22uy;0x61uy;0x33uy;0x38uy;0x65uy;0x30uy;0x65uy;0x35uy;0x36uy;0x38uy;0x63uy;
//      0x31uy;0x32uy;0x34uy;0x37uy;0x63uy;0x66uy;0x38uy;0x61uy;0x65uy;0x36uy;0x32uy;
//      0x33uy;0x62uy;0x32uy;0x31uy;0x34uy;0x36uy;0x38uy;0x33uy;0x34uy;0x61uy;0x65uy;
//      0x22uy;0x7duy;0x00uy;0x82uy;0x04uy;0x00uy;0x00uy;0x6cuy;0x46uy;0x31uy;0x63uy;
//      0x63uy;0x30uy;0x31uy;0x38uy;0x62uy;0x38uy;0x63uy;0x30uy;0x31uy;0x62uy;0x34uy;
//      0x30uy;0x62uy;0x36uy;0x39uy;0x32uy;0x34uy;0x32uy;0x33uy;0x38uy;0x65uy;0x32uy;
//      0x65uy;0x35uy;0x66uy;0x32uy;0x35uy;0x34uy;0x31uy;0x39uy;0x3auy;0x32uy;0x31uy;
//      0xaeuy;0x03uy;0x7buy;0x22uy;0x72uy;0x65uy;0x74uy;0x61uy;0x69uy;0x6cuy;0x53uy;
//      0x6buy;0x75uy;0x49uy;0x64uy;0x22uy;0x3auy;0x22uy;0x31uy;0x63uy;0x63uy;0x30uy;
//      0x31uy;0x38uy;0x62uy;0x38uy;0x63uy;0x30uy;0x31uy;0x62uy;0x34uy;0x30uy;0x62uy;
//      0x36uy;0x39uy;0x32uy;0x34uy;0x32uy;0x33uy;0x38uy;0x65uy;0x32uy;0x65uy;0x35uy;
//      0x66uy;0x32uy;0x35uy;0x34uy;0x31uy;0x39uy;0x22uy;0x2cuy;0x22uy;0x6duy;0x61uy;
//      0x72uy;0x74uy;0x49uy;0x64uy;0x22uy;0x3auy;0x32uy;0x31uy;0x2cuy;0x22uy;0x72uy;
//      0x65uy;0x63uy;0x61uy;0x6cuy;0x63uy;0x75uy;0x6cuy;0x61uy;0x74uy;0x69uy;0x6fuy;
//      0x6euy;0x54uy;0x68uy;0x72uy;0x65uy;0x73uy;0x68uy;0x6fuy;0x6cuy;0x64uy;0x22uy;
//      0x3auy;0x30uy;0x2cuy;0x22uy;0x74uy;0x69uy;0x6duy;0x65uy;0x73uy;0x74uy;0x61uy;
//      0x6duy;0x70uy;0x22uy;0x3auy;0x22uy;0x32uy;0x30uy;0x31uy;0x38uy;0x2duy;0x30uy;
//      0x32uy;0x2duy;0x30uy;0x32uy;0x54uy;0x32uy;0x32uy;0x3auy;0x30uy;0x36uy;0x3auy;
//      0x31uy;0x37uy;0x5auy;0x22uy;0x2cuy;0x22uy;0x72uy;0x65uy;0x61uy;0x73uy;0x6fuy;
//      0x6euy;0x22uy;0x3auy;0x22uy;0x45uy;0x76uy;0x65uy;0x6euy;0x74uy;0x22uy;0x2cuy;
//      0x22uy;0x63uy;0x6fuy;0x6duy;0x6duy;0x65uy;0x6euy;0x74uy;0x22uy;0x3auy;0x22uy;
//      0x52uy;0x65uy;0x70uy;0x72uy;0x69uy;0x63uy;0x65uy;0x20uy;0x61uy;0x6cuy;0x6cuy;
//      0x20uy;0x66uy;0x72uy;0x6fuy;0x6duy;0x20uy;0x4euy;0x6fuy;0x76uy;0x61uy;0x22uy;
//      0x2cuy;0x22uy;0x69uy;0x64uy;0x22uy;0x3auy;0x22uy;0x34uy;0x37uy;0x64uy;0x35uy;
//      0x30uy;0x34uy;0x66uy;0x66uy;0x32uy;0x38uy;0x62uy;0x64uy;0x34uy;0x38uy;0x30uy;
//      0x31uy;0x38uy;0x64uy;0x35uy;0x64uy;0x36uy;0x35uy;0x32uy;0x32uy;0x36uy;0x32uy;
//      0x66uy;0x37uy;0x30uy;0x31uy;0x33uy;0x38uy;0x22uy;0x7duy;0x00uy;0x82uy;0x04uy;
//      0x00uy;0x00uy;0x6euy;0x46uy;0x31uy;0x62uy;0x63uy;0x30uy;0x30uy;0x37uy;0x64uy;
//      0x61uy;0x36uy;0x36uy;0x30uy;0x35uy;0x34uy;0x35uy;0x65uy;0x34uy;0x62uy;0x61uy;
//      0x31uy;0x32uy;0x34uy;0x39uy;0x61uy;0x38uy;0x37uy;0x38uy;0x36uy;0x30uy;0x32uy;
//      0x61uy;0x38uy;0x64uy;0x3auy;0x32uy;0x31uy;0xaeuy;0x03uy;0x7buy;0x22uy;0x72uy;
//      0x65uy;0x74uy;0x61uy;0x69uy;0x6cuy;0x53uy;0x6buy;0x75uy;0x49uy;0x64uy;0x22uy;
//      0x3auy;0x22uy;0x31uy;0x62uy;0x63uy;0x30uy;0x30uy;0x37uy;0x64uy;0x61uy;0x36uy;
//      0x36uy;0x30uy;0x35uy;0x34uy;0x35uy;0x65uy;0x34uy;0x62uy;0x61uy;0x31uy;0x32uy;
//      0x34uy;0x39uy;0x61uy;0x38uy;0x37uy;0x38uy;0x36uy;0x30uy;0x32uy;0x61uy;0x38uy;
//      0x64uy;0x22uy;0x2cuy;0x22uy;0x6duy;0x61uy;0x72uy;0x74uy;0x49uy;0x64uy;0x22uy;
//      0x3auy;0x32uy;0x31uy;0x2cuy;0x22uy;0x72uy;0x65uy;0x63uy;0x61uy;0x6cuy;0x63uy;
//      0x75uy;0x6cuy;0x61uy;0x74uy;0x69uy;0x6fuy;0x6euy;0x54uy;0x68uy;0x72uy;0x65uy;
//      0x73uy;0x68uy;0x6fuy;0x6cuy;0x64uy;0x22uy;0x3auy;0x30uy;0x2cuy;0x22uy;0x74uy;
//      0x69uy;0x6duy;0x65uy;0x73uy;0x74uy;0x61uy;0x6duy;0x70uy;0x22uy;0x3auy;0x22uy;
//      0x32uy;0x30uy;0x31uy;0x38uy;0x2duy;0x30uy;0x32uy;0x2duy;0x30uy;0x32uy;0x54uy;
//      0x32uy;0x32uy;0x3auy;0x30uy;0x36uy;0x3auy;0x31uy;0x37uy;0x5auy;0x22uy;0x2cuy;
//      0x22uy;0x72uy;0x65uy;0x61uy;0x73uy;0x6fuy;0x6euy;0x22uy;0x3auy;0x22uy;0x45uy;
//      0x76uy;0x65uy;0x6euy;0x74uy;0x22uy;0x2cuy;0x22uy;0x63uy;0x6fuy;0x6duy;0x6duy;
//      0x65uy;0x6euy;0x74uy;0x22uy;0x3auy;0x22uy;0x52uy;0x65uy;0x70uy;0x72uy;0x69uy;
//      0x63uy;0x65uy;0x20uy;0x61uy;0x6cuy;0x6cuy;0x20uy;0x66uy;0x72uy;0x6fuy;0x6duy;
//      0x20uy;0x4euy;0x6fuy;0x76uy;0x61uy;0x22uy;0x2cuy;0x22uy;0x69uy;0x64uy;0x22uy;
//      0x3auy;0x22uy;0x33uy;0x31uy;0x37uy;0x64uy;0x35uy;0x64uy;0x64uy;0x35uy;0x37uy;
//      0x30uy;0x33uy;0x65uy;0x34uy;0x32uy;0x31uy;0x65uy;0x39uy;0x33uy;0x63uy;0x61uy;
//      0x63uy;0x30uy;0x37uy;0x65uy;0x63uy;0x33uy;0x33uy;0x34uy;0x38uy;0x62uy;0x38uy;
//      0x64uy;0x22uy;0x7duy;0x00uy;0x82uy;0x04uy;0x00uy;0x00uy;0x70uy;0x46uy;0x31uy;
//      0x61uy;0x39uy;0x30uy;0x30uy;0x34uy;0x31uy;0x32uy;0x37uy;0x30uy;0x30uy;0x36uy;
//      0x34uy;0x38uy;0x61uy;0x30uy;0x61uy;0x65uy;0x31uy;0x39uy;0x34uy;0x65uy;0x66uy;
//      0x66uy;0x61uy;0x33uy;0x63uy;0x32uy;0x35uy;0x34uy;0x64uy;0x65uy;0x3auy;0x32uy;
//      0x31uy;0xaeuy;0x03uy;0x7buy;0x22uy;0x72uy;0x65uy;0x74uy;0x61uy;0x69uy;0x6cuy;
//      0x53uy;0x6buy;0x75uy;0x49uy;0x64uy;0x22uy;0x3auy;0x22uy;0x31uy;0x61uy;0x39uy;
//      0x30uy;0x30uy;0x34uy;0x31uy;0x32uy;0x37uy;0x30uy;0x30uy;0x36uy;0x34uy;0x38uy;
//      0x61uy;0x30uy;0x61uy;0x65uy;0x31uy;0x39uy;0x34uy;0x65uy;0x66uy;0x66uy;0x61uy;
//      0x33uy;0x63uy;0x32uy;0x35uy;0x34uy;0x64uy;0x65uy;0x22uy;0x2cuy;0x22uy;0x6duy;
//      0x61uy;0x72uy;0x74uy;0x49uy;0x64uy;0x22uy;0x3auy;0x32uy;0x31uy;0x2cuy;0x22uy;
//      0x72uy;0x65uy;0x63uy;0x61uy;0x6cuy;0x63uy;0x75uy;0x6cuy;0x61uy;0x74uy;0x69uy;
//      0x6fuy;0x6euy;0x54uy;0x68uy;0x72uy;0x65uy;0x73uy;0x68uy;0x6fuy;0x6cuy;0x64uy;
//      0x22uy;0x3auy;0x30uy;0x2cuy;0x22uy;0x74uy;0x69uy;0x6duy;0x65uy;0x73uy;0x74uy;
//      0x61uy;0x6duy;0x70uy;0x22uy;0x3auy;0x22uy;0x32uy;0x30uy;0x31uy;0x38uy;0x2duy;
//      0x30uy;0x32uy;0x2duy;0x30uy;0x32uy;0x54uy;0x32uy;0x32uy;0x3auy;0x30uy;0x36uy;
//      0x3auy;0x31uy;0x37uy;0x5auy;0x22uy;0x2cuy;0x22uy;0x72uy;0x65uy;0x61uy;0x73uy;
//      0x6fuy;0x6euy;0x22uy;0x3auy;0x22uy;0x45uy;0x76uy;0x65uy;0x6euy;0x74uy;0x22uy;
//      0x2cuy;0x22uy;0x63uy;0x6fuy;0x6duy;0x6duy;0x65uy;0x6euy;0x74uy;0x22uy;0x3auy;
//      0x22uy;0x52uy;0x65uy;0x70uy;0x72uy;0x69uy;0x63uy;0x65uy;0x20uy;0x61uy;0x6cuy;
//      0x6cuy;0x20uy;0x66uy;0x72uy;0x6fuy;0x6duy;0x20uy;0x4euy;0x6fuy;0x76uy;0x61uy;
//      0x22uy;0x2cuy;0x22uy;0x69uy;0x64uy;0x22uy;0x3auy;0x22uy;0x36uy;0x31uy;0x33uy;
//      0x34uy;0x37uy;0x36uy;0x35uy;0x64uy;0x34uy;0x33uy;0x34uy;0x62uy;0x34uy;0x62uy;
//      0x30uy;0x64uy;0x39uy;0x39uy;0x39uy;0x31uy;0x64uy;0x61uy;0x61uy;0x65uy;0x65uy;
//      0x36uy;0x31uy;0x65uy;0x61uy;0x31uy;0x32uy;0x66uy;0x22uy;0x7duy;0x00uy;0x82uy;
//      0x04uy;0x00uy;0x00uy;0x72uy;0x46uy;0x31uy;0x39uy;0x66uy;0x30uy;0x30uy;0x36uy;
//      0x39uy;0x34uy;0x35uy;0x36uy;0x38uy;0x34uy;0x34uy;0x32uy;0x31uy;0x36uy;0x39uy;
//      0x61uy;0x62uy;0x33uy;0x64uy;0x35uy;0x62uy;0x34uy;0x62uy;0x36uy;0x63uy;0x33uy;
//      0x62uy;0x64uy;0x38uy;0x30uy;0x3auy;0x32uy;0x31uy;0xaeuy;0x03uy;0x7buy;0x22uy;
//      0x72uy;0x65uy;0x74uy;0x61uy;0x69uy;0x6cuy;0x53uy;0x6buy;0x75uy;0x49uy;0x64uy;
//      0x22uy;0x3auy;0x22uy;0x31uy;0x39uy;0x66uy;0x30uy;0x30uy;0x36uy;0x39uy;0x34uy;
//      0x35uy;0x36uy;0x38uy;0x34uy;0x34uy;0x32uy;0x31uy;0x36uy;0x39uy;0x61uy;0x62uy;
//      0x33uy;0x64uy;0x35uy;0x62uy;0x34uy;0x62uy;0x36uy;0x63uy;0x33uy;0x62uy;0x64uy;
//      0x38uy;0x30uy;0x22uy;0x2cuy;0x22uy;0x6duy;0x61uy;0x72uy;0x74uy;0x49uy;0x64uy;
//      0x22uy;0x3auy;0x32uy;0x31uy;0x2cuy;0x22uy;0x72uy;0x65uy;0x63uy;0x61uy;0x6cuy;
//      0x63uy;0x75uy;0x6cuy;0x61uy;0x74uy;0x69uy;0x6fuy;0x6euy;0x54uy;0x68uy;0x72uy;
//      0x65uy;0x73uy;0x68uy;0x6fuy;0x6cuy;0x64uy;0x22uy;0x3auy;0x30uy;0x2cuy;0x22uy;
//      0x74uy;0x69uy;0x6duy;0x65uy;0x73uy;0x74uy;0x61uy;0x6duy;0x70uy;0x22uy;0x3auy;
//      0x22uy;0x32uy;0x30uy;0x31uy;0x38uy;0x2duy;0x30uy;0x32uy;0x2duy;0x30uy;0x32uy;
//      0x54uy;0x32uy;0x32uy;0x3auy;0x30uy;0x36uy;0x3auy;0x31uy;0x37uy;0x5auy;0x22uy;
//      0x2cuy;0x22uy;0x72uy;0x65uy;0x61uy;0x73uy;0x6fuy;0x6euy;0x22uy;0x3auy;0x22uy;
//      0x45uy;0x76uy;0x65uy;0x6euy;0x74uy;0x22uy;0x2cuy;0x22uy;0x63uy;0x6fuy;0x6duy;
//      0x6duy;0x65uy;0x6euy;0x74uy;0x22uy;0x3auy;0x22uy;0x52uy;0x65uy;0x70uy;0x72uy;
//      0x69uy;0x63uy;0x65uy;0x20uy;0x61uy;0x6cuy;0x6cuy;0x20uy;0x66uy;0x72uy;0x6fuy;
//      0x6duy;0x20uy;0x4euy;0x6fuy;0x76uy;0x61uy;0x22uy;0x2cuy;0x22uy;0x69uy;0x64uy;
//      0x22uy;0x3auy;0x22uy;0x63uy;0x39uy;0x65uy;0x32uy;0x32uy;0x63uy;0x31uy;0x30uy;
//      0x35uy;0x66uy;0x61uy;0x39uy;0x34uy;0x38uy;0x63uy;0x36uy;0x39uy;0x62uy;0x63uy;
//      0x35uy;0x30uy;0x63uy;0x36uy;0x63uy;0x61uy;0x31uy;0x31uy;0x30uy;0x31uy;0x62uy;
//      0x65uy;0x32uy;0x22uy;0x7duy;0x00uy;0x82uy;0x04uy;0x00uy;0x00uy;0x74uy;0x46uy;
//      0x31uy;0x38uy;0x62uy;0x30uy;0x30uy;0x36uy;0x34uy;0x66uy;0x61uy;0x64uy;0x61uy;
//      0x37uy;0x34uy;0x35uy;0x32uy;0x64uy;0x61uy;0x34uy;0x62uy;0x61uy;0x31uy;0x66uy;
//      0x61uy;0x35uy;0x39uy;0x39uy;0x62uy;0x61uy;0x32uy;0x31uy;0x62uy;0x37uy;0x3auy;
//      0x32uy;0x31uy;0xaeuy;0x03uy;0x7buy;0x22uy;0x72uy;0x65uy;0x74uy;0x61uy;0x69uy;
//      0x6cuy;0x53uy;0x6buy;0x75uy;0x49uy;0x64uy;0x22uy;0x3auy;0x22uy;0x31uy;0x38uy;
//      0x62uy;0x30uy;0x30uy;0x36uy;0x34uy;0x66uy;0x61uy;0x64uy;0x61uy;0x37uy;0x34uy;
//      0x35uy;0x32uy;0x64uy;0x61uy;0x34uy;0x62uy;0x61uy;0x31uy;0x66uy;0x61uy;0x35uy;
//      0x39uy;0x39uy;0x62uy;0x61uy;0x32uy;0x31uy;0x62uy;0x37uy;0x22uy;0x2cuy;0x22uy;
//      0x6duy;0x61uy;0x72uy;0x74uy;0x49uy;0x64uy;0x22uy;0x3auy;0x32uy;0x31uy;0x2cuy;
//      0x22uy;0x72uy;0x65uy;0x63uy;0x61uy;0x6cuy;0x63uy;0x75uy;0x6cuy;0x61uy;0x74uy;
//      0x69uy;0x6fuy;0x6euy;0x54uy;0x68uy;0x72uy;0x65uy;0x73uy;0x68uy;0x6fuy;0x6cuy;
//      0x64uy;0x22uy;0x3auy;0x30uy;0x2cuy;0x22uy;0x74uy;0x69uy;0x6duy;0x65uy;0x73uy;
//      0x74uy;0x61uy;0x6duy;0x70uy;0x22uy;0x3auy;0x22uy;0x32uy;0x30uy;0x31uy;0x38uy;
//      0x2duy;0x30uy;0x32uy;0x2duy;0x30uy;0x32uy;0x54uy;0x32uy;0x32uy;0x3auy;0x30uy;
//      0x36uy;0x3auy;0x31uy;0x37uy;0x5auy;0x22uy;0x2cuy;0x22uy;0x72uy;0x65uy;0x61uy;
//      0x73uy;0x6fuy;0x6euy;0x22uy;0x3auy;0x22uy;0x45uy;0x76uy;0x65uy;0x6euy;0x74uy;
//      0x22uy;0x2cuy;0x22uy;0x63uy;0x6fuy;0x6duy;0x6duy;0x65uy;0x6euy;0x74uy;0x22uy;
//      0x3auy;0x22uy;0x52uy;0x65uy;0x70uy;0x72uy;0x69uy;0x63uy;0x65uy;0x20uy;0x61uy;
//      0x6cuy;0x6cuy;0x20uy;0x66uy;0x72uy;0x6fuy;0x6duy;0x20uy;0x4euy;0x6fuy;0x76uy;
//      0x61uy;0x22uy;0x2cuy;0x22uy;0x69uy;0x64uy;0x22uy;0x3auy;0x22uy;0x31uy;0x35uy;
//      0x63uy;0x33uy;0x35uy;0x64uy;0x30uy;0x66uy;0x30uy;0x63uy;0x65uy;0x64uy;0x34uy;
//      0x38uy;0x38uy;0x61uy;0x38uy;0x66uy;0x62uy;0x38uy;0x62uy;0x37uy;0x35uy;0x33uy;
//      0x64uy;0x62uy;0x32uy;0x35uy;0x34uy;0x30uy;0x35uy;0x38uy;0x22uy;0x7duy;0x00uy;
//      0x82uy;0x04uy;0x00uy;0x00uy;0x76uy;0x46uy;0x31uy;0x37uy;0x64uy;0x30uy;0x31uy;
//      0x61uy;0x36uy;0x63uy;0x36uy;0x35uy;0x37uy;0x38uy;0x34uy;0x64uy;0x66uy;0x38uy;
//      0x38uy;0x34uy;0x34uy;0x62uy;0x62uy;0x39uy;0x36uy;0x38uy;0x62uy;0x37uy;0x34uy;
//      0x37uy;0x37uy;0x64uy;0x61uy;0x62uy;0x3auy;0x32uy;0x31uy;0xaeuy;0x03uy;0x7buy;
//      0x22uy;0x72uy;0x65uy;0x74uy;0x61uy;0x69uy;0x6cuy;0x53uy;0x6buy;0x75uy;0x49uy;
//      0x64uy;0x22uy;0x3auy;0x22uy;0x31uy;0x37uy;0x64uy;0x30uy;0x31uy;0x61uy;0x36uy;
//      0x63uy;0x36uy;0x35uy;0x37uy;0x38uy;0x34uy;0x64uy;0x66uy;0x38uy;0x38uy;0x34uy;
//      0x34uy;0x62uy;0x62uy;0x39uy;0x36uy;0x38uy;0x62uy;0x37uy;0x34uy;0x37uy;0x37uy;
//      0x64uy;0x61uy;0x62uy;0x22uy;0x2cuy;0x22uy;0x6duy;0x61uy;0x72uy;0x74uy;0x49uy;
//      0x64uy;0x22uy;0x3auy;0x32uy;0x31uy;0x2cuy;0x22uy;0x72uy;0x65uy;0x63uy;0x61uy;
//      0x6cuy;0x63uy;0x75uy;0x6cuy;0x61uy;0x74uy;0x69uy;0x6fuy;0x6euy;0x54uy;0x68uy;
//      0x72uy;0x65uy;0x73uy;0x68uy;0x6fuy;0x6cuy;0x64uy;0x22uy;0x3auy;0x30uy;0x2cuy;
//      0x22uy;0x74uy;0x69uy;0x6duy;0x65uy;0x73uy;0x74uy;0x61uy;0x6duy;0x70uy;0x22uy;
//      0x3auy;0x22uy;0x32uy;0x30uy;0x31uy;0x38uy;0x2duy;0x30uy;0x32uy;0x2duy;0x30uy;
//      0x32uy;0x54uy;0x32uy;0x32uy;0x3auy;0x30uy;0x36uy;0x3auy;0x31uy;0x37uy;0x5auy;
//      0x22uy;0x2cuy;0x22uy;0x72uy;0x65uy;0x61uy;0x73uy;0x6fuy;0x6euy;0x22uy;0x3auy;
//      0x22uy;0x45uy;0x76uy;0x65uy;0x6euy;0x74uy;0x22uy;0x2cuy;0x22uy;0x63uy;0x6fuy;
//      0x6duy;0x6duy;0x65uy;0x6euy;0x74uy;0x22uy;0x3auy;0x22uy;0x52uy;0x65uy;0x70uy;
//      0x72uy;0x69uy;0x63uy;0x65uy;0x20uy;0x61uy;0x6cuy;0x6cuy;0x20uy;0x66uy;0x72uy;
//      0x6fuy;0x6duy;0x20uy;0x4euy;0x6fuy;0x76uy;0x61uy;0x22uy;0x2cuy;0x22uy;0x69uy;
//      0x64uy;0x22uy;0x3auy;0x22uy;0x61uy;0x37uy;0x31uy;0x62uy;0x66uy;0x61uy;0x31uy;
//      0x61uy;0x63uy;0x38uy;0x64uy;0x31uy;0x34uy;0x32uy;0x36uy;0x33uy;0x61uy;0x65uy;
//      0x39uy;0x35uy;0x31uy;0x30uy;0x35uy;0x31uy;0x32uy;0x35uy;0x33uy;0x38uy;0x32uy;
//      0x61uy;0x36uy;0x34uy;0x22uy;0x7duy;0x00uy;0x82uy;0x04uy;0x00uy;0x00uy;0x78uy;
//      0x46uy;0x31uy;0x37uy;0x32uy;0x30uy;0x30uy;0x61uy;0x38uy;0x61uy;0x37uy;0x36uy;
//      0x34uy;0x64uy;0x34uy;0x36uy;0x38uy;0x62uy;0x61uy;0x35uy;0x63uy;0x35uy;0x30uy;
//      0x61uy;0x65uy;0x66uy;0x35uy;0x63uy;0x31uy;0x61uy;0x62uy;0x39uy;0x30uy;0x63uy;
//      0x3auy;0x32uy;0x31uy;0xaeuy;0x03uy;0x7buy;0x22uy;0x72uy;0x65uy;0x74uy;0x61uy;
//      0x69uy;0x6cuy;0x53uy;0x6buy;0x75uy;0x49uy;0x64uy;0x22uy;0x3auy;0x22uy;0x31uy;
//      0x37uy;0x32uy;0x30uy;0x30uy;0x61uy;0x38uy;0x61uy;0x37uy;0x36uy;0x34uy;0x64uy;
//      0x34uy;0x36uy;0x38uy;0x62uy;0x61uy;0x35uy;0x63uy;0x35uy;0x30uy;0x61uy;0x65uy;
//      0x66uy;0x35uy;0x63uy;0x31uy;0x61uy;0x62uy;0x39uy;0x30uy;0x63uy;0x22uy;0x2cuy;
//      0x22uy;0x6duy;0x61uy;0x72uy;0x74uy;0x49uy;0x64uy;0x22uy;0x3auy;0x32uy;0x31uy;
//      0x2cuy;0x22uy;0x72uy;0x65uy;0x63uy;0x61uy;0x6cuy;0x63uy;0x75uy;0x6cuy;0x61uy;
//      0x74uy;0x69uy;0x6fuy;0x6euy;0x54uy;0x68uy;0x72uy;0x65uy;0x73uy;0x68uy;0x6fuy;
//      0x6cuy;0x64uy;0x22uy;0x3auy;0x30uy;0x2cuy;0x22uy;0x74uy;0x69uy;0x6duy;0x65uy;
//      0x73uy;0x74uy;0x61uy;0x6duy;0x70uy;0x22uy;0x3auy;0x22uy;0x32uy;0x30uy;0x31uy;
//      0x38uy;0x2duy;0x30uy;0x32uy;0x2duy;0x30uy;0x32uy;0x54uy;0x32uy;0x32uy;0x3auy;
//      0x30uy;0x36uy;0x3auy;0x31uy;0x37uy;0x5auy;0x22uy;0x2cuy;0x22uy;0x72uy;0x65uy;
//      0x61uy;0x73uy;0x6fuy;0x6euy;0x22uy;0x3auy;0x22uy;0x45uy;0x76uy;0x65uy;0x6euy;
//      0x74uy;0x22uy;0x2cuy;0x22uy;0x63uy;0x6fuy;0x6duy;0x6duy;0x65uy;0x6euy;0x74uy;
//      0x22uy;0x3auy;0x22uy;0x52uy;0x65uy;0x70uy;0x72uy;0x69uy;0x63uy;0x65uy;0x20uy;
//      0x61uy;0x6cuy;0x6cuy;0x20uy;0x66uy;0x72uy;0x6fuy;0x6duy;0x20uy;0x4euy;0x6fuy;
//      0x76uy;0x61uy;0x22uy;0x2cuy;0x22uy;0x69uy;0x64uy;0x22uy;0x3auy;0x22uy;0x39uy;
//      0x36uy;0x33uy;0x63uy;0x38uy;0x64uy;0x33uy;0x36uy;0x64uy;0x61uy;0x38uy;0x35uy;
//      0x34uy;0x64uy;0x36uy;0x34uy;0x62uy;0x38uy;0x39uy;0x64uy;0x38uy;0x35uy;0x32uy;
//      0x66uy;0x62uy;0x66uy;0x37uy;0x33uy;0x39uy;0x64uy;0x33uy;0x61uy;0x22uy;0x7duy;
//      0x00uy;0x82uy;0x04uy;0x00uy;0x00uy;0x7auy;0x46uy;0x31uy;0x35uy;0x63uy;0x30uy;
//      0x30uy;0x37uy;0x36uy;0x34uy;0x64uy;0x30uy;0x38uy;0x66uy;0x34uy;0x62uy;0x61uy;
//      0x33uy;0x62uy;0x39uy;0x32uy;0x65uy;0x62uy;0x34uy;0x61uy;0x35uy;0x63uy;0x38uy;
//      0x35uy;0x34uy;0x33uy;0x35uy;0x66uy;0x65uy;0x3auy;0x32uy;0x31uy;0xaeuy;0x03uy;
//      0x7buy;0x22uy;0x72uy;0x65uy;0x74uy;0x61uy;0x69uy;0x6cuy;0x53uy;0x6buy;0x75uy;
//      0x49uy;0x64uy;0x22uy;0x3auy;0x22uy;0x31uy;0x35uy;0x63uy;0x30uy;0x30uy;0x37uy;
//      0x36uy;0x34uy;0x64uy;0x30uy;0x38uy;0x66uy;0x34uy;0x62uy;0x61uy;0x33uy;0x62uy;
//      0x39uy;0x32uy;0x65uy;0x62uy;0x34uy;0x61uy;0x35uy;0x63uy;0x38uy;0x35uy;0x34uy;
//      0x33uy;0x35uy;0x66uy;0x65uy;0x22uy;0x2cuy;0x22uy;0x6duy;0x61uy;0x72uy;0x74uy;
//      0x49uy;0x64uy;0x22uy;0x3auy;0x32uy;0x31uy;0x2cuy;0x22uy;0x72uy;0x65uy;0x63uy;
//      0x61uy;0x6cuy;0x63uy;0x75uy;0x6cuy;0x61uy;0x74uy;0x69uy;0x6fuy;0x6euy;0x54uy;
//      0x68uy;0x72uy;0x65uy;0x73uy;0x68uy;0x6fuy;0x6cuy;0x64uy;0x22uy;0x3auy;0x30uy;
//      0x2cuy;0x22uy;0x74uy;0x69uy;0x6duy;0x65uy;0x73uy;0x74uy;0x61uy;0x6duy;0x70uy;
//      0x22uy;0x3auy;0x22uy;0x32uy;0x30uy;0x31uy;0x38uy;0x2duy;0x30uy;0x32uy;0x2duy;
//      0x30uy;0x32uy;0x54uy;0x32uy;0x32uy;0x3auy;0x30uy;0x36uy;0x3auy;0x31uy;0x37uy;
//      0x5auy;0x22uy;0x2cuy;0x22uy;0x72uy;0x65uy;0x61uy;0x73uy;0x6fuy;0x6euy;0x22uy;
//      0x3auy;0x22uy;0x45uy;0x76uy;0x65uy;0x6euy;0x74uy;0x22uy;0x2cuy;0x22uy;0x63uy;
//      0x6fuy;0x6duy;0x6duy;0x65uy;0x6euy;0x74uy;0x22uy;0x3auy;0x22uy;0x52uy;0x65uy;
//      0x70uy;0x72uy;0x69uy;0x63uy;0x65uy;0x20uy;0x61uy;0x6cuy;0x6cuy;0x20uy;0x66uy;
//      0x72uy;0x6fuy;0x6duy;0x20uy;0x4euy;0x6fuy;0x76uy;0x61uy;0x22uy;0x2cuy;0x22uy;
//      0x69uy;0x64uy;0x22uy;0x3auy;0x22uy;0x37uy;0x66uy;0x65uy;0x35uy;0x63uy;0x39uy;
//      0x32uy;0x66uy;0x65uy;0x36uy;0x36uy;0x34uy;0x34uy;0x62uy;0x62uy;0x64uy;0x62uy;
//      0x61uy;0x63uy;0x64uy;0x36uy;0x64uy;0x62uy;0x66uy;0x37uy;0x32uy;0x39uy;0x31uy;
//      0x31uy;0x61uy;0x61uy;0x33uy;0x22uy;0x7duy;0x00uy;0x82uy;0x04uy;0x00uy;0x00uy;
//      0x7cuy;0x46uy;0x31uy;0x34uy;0x64uy;0x30uy;0x30uy;0x37uy;0x33uy;0x63uy;0x63uy;
//      0x62uy;0x37uy;0x36uy;0x34uy;0x30uy;0x66uy;0x61uy;0x62uy;0x64uy;0x64uy;0x62uy;
//      0x36uy;0x37uy;0x37uy;0x30uy;0x63uy;0x62uy;0x38uy;0x33uy;0x36uy;0x34uy;0x36uy;
//      0x37uy;0x3auy;0x32uy;0x31uy;0xaeuy;0x03uy;0x7buy;0x22uy;0x72uy;0x65uy;0x74uy;
//      0x61uy;0x69uy;0x6cuy;0x53uy;0x6buy;0x75uy;0x49uy;0x64uy;0x22uy;0x3auy;0x22uy;
//      0x31uy;0x34uy;0x64uy;0x30uy;0x30uy;0x37uy;0x33uy;0x63uy;0x63uy;0x62uy;0x37uy;
//      0x36uy;0x34uy;0x30uy;0x66uy;0x61uy;0x62uy;0x64uy;0x64uy;0x62uy;0x36uy;0x37uy;
//      0x37uy;0x30uy;0x63uy;0x62uy;0x38uy;0x33uy;0x36uy;0x34uy;0x36uy;0x37uy;0x22uy;
//      0x2cuy;0x22uy;0x6duy;0x61uy;0x72uy;0x74uy;0x49uy;0x64uy;0x22uy;0x3auy;0x32uy;
//      0x31uy;0x2cuy;0x22uy;0x72uy;0x65uy;0x63uy;0x61uy;0x6cuy;0x63uy;0x75uy;0x6cuy;
//      0x61uy;0x74uy;0x69uy;0x6fuy;0x6euy;0x54uy;0x68uy;0x72uy;0x65uy;0x73uy;0x68uy;
//      0x6fuy;0x6cuy;0x64uy;0x22uy;0x3auy;0x30uy;0x2cuy;0x22uy;0x74uy;0x69uy;0x6duy;
//      0x65uy;0x73uy;0x74uy;0x61uy;0x6duy;0x70uy;0x22uy;0x3auy;0x22uy;0x32uy;0x30uy;
//      0x31uy;0x38uy;0x2duy;0x30uy;0x32uy;0x2duy;0x30uy;0x32uy;0x54uy;0x32uy;0x32uy;
//      0x3auy;0x30uy;0x36uy;0x3auy;0x31uy;0x37uy;0x5auy;0x22uy;0x2cuy;0x22uy;0x72uy;
//      0x65uy;0x61uy;0x73uy;0x6fuy;0x6euy;0x22uy;0x3auy;0x22uy;0x45uy;0x76uy;0x65uy;
//      0x6euy;0x74uy;0x22uy;0x2cuy;0x22uy;0x63uy;0x6fuy;0x6duy;0x6duy;0x65uy;0x6euy;
//      0x74uy;0x22uy;0x3auy;0x22uy;0x52uy;0x65uy;0x70uy;0x72uy;0x69uy;0x63uy;0x65uy;
//      0x20uy;0x61uy;0x6cuy;0x6cuy;0x20uy;0x66uy;0x72uy;0x6fuy;0x6duy;0x20uy;0x4euy;
//      0x6fuy;0x76uy;0x61uy;0x22uy;0x2cuy;0x22uy;0x69uy;0x64uy;0x22uy;0x3auy;0x22uy;
//      0x63uy;0x33uy;0x61uy;0x38uy;0x39uy;0x65uy;0x64uy;0x65uy;0x33uy;0x31uy;0x36uy;
//      0x31uy;0x34uy;0x31uy;0x31uy;0x30uy;0x61uy;0x65uy;0x32uy;0x36uy;0x30uy;0x66uy;
//      0x34uy;0x63uy;0x30uy;0x35uy;0x39uy;0x31uy;0x63uy;0x66uy;0x30uy;0x63uy;0x22uy;
//      0x7duy;0x00uy;0x82uy;0x04uy;0x00uy;0x00uy;0x7euy;0x46uy;0x31uy;0x33uy;0x63uy;
//      0x30uy;0x31uy;0x66uy;0x31uy;0x64uy;0x35uy;0x62uy;0x62uy;0x35uy;0x34uy;0x31uy;
//      0x64uy;0x65uy;0x39uy;0x66uy;0x66uy;0x39uy;0x63uy;0x30uy;0x34uy;0x62uy;0x37uy;
//      0x39uy;0x39uy;0x32uy;0x32uy;0x63uy;0x63uy;0x32uy;0x3auy;0x32uy;0x31uy;0xaeuy;
//      0x03uy;0x7buy;0x22uy;0x72uy;0x65uy;0x74uy;0x61uy;0x69uy;0x6cuy;0x53uy;0x6buy;
//      0x75uy;0x49uy;0x64uy;0x22uy;0x3auy;0x22uy;0x31uy;0x33uy;0x63uy;0x30uy;0x31uy;
//      0x66uy;0x31uy;0x64uy;0x35uy;0x62uy;0x62uy;0x35uy;0x34uy;0x31uy;0x64uy;0x65uy;
//      0x39uy;0x66uy;0x66uy;0x39uy;0x63uy;0x30uy;0x34uy;0x62uy;0x37uy;0x39uy;0x39uy;
//      0x32uy;0x32uy;0x63uy;0x63uy;0x32uy;0x22uy;0x2cuy;0x22uy;0x6duy;0x61uy;0x72uy;
//      0x74uy;0x49uy;0x64uy;0x22uy;0x3auy;0x32uy;0x31uy;0x2cuy;0x22uy;0x72uy;0x65uy;
//      0x63uy;0x61uy;0x6cuy;0x63uy;0x75uy;0x6cuy;0x61uy;0x74uy;0x69uy;0x6fuy;0x6euy;
//      0x54uy;0x68uy;0x72uy;0x65uy;0x73uy;0x68uy;0x6fuy;0x6cuy;0x64uy;0x22uy;0x3auy;
//      0x30uy;0x2cuy;0x22uy;0x74uy;0x69uy;0x6duy;0x65uy;0x73uy;0x74uy;0x61uy;0x6duy;
//      0x70uy;0x22uy;0x3auy;0x22uy;0x32uy;0x30uy;0x31uy;0x38uy;0x2duy;0x30uy;0x32uy;
//      0x2duy;0x30uy;0x32uy;0x54uy;0x32uy;0x32uy;0x3auy;0x30uy;0x36uy;0x3auy;0x31uy;
//      0x37uy;0x5auy;0x22uy;0x2cuy;0x22uy;0x72uy;0x65uy;0x61uy;0x73uy;0x6fuy;0x6euy;
//      0x22uy;0x3auy;0x22uy;0x45uy;0x76uy;0x65uy;0x6euy;0x74uy;0x22uy;0x2cuy;0x22uy;
//      0x63uy;0x6fuy;0x6duy;0x6duy;0x65uy;0x6euy;0x74uy;0x22uy;0x3auy;0x22uy;0x52uy;
//      0x65uy;0x70uy;0x72uy;0x69uy;0x63uy;0x65uy;0x20uy;0x61uy;0x6cuy;0x6cuy;0x20uy;
//      0x66uy;0x72uy;0x6fuy;0x6duy;0x20uy;0x4euy;0x6fuy;0x76uy;0x61uy;0x22uy;0x2cuy;
//      0x22uy;0x69uy;0x64uy;0x22uy;0x3auy;0x22uy;0x64uy;0x37uy;0x62uy;0x64uy;0x36uy;
//      0x38uy;0x37uy;0x36uy;0x34uy;0x37uy;0x33uy;0x62uy;0x34uy;0x30uy;0x35uy;0x38uy;
//      0x39uy;0x33uy;0x62uy;0x35uy;0x36uy;0x34uy;0x39uy;0x37uy;0x30uy;0x32uy;0x64uy;
//      0x62uy;0x30uy;0x30uy;0x38uy;0x32uy;0x22uy;0x7duy;0x00uy;0x84uy;0x04uy;0x00uy;
//      0x00uy;0x80uy;0x01uy;0x46uy;0x31uy;0x66uy;0x61uy;0x30uy;0x31uy;0x39uy;0x65uy;
//      0x31uy;0x37uy;0x65uy;0x35uy;0x63uy;0x34uy;0x31uy;0x39uy;0x35uy;0x61uy;0x30uy;
//      0x34uy;0x30uy;0x38uy;0x36uy;0x37uy;0x61uy;0x63uy;0x62uy;0x63uy;0x66uy;0x64uy;
//      0x30uy;0x38uy;0x65uy;0x3auy;0x32uy;0x31uy;0xaeuy;0x03uy;0x7buy;0x22uy;0x72uy;
//      0x65uy;0x74uy;0x61uy;0x69uy;0x6cuy;0x53uy;0x6buy;0x75uy;0x49uy;0x64uy;0x22uy;
//      0x3auy;0x22uy;0x31uy;0x66uy;0x61uy;0x30uy;0x31uy;0x39uy;0x65uy;0x31uy;0x37uy;
//      0x65uy;0x35uy;0x63uy;0x34uy;0x31uy;0x39uy;0x35uy;0x61uy;0x30uy;0x34uy;0x30uy;
//      0x38uy;0x36uy;0x37uy;0x61uy;0x63uy;0x62uy;0x63uy;0x66uy;0x64uy;0x30uy;0x38uy;
//      0x65uy;0x22uy;0x2cuy;0x22uy;0x6duy;0x61uy;0x72uy;0x74uy;0x49uy;0x64uy;0x22uy;
//      0x3auy;0x32uy;0x31uy;0x2cuy;0x22uy;0x72uy;0x65uy;0x63uy;0x61uy;0x6cuy;0x63uy;
//      0x75uy;0x6cuy;0x61uy;0x74uy;0x69uy;0x6fuy;0x6euy;0x54uy;0x68uy;0x72uy;0x65uy;
//      0x73uy;0x68uy;0x6fuy;0x6cuy;0x64uy;0x22uy;0x3auy;0x30uy;0x2cuy;0x22uy;0x74uy;
//      0x69uy;0x6duy;0x65uy;0x73uy;0x74uy;0x61uy;0x6duy;0x70uy;0x22uy;0x3auy;0x22uy;
//      0x32uy;0x30uy;0x31uy;0x38uy;0x2duy;0x30uy;0x32uy;0x2duy;0x30uy;0x32uy;0x54uy;
//      0x32uy;0x32uy;0x3auy;0x30uy;0x36uy;0x3auy;0x31uy;0x38uy;0x5auy;0x22uy;0x2cuy;
//      0x22uy;0x72uy;0x65uy;0x61uy;0x73uy;0x6fuy;0x6euy;0x22uy;0x3auy;0x22uy;0x45uy;
//      0x76uy;0x65uy;0x6euy;0x74uy;0x22uy;0x2cuy;0x22uy;0x63uy;0x6fuy;0x6duy;0x6duy;
//      0x65uy;0x6euy;0x74uy;0x22uy;0x3auy;0x22uy;0x52uy;0x65uy;0x70uy;0x72uy;0x69uy;
//      0x63uy;0x65uy;0x20uy;0x61uy;0x6cuy;0x6cuy;0x20uy;0x66uy;0x72uy;0x6fuy;0x6duy;
//      0x20uy;0x4euy;0x6fuy;0x76uy;0x61uy;0x22uy;0x2cuy;0x22uy;0x69uy;0x64uy;0x22uy;
//      0x3auy;0x22uy;0x35uy;0x36uy;0x38uy;0x37uy;0x66uy;0x63uy;0x66uy;0x38uy;0x64uy;
//      0x37uy;0x32uy;0x36uy;0x34uy;0x32uy;0x66uy;0x63uy;0x62uy;0x62uy;0x33uy;0x34uy;
//      0x35uy;0x32uy;0x30uy;0x62uy;0x30uy;0x34uy;0x65uy;0x34uy;0x32uy;0x62uy;0x66uy;
//      0x64uy;0x22uy;0x7duy;0x00uy;0x84uy;0x04uy;0x00uy;0x00uy;0x82uy;0x01uy;0x46uy;
//      0x31uy;0x35uy;0x39uy;0x30uy;0x31uy;0x61uy;0x65uy;0x39uy;0x35uy;0x31uy;0x35uy;
//      0x62uy;0x34uy;0x35uy;0x64uy;0x37uy;0x38uy;0x39uy;0x36uy;0x38uy;0x32uy;0x31uy;
//      0x63uy;0x39uy;0x61uy;0x30uy;0x36uy;0x39uy;0x61uy;0x33uy;0x31uy;0x31uy;0x3auy;
//      0x32uy;0x31uy;0xaeuy;0x03uy;0x7buy;0x22uy;0x72uy;0x65uy;0x74uy;0x61uy;0x69uy;
//      0x6cuy;0x53uy;0x6buy;0x75uy;0x49uy;0x64uy;0x22uy;0x3auy;0x22uy;0x31uy;0x35uy;
//      0x39uy;0x30uy;0x31uy;0x61uy;0x65uy;0x39uy;0x35uy;0x31uy;0x35uy;0x62uy;0x34uy;
//      0x35uy;0x64uy;0x37uy;0x38uy;0x39uy;0x36uy;0x38uy;0x32uy;0x31uy;0x63uy;0x39uy;
//      0x61uy;0x30uy;0x36uy;0x39uy;0x61uy;0x33uy;0x31uy;0x31uy;0x22uy;0x2cuy;0x22uy;
//      0x6duy;0x61uy;0x72uy;0x74uy;0x49uy;0x64uy;0x22uy;0x3auy;0x32uy;0x31uy;0x2cuy;
//      0x22uy;0x72uy;0x65uy;0x63uy;0x61uy;0x6cuy;0x63uy;0x75uy;0x6cuy;0x61uy;0x74uy;
//      0x69uy;0x6fuy;0x6euy;0x54uy;0x68uy;0x72uy;0x65uy;0x73uy;0x68uy;0x6fuy;0x6cuy;
//      0x64uy;0x22uy;0x3auy;0x30uy;0x2cuy;0x22uy;0x74uy;0x69uy;0x6duy;0x65uy;0x73uy;
//      0x74uy;0x61uy;0x6duy;0x70uy;0x22uy;0x3auy;0x22uy;0x32uy;0x30uy;0x31uy;0x38uy;
//      0x2duy;0x30uy;0x32uy;0x2duy;0x30uy;0x32uy;0x54uy;0x32uy;0x32uy;0x3auy;0x30uy;
//      0x36uy;0x3auy;0x31uy;0x38uy;0x5auy;0x22uy;0x2cuy;0x22uy;0x72uy;0x65uy;0x61uy;
//      0x73uy;0x6fuy;0x6euy;0x22uy;0x3auy;0x22uy;0x45uy;0x76uy;0x65uy;0x6euy;0x74uy;
//      0x22uy;0x2cuy;0x22uy;0x63uy;0x6fuy;0x6duy;0x6duy;0x65uy;0x6euy;0x74uy;0x22uy;
//      0x3auy;0x22uy;0x52uy;0x65uy;0x70uy;0x72uy;0x69uy;0x63uy;0x65uy;0x20uy;0x61uy;
//      0x6cuy;0x6cuy;0x20uy;0x66uy;0x72uy;0x6fuy;0x6duy;0x20uy;0x4euy;0x6fuy;0x76uy;
//      0x61uy;0x22uy;0x2cuy;0x22uy;0x69uy;0x64uy;0x22uy;0x3auy;0x22uy;0x31uy;0x31uy;
//      0x37uy;0x31uy;0x36uy;0x66uy;0x35uy;0x34uy;0x31uy;0x64uy;0x65uy;0x34uy;0x34uy;
//      0x62uy;0x37uy;0x32uy;0x38uy;0x64uy;0x35uy;0x66uy;0x36uy;0x39uy;0x30uy;0x37uy;
//      0x62uy;0x36uy;0x31uy;0x33uy;0x32uy;0x30uy;0x38uy;0x33uy;0x22uy;0x7duy;0x00uy;
//      0x84uy;0x04uy;0x00uy;0x00uy;0x84uy;0x01uy;0x46uy;0x31uy;0x38uy;0x38uy;0x30uy;
//      0x31uy;0x37uy;0x34uy;0x38uy;0x36uy;0x61uy;0x62uy;0x66uy;0x34uy;0x38uy;0x34uy;
//      0x66uy;0x61uy;0x39uy;0x37uy;0x32uy;0x63uy;0x38uy;0x63uy;0x64uy;0x63uy;0x32uy;
//      0x38uy;0x63uy;0x39uy;0x39uy;0x38uy;0x64uy;0x3auy;0x32uy;0x31uy;0xaeuy;0x03uy;
//      0x7buy;0x22uy;0x72uy;0x65uy;0x74uy;0x61uy;0x69uy;0x6cuy;0x53uy;0x6buy;0x75uy;
//      0x49uy;0x64uy;0x22uy;0x3auy;0x22uy;0x31uy;0x38uy;0x38uy;0x30uy;0x31uy;0x37uy;
//      0x34uy;0x38uy;0x36uy;0x61uy;0x62uy;0x66uy;0x34uy;0x38uy;0x34uy;0x66uy;0x61uy;
//      0x39uy;0x37uy;0x32uy;0x63uy;0x38uy;0x63uy;0x64uy;0x63uy;0x32uy;0x38uy;0x63uy;
//      0x39uy;0x39uy;0x38uy;0x64uy;0x22uy;0x2cuy;0x22uy;0x6duy;0x61uy;0x72uy;0x74uy;
//      0x49uy;0x64uy;0x22uy;0x3auy;0x32uy;0x31uy;0x2cuy;0x22uy;0x72uy;0x65uy;0x63uy;
//      0x61uy;0x6cuy;0x63uy;0x75uy;0x6cuy;0x61uy;0x74uy;0x69uy;0x6fuy;0x6euy;0x54uy;
//      0x68uy;0x72uy;0x65uy;0x73uy;0x68uy;0x6fuy;0x6cuy;0x64uy;0x22uy;0x3auy;0x30uy;
//      0x2cuy;0x22uy;0x74uy;0x69uy;0x6duy;0x65uy;0x73uy;0x74uy;0x61uy;0x6duy;0x70uy;
//      0x22uy;0x3auy;0x22uy;0x32uy;0x30uy;0x31uy;0x38uy;0x2duy;0x30uy;0x32uy;0x2duy;
//      0x30uy;0x32uy;0x54uy;0x32uy;0x32uy;0x3auy;0x30uy;0x36uy;0x3auy;0x31uy;0x38uy;
//      0x5auy;0x22uy;0x2cuy;0x22uy;0x72uy;0x65uy;0x61uy;0x73uy;0x6fuy;0x6euy;0x22uy;
//      0x3auy;0x22uy;0x45uy;0x76uy;0x65uy;0x6euy;0x74uy;0x22uy;0x2cuy;0x22uy;0x63uy;
//      0x6fuy;0x6duy;0x6duy;0x65uy;0x6euy;0x74uy;0x22uy;0x3auy;0x22uy;0x52uy;0x65uy;
//      0x70uy;0x72uy;0x69uy;0x63uy;0x65uy;0x20uy;0x61uy;0x6cuy;0x6cuy;0x20uy;0x66uy;
//      0x72uy;0x6fuy;0x6duy;0x20uy;0x4euy;0x6fuy;0x76uy;0x61uy;0x22uy;0x2cuy;0x22uy;
//      0x69uy;0x64uy;0x22uy;0x3auy;0x22uy;0x36uy;0x35uy;0x63uy;0x65uy;0x65uy;0x39uy;
//      0x62uy;0x30uy;0x66uy;0x66uy;0x37uy;0x65uy;0x34uy;0x66uy;0x30uy;0x39uy;0x38uy;
//      0x61uy;0x30uy;0x30uy;0x65uy;0x62uy;0x38uy;0x36uy;0x65uy;0x31uy;0x66uy;0x63uy;
//      0x30uy;0x64uy;0x66uy;0x65uy;0x22uy;0x7duy;0x00uy;0x84uy;0x04uy;0x00uy;0x00uy;
//      0x86uy;0x01uy;0x46uy;0x31uy;0x62uy;0x38uy;0x30uy;0x30uy;0x66uy;0x30uy;0x32uy;
//      0x65uy;0x33uy;0x32uy;0x37uy;0x34uy;0x61uy;0x30uy;0x33uy;0x62uy;0x63uy;0x63uy;
//      0x65uy;0x66uy;0x30uy;0x34uy;0x65uy;0x37uy;0x31uy;0x37uy;0x61uy;0x31uy;0x64uy;
//      0x34uy;0x65uy;0x3auy;0x32uy;0x31uy;0xaeuy;0x03uy;0x7buy;0x22uy;0x72uy;0x65uy;
//      0x74uy;0x61uy;0x69uy;0x6cuy;0x53uy;0x6buy;0x75uy;0x49uy;0x64uy;0x22uy;0x3auy;
//      0x22uy;0x31uy;0x62uy;0x38uy;0x30uy;0x30uy;0x66uy;0x30uy;0x32uy;0x65uy;0x33uy;
//      0x32uy;0x37uy;0x34uy;0x61uy;0x30uy;0x33uy;0x62uy;0x63uy;0x63uy;0x65uy;0x66uy;
//      0x30uy;0x34uy;0x65uy;0x37uy;0x31uy;0x37uy;0x61uy;0x31uy;0x64uy;0x34uy;0x65uy;
//      0x22uy;0x2cuy;0x22uy;0x6duy;0x61uy;0x72uy;0x74uy;0x49uy;0x64uy;0x22uy;0x3auy;
//      0x32uy;0x31uy;0x2cuy;0x22uy;0x72uy;0x65uy;0x63uy;0x61uy;0x6cuy;0x63uy;0x75uy;
//      0x6cuy;0x61uy;0x74uy;0x69uy;0x6fuy;0x6euy;0x54uy;0x68uy;0x72uy;0x65uy;0x73uy;
//      0x68uy;0x6fuy;0x6cuy;0x64uy;0x22uy;0x3auy;0x30uy;0x2cuy;0x22uy;0x74uy;0x69uy;
//      0x6duy;0x65uy;0x73uy;0x74uy;0x61uy;0x6duy;0x70uy;0x22uy;0x3auy;0x22uy;0x32uy;
//      0x30uy;0x31uy;0x38uy;0x2duy;0x30uy;0x32uy;0x2duy;0x30uy;0x32uy;0x54uy;0x32uy;
//      0x32uy;0x3auy;0x30uy;0x36uy;0x3auy;0x31uy;0x38uy;0x5auy;0x22uy;0x2cuy;0x22uy;
//      0x72uy;0x65uy;0x61uy;0x73uy;0x6fuy;0x6euy;0x22uy;0x3auy;0x22uy;0x45uy;0x76uy;
//      0x65uy;0x6euy;0x74uy;0x22uy;0x2cuy;0x22uy;0x63uy;0x6fuy;0x6duy;0x6duy;0x65uy;
//      0x6euy;0x74uy;0x22uy;0x3auy;0x22uy;0x52uy;0x65uy;0x70uy;0x72uy;0x69uy;0x63uy;
//      0x65uy;0x20uy;0x61uy;0x6cuy;0x6cuy;0x20uy;0x66uy;0x72uy;0x6fuy;0x6duy;0x20uy;
//      0x4euy;0x6fuy;0x76uy;0x61uy;0x22uy;0x2cuy;0x22uy;0x69uy;0x64uy;0x22uy;0x3auy;
//      0x22uy;0x37uy;0x66uy;0x66uy;0x66uy;0x31uy;0x33uy;0x66uy;0x61uy;0x37uy;0x64uy;
//      0x64uy;0x37uy;0x34uy;0x38uy;0x62uy;0x36uy;0x38uy;0x62uy;0x32uy;0x32uy;0x34uy;
//      0x63uy;0x30uy;0x32uy;0x65uy;0x31uy;0x39uy;0x32uy;0x64uy;0x38uy;0x37uy;0x32uy;
//      0x22uy;0x7duy;0x00uy;0x84uy;0x04uy;0x00uy;0x00uy;0x88uy;0x01uy;0x46uy;0x31uy;
//      0x64uy;0x61uy;0x30uy;0x30uy;0x63uy;0x31uy;0x66uy;0x63uy;0x31uy;0x62uy;0x66uy;
//      0x34uy;0x62uy;0x35uy;0x33uy;0x38uy;0x65uy;0x61uy;0x63uy;0x63uy;0x38uy;0x35uy;
//      0x35uy;0x32uy;0x36uy;0x64uy;0x64uy;0x65uy;0x63uy;0x37uy;0x61uy;0x3auy;0x32uy;
//      0x31uy;0xaeuy;0x03uy;0x7buy;0x22uy;0x72uy;0x65uy;0x74uy;0x61uy;0x69uy;0x6cuy;
//      0x53uy;0x6buy;0x75uy;0x49uy;0x64uy;0x22uy;0x3auy;0x22uy;0x31uy;0x64uy;0x61uy;
//      0x30uy;0x30uy;0x63uy;0x31uy;0x66uy;0x63uy;0x31uy;0x62uy;0x66uy;0x34uy;0x62uy;
//      0x35uy;0x33uy;0x38uy;0x65uy;0x61uy;0x63uy;0x63uy;0x38uy;0x35uy;0x35uy;0x32uy;
//      0x36uy;0x64uy;0x64uy;0x65uy;0x63uy;0x37uy;0x61uy;0x22uy;0x2cuy;0x22uy;0x6duy;
//      0x61uy;0x72uy;0x74uy;0x49uy;0x64uy;0x22uy;0x3auy;0x32uy;0x31uy;0x2cuy;0x22uy;
//      0x72uy;0x65uy;0x63uy;0x61uy;0x6cuy;0x63uy;0x75uy;0x6cuy;0x61uy;0x74uy;0x69uy;
//      0x6fuy;0x6euy;0x54uy;0x68uy;0x72uy;0x65uy;0x73uy;0x68uy;0x6fuy;0x6cuy;0x64uy;
//      0x22uy;0x3auy;0x30uy;0x2cuy;0x22uy;0x74uy;0x69uy;0x6duy;0x65uy;0x73uy;0x74uy;
//      0x61uy;0x6duy;0x70uy;0x22uy;0x3auy;0x22uy;0x32uy;0x30uy;0x31uy;0x38uy;0x2duy;
//      0x30uy;0x32uy;0x2duy;0x30uy;0x32uy;0x54uy;0x32uy;0x32uy;0x3auy;0x30uy;0x36uy;
//      0x3auy;0x31uy;0x38uy;0x5auy;0x22uy;0x2cuy;0x22uy;0x72uy;0x65uy;0x61uy;0x73uy;
//      0x6fuy;0x6euy;0x22uy;0x3auy;0x22uy;0x45uy;0x76uy;0x65uy;0x6euy;0x74uy;0x22uy;
//      0x2cuy;0x22uy;0x63uy;0x6fuy;0x6duy;0x6duy;0x65uy;0x6euy;0x74uy;0x22uy;0x3auy;
//      0x22uy;0x52uy;0x65uy;0x70uy;0x72uy;0x69uy;0x63uy;0x65uy;0x20uy;0x61uy;0x6cuy;
//      0x6cuy;0x20uy;0x66uy;0x72uy;0x6fuy;0x6duy;0x20uy;0x4euy;0x6fuy;0x76uy;0x61uy;
//      0x22uy;0x2cuy;0x22uy;0x69uy;0x64uy;0x22uy;0x3auy;0x22uy;0x38uy;0x35uy;0x38uy;
//      0x62uy;0x31uy;0x64uy;0x62uy;0x64uy;0x32uy;0x64uy;0x39uy;0x34uy;0x34uy;0x63uy;
//      0x61uy;0x37uy;0x39uy;0x33uy;0x39uy;0x39uy;0x64uy;0x66uy;0x66uy;0x63uy;0x36uy;
//      0x38uy;0x32uy;0x64uy;0x31uy;0x35uy;0x35uy;0x61uy;0x22uy;0x7duy;0x00uy;0x84uy;
//      0x04uy;0x00uy;0x00uy;0x8auy;0x01uy;0x46uy;0x31uy;0x63uy;0x64uy;0x30uy;0x30uy;
//      0x38uy;0x64uy;0x66uy;0x36uy;0x63uy;0x37uy;0x35uy;0x34uy;0x31uy;0x38uy;0x32uy;
//      0x39uy;0x39uy;0x63uy;0x37uy;0x31uy;0x63uy;0x64uy;0x36uy;0x37uy;0x62uy;0x63uy;
//      0x34uy;0x30uy;0x39uy;0x36uy;0x34uy;0x3auy;0x32uy;0x31uy;0xaeuy;0x03uy;0x7buy;
//      0x22uy;0x72uy;0x65uy;0x74uy;0x61uy;0x69uy;0x6cuy;0x53uy;0x6buy;0x75uy;0x49uy;
//      0x64uy;0x22uy;0x3auy;0x22uy;0x31uy;0x63uy;0x64uy;0x30uy;0x30uy;0x38uy;0x64uy;
//      0x66uy;0x36uy;0x63uy;0x37uy;0x35uy;0x34uy;0x31uy;0x38uy;0x32uy;0x39uy;0x39uy;
//      0x63uy;0x37uy;0x31uy;0x63uy;0x64uy;0x36uy;0x37uy;0x62uy;0x63uy;0x34uy;0x30uy;
//      0x39uy;0x36uy;0x34uy;0x22uy;0x2cuy;0x22uy;0x6duy;0x61uy;0x72uy;0x74uy;0x49uy;
//      0x64uy;0x22uy;0x3auy;0x32uy;0x31uy;0x2cuy;0x22uy;0x72uy;0x65uy;0x63uy;0x61uy;
//      0x6cuy;0x63uy;0x75uy;0x6cuy;0x61uy;0x74uy;0x69uy;0x6fuy;0x6euy;0x54uy;0x68uy;
//      0x72uy;0x65uy;0x73uy;0x68uy;0x6fuy;0x6cuy;0x64uy;0x22uy;0x3auy;0x30uy;0x2cuy;
//      0x22uy;0x74uy;0x69uy;0x6duy;0x65uy;0x73uy;0x74uy;0x61uy;0x6duy;0x70uy;0x22uy;
//      0x3auy;0x22uy;0x32uy;0x30uy;0x31uy;0x38uy;0x2duy;0x30uy;0x32uy;0x2duy;0x30uy;
//      0x32uy;0x54uy;0x32uy;0x32uy;0x3auy;0x30uy;0x36uy;0x3auy;0x31uy;0x38uy;0x5auy;
//      0x22uy;0x2cuy;0x22uy;0x72uy;0x65uy;0x61uy;0x73uy;0x6fuy;0x6euy;0x22uy;0x3auy;
//      0x22uy;0x45uy;0x76uy;0x65uy;0x6euy;0x74uy;0x22uy;0x2cuy;0x22uy;0x63uy;0x6fuy;
//      0x6duy;0x6duy;0x65uy;0x6euy;0x74uy;0x22uy;0x3auy;0x22uy;0x52uy;0x65uy;0x70uy;
//      0x72uy;0x69uy;0x63uy;0x65uy;0x20uy;0x61uy;0x6cuy;0x6cuy;0x20uy;0x66uy;0x72uy;
//      0x6fuy;0x6duy;0x20uy;0x4euy;0x6fuy;0x76uy;0x61uy;0x22uy;0x2cuy;0x22uy;0x69uy;
//      0x64uy;0x22uy;0x3auy;0x22uy;0x30uy;0x33uy;0x66uy;0x39uy;0x64uy;0x38uy;0x66uy;
//      0x34uy;0x36uy;0x38uy;0x38uy;0x31uy;0x34uy;0x36uy;0x36uy;0x38uy;0x38uy;0x36uy;
//      0x32uy;0x63uy;0x38uy;0x64uy;0x36uy;0x36uy;0x65uy;0x62uy;0x32uy;0x30uy;0x36uy;
//      0x65uy;0x35uy;0x65uy;0x22uy;0x7duy;0x00uy;0x84uy;0x04uy;0x00uy;0x00uy;0x8cuy;
//      0x01uy;0x46uy;0x31uy;0x31uy;0x39uy;0x30uy;0x31uy;0x36uy;0x39uy;0x65uy;0x36uy;
//      0x32uy;0x33uy;0x66uy;0x34uy;0x35uy;0x64uy;0x36uy;0x38uy;0x39uy;0x61uy;0x64uy;
//      0x62uy;0x65uy;0x31uy;0x34uy;0x31uy;0x34uy;0x61uy;0x63uy;0x39uy;0x31uy;0x39uy;
//      0x63uy;0x3auy;0x32uy;0x31uy;0xaeuy;0x03uy;0x7buy;0x22uy;0x72uy;0x65uy;0x74uy;
//      0x61uy;0x69uy;0x6cuy;0x53uy;0x6buy;0x75uy;0x49uy;0x64uy;0x22uy;0x3auy;0x22uy;
//      0x31uy;0x31uy;0x39uy;0x30uy;0x31uy;0x36uy;0x39uy;0x65uy;0x36uy;0x32uy;0x33uy;
//      0x66uy;0x34uy;0x35uy;0x64uy;0x36uy;0x38uy;0x39uy;0x61uy;0x64uy;0x62uy;0x65uy;
//      0x31uy;0x34uy;0x31uy;0x34uy;0x61uy;0x63uy;0x39uy;0x31uy;0x39uy;0x63uy;0x22uy;
//      0x2cuy;0x22uy;0x6duy;0x61uy;0x72uy;0x74uy;0x49uy;0x64uy;0x22uy;0x3auy;0x32uy;
//      0x31uy;0x2cuy;0x22uy;0x72uy;0x65uy;0x63uy;0x61uy;0x6cuy;0x63uy;0x75uy;0x6cuy;
//      0x61uy;0x74uy;0x69uy;0x6fuy;0x6euy;0x54uy;0x68uy;0x72uy;0x65uy;0x73uy;0x68uy;
//      0x6fuy;0x6cuy;0x64uy;0x22uy;0x3auy;0x30uy;0x2cuy;0x22uy;0x74uy;0x69uy;0x6duy;
//      0x65uy;0x73uy;0x74uy;0x61uy;0x6duy;0x70uy;0x22uy;0x3auy;0x22uy;0x32uy;0x30uy;
//      0x31uy;0x38uy;0x2duy;0x30uy;0x32uy;0x2duy;0x30uy;0x32uy;0x54uy;0x32uy;0x32uy;
//      0x3auy;0x30uy;0x36uy;0x3auy;0x31uy;0x38uy;0x5auy;0x22uy;0x2cuy;0x22uy;0x72uy;
//      0x65uy;0x61uy;0x73uy;0x6fuy;0x6euy;0x22uy;0x3auy;0x22uy;0x45uy;0x76uy;0x65uy;
//      0x6euy;0x74uy;0x22uy;0x2cuy;0x22uy;0x63uy;0x6fuy;0x6duy;0x6duy;0x65uy;0x6euy;
//      0x74uy;0x22uy;0x3auy;0x22uy;0x52uy;0x65uy;0x70uy;0x72uy;0x69uy;0x63uy;0x65uy;
//      0x20uy;0x61uy;0x6cuy;0x6cuy;0x20uy;0x66uy;0x72uy;0x6fuy;0x6duy;0x20uy;0x4euy;
//      0x6fuy;0x76uy;0x61uy;0x22uy;0x2cuy;0x22uy;0x69uy;0x64uy;0x22uy;0x3auy;0x22uy;
//      0x30uy;0x33uy;0x38uy;0x62uy;0x33uy;0x37uy;0x36uy;0x38uy;0x31uy;0x31uy;0x32uy;
//      0x63uy;0x34uy;0x37uy;0x37uy;0x31uy;0x38uy;0x31uy;0x64uy;0x34uy;0x31uy;0x33uy;
//      0x62uy;0x32uy;0x62uy;0x30uy;0x35uy;0x66uy;0x34uy;0x63uy;0x30uy;0x64uy;0x22uy;
//      0x7duy;0x00uy;0x84uy;0x04uy;0x00uy;0x00uy;0x8euy;0x01uy;0x46uy;0x31uy;0x62uy;
//      0x64uy;0x30uy;0x32uy;0x30uy;0x64uy;0x32uy;0x32uy;0x36uy;0x38uy;0x35uy;0x34uy;
//      0x34uy;0x63uy;0x37uy;0x61uy;0x63uy;0x63uy;0x62uy;0x66uy;0x61uy;0x36uy;0x35uy;
//      0x36uy;0x37uy;0x32uy;0x38uy;0x66uy;0x33uy;0x31uy;0x38uy;0x3auy;0x32uy;0x31uy;
//      0xaeuy;0x03uy;0x7buy;0x22uy;0x72uy;0x65uy;0x74uy;0x61uy;0x69uy;0x6cuy;0x53uy;
//      0x6buy;0x75uy;0x49uy;0x64uy;0x22uy;0x3auy;0x22uy;0x31uy;0x62uy;0x64uy;0x30uy;
//      0x32uy;0x30uy;0x64uy;0x32uy;0x32uy;0x36uy;0x38uy;0x35uy;0x34uy;0x34uy;0x63uy;
//      0x37uy;0x61uy;0x63uy;0x63uy;0x62uy;0x66uy;0x61uy;0x36uy;0x35uy;0x36uy;0x37uy;
//      0x32uy;0x38uy;0x66uy;0x33uy;0x31uy;0x38uy;0x22uy;0x2cuy;0x22uy;0x6duy;0x61uy;
//      0x72uy;0x74uy;0x49uy;0x64uy;0x22uy;0x3auy;0x32uy;0x31uy;0x2cuy;0x22uy;0x72uy;
//      0x65uy;0x63uy;0x61uy;0x6cuy;0x63uy;0x75uy;0x6cuy;0x61uy;0x74uy;0x69uy;0x6fuy;
//      0x6euy;0x54uy;0x68uy;0x72uy;0x65uy;0x73uy;0x68uy;0x6fuy;0x6cuy;0x64uy;0x22uy;
//      0x3auy;0x30uy;0x2cuy;0x22uy;0x74uy;0x69uy;0x6duy;0x65uy;0x73uy;0x74uy;0x61uy;
//      0x6duy;0x70uy;0x22uy;0x3auy;0x22uy;0x32uy;0x30uy;0x31uy;0x38uy;0x2duy;0x30uy;
//      0x32uy;0x2duy;0x30uy;0x32uy;0x54uy;0x32uy;0x32uy;0x3auy;0x30uy;0x36uy;0x3auy;
//      0x31uy;0x38uy;0x5auy;0x22uy;0x2cuy;0x22uy;0x72uy;0x65uy;0x61uy;0x73uy;0x6fuy;
//      0x6euy;0x22uy;0x3auy;0x22uy;0x45uy;0x76uy;0x65uy;0x6euy;0x74uy;0x22uy;0x2cuy;
//      0x22uy;0x63uy;0x6fuy;0x6duy;0x6duy;0x65uy;0x6euy;0x74uy;0x22uy;0x3auy;0x22uy;
//      0x52uy;0x65uy;0x70uy;0x72uy;0x69uy;0x63uy;0x65uy;0x20uy;0x61uy;0x6cuy;0x6cuy;
//      0x20uy;0x66uy;0x72uy;0x6fuy;0x6duy;0x20uy;0x4euy;0x6fuy;0x76uy;0x61uy;0x22uy;
//      0x2cuy;0x22uy;0x69uy;0x64uy;0x22uy;0x3auy;0x22uy;0x31uy;0x30uy;0x35uy;0x35uy;
//      0x39uy;0x31uy;0x64uy;0x36uy;0x39uy;0x65uy;0x32uy;0x39uy;0x34uy;0x35uy;0x61uy;
//      0x62uy;0x38uy;0x37uy;0x32uy;0x61uy;0x30uy;0x35uy;0x36uy;0x30uy;0x32uy;0x62uy;
//      0x32uy;0x61uy;0x31uy;0x33uy;0x39uy;0x34uy;0x22uy;0x7duy;0x00uy;0x84uy;0x04uy;
//      0x00uy;0x00uy;0x90uy;0x01uy;0x46uy;0x31uy;0x37uy;0x61uy;0x30uy;0x30uy;0x37uy;
//      0x61uy;0x36uy;0x64uy;0x36uy;0x35uy;0x30uy;0x34uy;0x33uy;0x37uy;0x31uy;0x62uy;
//      0x64uy;0x62uy;0x30uy;0x35uy;0x39uy;0x36uy;0x66uy;0x64uy;0x31uy;0x35uy;0x34uy;
//      0x36uy;0x65uy;0x63uy;0x63uy;0x3auy;0x32uy;0x31uy;0xaeuy;0x03uy;0x7buy;0x22uy;
//      0x72uy;0x65uy;0x74uy;0x61uy;0x69uy;0x6cuy;0x53uy;0x6buy;0x75uy;0x49uy;0x64uy;
//      0x22uy;0x3auy;0x22uy;0x31uy;0x37uy;0x61uy;0x30uy;0x30uy;0x37uy;0x61uy;0x36uy;
//      0x64uy;0x36uy;0x35uy;0x30uy;0x34uy;0x33uy;0x37uy;0x31uy;0x62uy;0x64uy;0x62uy;
//      0x30uy;0x35uy;0x39uy;0x36uy;0x66uy;0x64uy;0x31uy;0x35uy;0x34uy;0x36uy;0x65uy;
//      0x63uy;0x63uy;0x22uy;0x2cuy;0x22uy;0x6duy;0x61uy;0x72uy;0x74uy;0x49uy;0x64uy;
//      0x22uy;0x3auy;0x32uy;0x31uy;0x2cuy;0x22uy;0x72uy;0x65uy;0x63uy;0x61uy;0x6cuy;
//      0x63uy;0x75uy;0x6cuy;0x61uy;0x74uy;0x69uy;0x6fuy;0x6euy;0x54uy;0x68uy;0x72uy;
//      0x65uy;0x73uy;0x68uy;0x6fuy;0x6cuy;0x64uy;0x22uy;0x3auy;0x30uy;0x2cuy;0x22uy;
//      0x74uy;0x69uy;0x6duy;0x65uy;0x73uy;0x74uy;0x61uy;0x6duy;0x70uy;0x22uy;0x3auy;
//      0x22uy;0x32uy;0x30uy;0x31uy;0x38uy;0x2duy;0x30uy;0x32uy;0x2duy;0x30uy;0x32uy;
//      0x54uy;0x32uy;0x32uy;0x3auy;0x30uy;0x36uy;0x3auy;0x31uy;0x38uy;0x5auy;0x22uy;
//      0x2cuy;0x22uy;0x72uy;0x65uy;0x61uy;0x73uy;0x6fuy;0x6euy;0x22uy;0x3auy;0x22uy;
//      0x45uy;0x76uy;0x65uy;0x6euy;0x74uy;
//    |]
//    |> BinaryZipper

//  let res = FetchResponse.Read (2s, responseData)
//  let (topicName, topicData) = res.topics.[0]
//  let (partition, errorCode, highWatermark, _, _, _, messageSetSize, messageSet) = topicData.[0]
//  let firstMessage = messageSet.messages.[0]
//  let lastMessage = messageSet.messages.[messageSet.messages.Length - 1]


//  Assert.AreEqual("superman-member-price-request-shoebuy-p4", topicName)
//  Assert.AreEqual(1, partition)
//  Assert.AreEqual(0, errorCode)
//  Assert.AreEqual(1631, highWatermark)
//  Assert.AreEqual(72, messageSet.messages.Length)
//  Assert.AreEqual(64, firstMessage.offset)
//  Assert.AreEqual(135, lastMessage.offset)
//  Assert.AreEqual("106010ce8ff6459fb7cf47d4251380d1:21", System.Text.Encoding.UTF8.GetString(firstMessage.message.key.Array, firstMessage.message.key.Offset, firstMessage.message.key.Count))

//[<Test>]
let ``FetchResponse should read v2`` () =
  let file = @"C:\Users\eulerfx\Documents\GitHub\kafunk\tests\kafunk.Tests\FetchResponse_RecordBatch.bin"
  let buf = System.IO.File.ReadAllBytes(file) |> Binary.ofArray |> BinaryZipper
  let res = FetchResponse.Read (2s, buf)
  
  ()
  
  


//[<Test>]
//let ``ProduceResponse.read should decode ProduceResponse``() =
//  let data =
//    Binary.ofArray [|
//      0uy;0uy;0uy;1uy;0uy;4uy;116uy;101uy;115uy;116uy;0uy;0uy;0uy;1uy;0uy;0uy;
//      0uy;0uy;0uy;0uy;0uy;0uy;0uy;0uy;0uy;0uy;0uy;8uy; |]
//  let bz = BinaryZipper(data)
//  let (res:ProduceResponse) = ProduceResponse.Read (1s,bz)
//  let topicName, ps = let x = res.topics.[0] in x.topic, x.partitions
//  let p, ec, off, _ts = let x = ps.[0] in x.partition, x.errorCode, x.offset, x.timestamp
//  Assert.AreEqual("test", topicName)
//  Assert.AreEqual(0, p)
//  Assert.AreEqual(0s, ec)
//  Assert.AreEqual(8L, off)

//[<Test>]
//let ``ProduceRequest.write should encode ProduceRequest``() =
//  let req = 
//    ProduceRequest(RequiredAcks.Local)
//    ProduceRequest.ofMessageSetTopics 
//      [| "test", [| 0, (MessageSet.ofMessage (Message.ofBytes "hello world"B None)) |] |] RequiredAcks.Local 0
//  let data = toArraySeg ProduceRequest.size (fun x -> ProduceRequest.write (1s,x)) req |> Binary.toArray |> Array.toList
//  let expected = [
//    0uy;1uy;0uy;0uy;3uy;232uy;0uy;0uy;0uy;1uy;0uy;4uy;116uy;101uy;115uy;116uy;
//    0uy;0uy;0uy;1uy;0uy;0uy;0uy;0uy;0uy;0uy;0uy;37uy;0uy;0uy;0uy;0uy;0uy;0uy;
//    0uy;0uy;0uy;0uy;0uy;25uy;115uy;172uy;247uy;124uy;0uy;0uy;255uy;255uy;255uy;
//    255uy;0uy;0uy;0uy;11uy;104uy;101uy;108uy;108uy;111uy;32uy;119uy;111uy;
//    114uy;108uy;100uy; ]
//  Assert.True((expected = data))